@@ -1,22 +1,16 @@
-/******************************************************************************
- *                    _   _____   __________                                  *
- *                   | | / / _ | / __/_  __/     Visibility                   *
- *                   | |/ / __ |_\ \  / /          Across                     *
- *                   |___/_/ |_/___/ /_/       Space and Time                 *
- *                                                                            *
- * This file is part of VAST. It is subject to the license terms in the       *
- * LICENSE file found in the top-level directory of this distribution and at  *
- * http://vast.io/license. No part of VAST, including this file, may be       *
- * copied, modified, propagated, or distributed except according to the terms *
- * contained in the LICENSE file.                                             *
- ******************************************************************************/
+//    _   _____   __________
+//   | | / / _ | / __/_  __/     Visibility
+//   | |/ / __ |_\ \  / /          Across
+//   |___/_/ |_/___/ /_/       Space and Time
+//
+// SPDX-FileCopyrightText: (c) 2019 The VAST Contributors
+// SPDX-License-Identifier: BSD-3-Clause
 
 #include "vast/config.hpp"
 
 #if VAST_ENABLE_ARROW
 
 #  include "vast/arrow_table_slice.hpp"
-
 #  include "vast/arrow_table_slice_builder.hpp"
 #  include "vast/detail/byte_swap.hpp"
 #  include "vast/detail/narrow.hpp"
@@ -33,6 +27,7 @@
 #  include <arrow/ipc/api.h>
 
 #  include <type_traits>
+#  include <utility>
 
 namespace vast {
 
@@ -94,6 +89,33 @@ private:
   std::shared_ptr<arrow::Array> arr_;
 };
 
+class arrow_record_view
+  : public container_view<std::pair<std::string_view, data_view>> {
+public:
+  explicit arrow_record_view(record_type type, const arrow::StructArray& arr,
+                             int64_t row)
+    : type_{std::move(type)}, arr_{arr}, row_{row} {
+    // nop
+  }
+
+  value_type at(size_type i) const override {
+    const auto& field = type_.fields[i];
+    auto col = arr_.field(i);
+    VAST_ASSERT(col);
+    VAST_ASSERT(col->Equals(arr_.GetFieldByName(field.name)));
+    return {field.name, value_at(field.type, *col, row_)};
+  }
+
+  size_type size() const noexcept override {
+    return arr_.num_fields();
+  }
+
+private:
+  const record_type type_;
+  const arrow::StructArray& arr_;
+  const int64_t row_;
+};
+
 // -- decoding of Arrow column arrays ------------------------------------------
 
 // Safe ourselves redundant boilerplate code for dispatching to the visitor.
@@ -158,6 +180,12 @@ void decode(const type& t, const arrow::ListArray& arr, F& f) {
             kind(t));
 }
 
+template <class F>
+void decode(const type& t, const arrow::StructArray& arr, F& f) {
+  DECODE_TRY_DISPATCH(record);
+  VAST_WARN("{} expected to decode a record but got a {}", __func__, kind(t));
+}
+
 template <class F>
 void decode(const type& t, const arrow::Array& arr, F& f) {
   switch (arr.type_id()) {
@@ -183,6 +211,9 @@ void decode(const type& t, const arrow::Array& arr, F& f) {
     case arrow::Type::LIST: {
       return decode(t, static_cast<const arrow::ListArray&>(arr), f);
     }
+    case arrow::Type::STRUCT: {
+      return decode(t, static_cast<const arrow::StructArray&>(arr), f);
+    }
     // -- lift floating point values to real -----------------------------
     case arrow::Type::HALF_FLOAT: {
       using array_type = arrow::NumericArray<arrow::HalfFloatType>;
@@ -245,9 +276,8 @@ auto real_at = [](const auto& arr, int64_t row) {
   return static_cast<real>(arr.Value(row));
 };
 
-auto integer_at = [](const auto& arr, int64_t row) {
-  return static_cast<integer>(arr.Value(row));
-};
+auto integer_at
+  = [](const auto& arr, int64_t row) { return integer{arr.Value(row)}; };
 
 auto count_at = [](const auto& arr, int64_t row) {
   return static_cast<count>(arr.Value(row));
@@ -284,7 +314,7 @@ auto subnet_at(const arrow::FixedSizeBinaryArray& arr, int64_t row) {
 }
 
 auto timestamp_at(const arrow::TimestampArray& arr, int64_t row) {
-  auto ts_value = static_cast<integer>(arr.Value(row));
+  auto ts_value = static_cast<integer::value_type>(arr.Value(row));
   duration time_since_epoch{0};
   auto& ts_type = static_cast<const arrow::TimestampType&>(*arr.type());
   switch (ts_type.unit()) {
@@ -337,6 +367,12 @@ auto map_at(type key_type, type value_type, const arrow::ListArray& arr,
   return map_view_handle{map_view_ptr{std::move(ptr)}};
 }
 
+auto record_at(const record_type& type, const arrow::StructArray& arr,
+               int64_t row) {
+  auto ptr = caf::make_counted<arrow_record_view>(type, arr, row);
+  return record_view_handle{record_view_ptr{std::move(ptr)}};
+}
+
 class row_picker {
 public:
   row_picker(size_t row) : row_(detail::narrow_cast<int64_t>(row)) {
@@ -416,6 +452,12 @@ public:
     }
   }
 
+  void operator()(const arrow::StructArray& arr, const record_type& t) {
+    if (arr.IsNull(row_))
+      return;
+    result_ = record_at(t, arr, row_);
+  }
+
 private:
   data_view result_;
   int64_t row_;
@@ -509,6 +551,11 @@ public:
     }
   }
 
+  void operator()(const arrow::StructArray& arr, const record_type& t) {
+    apply(arr,
+          [&](const auto& arr, int64_t row) { return record_at(t, arr, row); });
+  }
+
 private:
   int64_t offset_;
   value_index& idx_;
@@ -659,8 +706,8 @@ template <class FlatBuffer>
 data_view arrow_table_slice<FlatBuffer>::at(table_slice::size_type row,
                                             table_slice::size_type column,
                                             const type& t) const {
-  VAST_ASSERT(state_.layout.at(*state_.layout.offset_from_index(column))->type
-              == t);
+  VAST_ASSERT(congruent(
+    state_.layout.at(*state_.layout.offset_from_index(column))->type, t));
   auto&& batch = record_batch();
   VAST_ASSERT(batch);
   auto array = batch->column(detail::narrow_cast<int>(column));
