@@ -1,29 +1,24 @@
-/******************************************************************************
- *                    _   _____   __________                                  *
- *                   | | / / _ | / __/_  __/     Visibility                   *
- *                   | |/ / __ |_\ \  / /          Across                     *
- *                   |___/_/ |_/___/ /_/       Space and Time                 *
- *                                                                            *
- * This file is part of VAST. It is subject to the license terms in the       *
- * LICENSE file found in the top-level directory of this distribution and at  *
- * http://vast.io/license. No part of VAST, including this file, may be       *
- * copied, modified, propagated, or distributed except according to the terms *
- * contained in the LICENSE file.                                             *
- ******************************************************************************/
+//    _   _____   __________
+//   | | / / _ | / __/_  __/     Visibility
+//   | |/ / __ |_\ \  / /          Across
+//   |___/_/ |_/___/ /_/       Space and Time
+//
+// SPDX-FileCopyrightText: (c) 2019 The VAST Contributors
+// SPDX-License-Identifier: BSD-3-Clause
 
 #include "vast/config.hpp"
 
 #if VAST_ENABLE_ARROW
 
-#  include "vast/arrow_table_slice_builder.hpp"
-
 #  include "vast/arrow_table_slice.hpp"
+#  include "vast/arrow_table_slice_builder.hpp"
 #  include "vast/detail/byte_swap.hpp"
 #  include "vast/detail/narrow.hpp"
 #  include "vast/detail/overload.hpp"
 #  include "vast/die.hpp"
 #  include "vast/fbs/table_slice.hpp"
 #  include "vast/fbs/utils.hpp"
+#  include "vast/logger.hpp"
 
 #  include <arrow/api.h>
 #  include <arrow/io/api.h>
@@ -61,17 +56,31 @@ struct primitive_column_builder_trait_base
 template <class T>
 struct column_builder_trait;
 
-#define PRIMITIVE_COLUMN_BUILDER_TRAIT(VastType, ArrowType)                    \
-  template <>                                                                  \
-  struct column_builder_trait<VastType>                                        \
-    : primitive_column_builder_trait_base<VastType, ArrowType> {}
+#  define PRIMITIVE_COLUMN_BUILDER_TRAIT(VastType, ArrowType)                  \
+    template <>                                                                \
+    struct column_builder_trait<VastType>                                      \
+      : primitive_column_builder_trait_base<VastType, ArrowType> {}
 
 PRIMITIVE_COLUMN_BUILDER_TRAIT(bool_type, arrow::BooleanType);
-PRIMITIVE_COLUMN_BUILDER_TRAIT(integer_type, arrow::Int64Type);
 PRIMITIVE_COLUMN_BUILDER_TRAIT(count_type, arrow::UInt64Type);
 PRIMITIVE_COLUMN_BUILDER_TRAIT(real_type, arrow::DoubleType);
 
-#undef PRIMITIVE_COLUMN_BUILDER_TRAIT
+#  undef PRIMITIVE_COLUMN_BUILDER_TRAIT
+
+template <>
+struct column_builder_trait<integer_type>
+  : column_builder_trait_base<integer_type, arrow::Int64Type> {
+  using super = column_builder_trait_base<integer_type, arrow::Int64Type>;
+
+  static auto make_arrow_type() {
+    return super::type_singleton();
+  }
+
+  static bool
+  append(typename super::BuilderType& builder, typename super::view_type x) {
+    return builder.Append(x.value).ok();
+  }
+};
 
 template <>
 struct column_builder_trait<time_type>
@@ -262,7 +271,8 @@ public:
     return result;
   }
 
-  std::shared_ptr<arrow::ArrayBuilder> arrow_builder() const override {
+  [[nodiscard]] std::shared_ptr<arrow::ArrayBuilder>
+  arrow_builder() const override {
     return arrow_builder_;
   }
 
@@ -276,16 +286,14 @@ private:
   std::shared_ptr<arrow_builder_type> arrow_builder_;
 };
 
-template <class SequenceType>
-class sequence_column_builder
-  : public arrow_table_slice_builder::column_builder {
+class list_column_builder : public arrow_table_slice_builder::column_builder {
 public:
   using arrow_builder_type = arrow::ListBuilder;
 
-  using data_type = typename type_traits<SequenceType>::data_type;
+  using data_type = typename type_traits<list_type>::data_type;
 
-  sequence_column_builder(arrow::MemoryPool* pool,
-                          std::unique_ptr<column_builder> nested)
+  list_column_builder(arrow::MemoryPool* pool,
+                      std::unique_ptr<column_builder> nested)
     : nested_(std::move(nested)) {
     reset(pool, nested_->arrow_builder());
   }
@@ -313,7 +321,8 @@ public:
     return result;
   }
 
-  std::shared_ptr<arrow::ArrayBuilder> arrow_builder() const override {
+  [[nodiscard]] std::shared_ptr<arrow::ArrayBuilder>
+  arrow_builder() const override {
     return arrow_builder_;
   }
 
@@ -376,7 +385,8 @@ public:
     return result;
   }
 
-  std::shared_ptr<arrow::ArrayBuilder> arrow_builder() const override {
+  [[nodiscard]] std::shared_ptr<arrow::ArrayBuilder>
+  arrow_builder() const override {
     return list_builder_;
   }
 
@@ -388,6 +398,61 @@ private:
   std::unique_ptr<column_builder> val_builder_;
 };
 
+class record_column_builder : public arrow_table_slice_builder::column_builder {
+public:
+  using data_type = view<record>;
+
+  record_column_builder(
+    arrow::MemoryPool* pool, std::shared_ptr<arrow::DataType> struct_type,
+    std::vector<std::unique_ptr<column_builder>>&& field_builders)
+    : field_builders_{std::move(field_builders)} {
+    auto fields = std::vector<std::shared_ptr<arrow::ArrayBuilder>>{};
+    fields.reserve(field_builders_.size());
+    for (auto& field_builder : field_builders_) {
+      auto underlying_builder = field_builder->arrow_builder();
+      VAST_ASSERT(underlying_builder);
+      fields.push_back(std::move(underlying_builder));
+    }
+    struct_builder_ = std::make_shared<arrow::StructBuilder>(struct_type, pool,
+                                                             std::move(fields));
+  }
+
+  bool add(data_view x) override {
+    if (caf::holds_alternative<view<caf::none_t>>(x)) {
+      auto status = struct_builder_->AppendNull();
+      return status.ok();
+    }
+    // Verify that we're actually holding a record.
+    auto* xptr = caf::get_if<data_type>(&x);
+    if (!xptr)
+      return false;
+    if (auto status = struct_builder_->Append(); !status.ok())
+      return false;
+    const auto& r = **xptr;
+    for (size_t i = 0; i < r.size(); ++i)
+      if (!field_builders_[i]->add(r.at(i).second))
+        return false;
+    return true;
+  }
+
+  std::shared_ptr<arrow::Array> finish() override {
+    std::shared_ptr<arrow::Array> result;
+    if (!struct_builder_->Finish(&result).ok())
+      die("failed to finish Arrow column builder");
+    return result;
+  }
+
+  [[nodiscard]] std::shared_ptr<arrow::ArrayBuilder>
+  arrow_builder() const override {
+    return struct_builder_;
+  }
+
+private:
+  std::shared_ptr<arrow::StructBuilder> struct_builder_;
+
+  std::vector<std::unique_ptr<column_builder>> field_builders_;
+};
+
 } // namespace
 
 // -- member types -------------------------------------------------------------
@@ -401,15 +466,12 @@ arrow_table_slice_builder::column_builder::make(const type& t,
                                                 arrow::MemoryPool* pool) {
   auto f = detail::overload{
     [=](const auto& x) -> std::unique_ptr<column_builder> {
-      using type = std::decay_t<decltype(x)>;
-      if constexpr (std::is_same_v<type, list_type>) {
-        auto nested = column_builder::make(x.value_type, pool);
-        return std::make_unique<sequence_column_builder<type>>(
-          pool, std::move(nested));
-      } else {
-        return std::make_unique<column_builder_impl_t<std::decay_t<decltype(x)>>>(
-          pool);
-      }
+      return std::make_unique<column_builder_impl_t<std::decay_t<decltype(x)>>>(
+        pool);
+    },
+    [=](const list_type& x) -> std::unique_ptr<column_builder> {
+      auto nested = column_builder::make(x.value_type, pool);
+      return std::make_unique<list_column_builder>(pool, std::move(nested));
     },
     [=](const map_type& x) -> std::unique_ptr<column_builder> {
       auto key_builder = column_builder::make(x.key_type, pool);
@@ -419,8 +481,13 @@ arrow_table_slice_builder::column_builder::make(const type& t,
                                                   std::move(key_builder),
                                                   std::move(value_builder));
     },
-    [=](const record_type&) -> std::unique_ptr<column_builder> {
-      die("expected flat layout!");
+    [=](const record_type& x) -> std::unique_ptr<column_builder> {
+      auto field_builders = std::vector<std::unique_ptr<column_builder>>{};
+      field_builders.reserve(x.fields.size());
+      for (const auto& field : x.fields)
+        field_builders.push_back(column_builder::make(field.type, pool));
+      return std::make_unique<record_column_builder>(pool, make_arrow_type(x),
+                                                     std::move(field_builders));
     },
     [=](const alias_type& x) -> std::unique_ptr<column_builder> {
       return column_builder::make(x.value_type, pool);
@@ -446,7 +513,9 @@ arrow_table_slice_builder::~arrow_table_slice_builder() noexcept {
 // -- properties ---------------------------------------------------------------
 
 size_t arrow_table_slice_builder::columns() const noexcept {
-  return flat_layout_.fields.size();
+  auto result = schema_->num_fields();
+  VAST_ASSERT(result >= 0);
+  return detail::narrow_cast<size_t>(result);
 }
 
 table_slice arrow_table_slice_builder::finish(
@@ -455,19 +524,28 @@ table_slice arrow_table_slice_builder::finish(
   // of fields in the layout.
   VAST_ASSERT(column_ == 0);
   // Pack layout.
-  auto layout_buffer
-    = serialized_layout.empty()
-        ? *fbs::serialize_bytes(builder_, layout())
-        : builder_.CreateVector(
-          reinterpret_cast<const unsigned char*>(serialized_layout.data()),
-          serialized_layout.size());
+  auto use_layout = [&](const auto& buf) {
+    return builder_.CreateVector(
+      reinterpret_cast<const unsigned char*>(buf.data()), buf.size());
+  };
+  auto gen_layout = [&]() {
+    caf::binary_serializer source(nullptr, serialized_layout_cache_);
+    auto error = source(layout());
+    VAST_ASSERT(error == caf::no_error);
+    return use_layout(serialized_layout_cache_);
+  };
+  auto layout_buffer = !serialized_layout.empty()
+                         ? use_layout(serialized_layout)
+                         : (!serialized_layout_cache_.empty()
+                              ? use_layout(serialized_layout_cache_)
+                              : gen_layout());
   // Pack schema.
-#if ARROW_VERSION_MAJOR >= 2
+#  if ARROW_VERSION_MAJOR >= 2
   auto flat_schema = arrow::ipc::SerializeSchema(*schema_).ValueOrDie();
-#else
+#  else
   auto flat_schema
     = arrow::ipc::SerializeSchema(*schema_, nullptr).ValueOrDie();
-#endif
+#  endif
   auto schema_buffer
     = builder_.CreateVector(flat_schema->data(), flat_schema->size());
   // Pack record batch.
@@ -498,6 +576,50 @@ table_slice arrow_table_slice_builder::finish(
   return table_slice{std::move(chunk), table_slice::verify::no, layout()};
 }
 
+table_slice arrow_table_slice_builder::create(
+  const std::shared_ptr<arrow::RecordBatch>& record_batch,
+  const record_type& layout, size_t initial_buffer_size) {
+  VAST_ASSERT(record_batch->schema()->Equals(make_arrow_schema(layout)),
+              "record layout doesn't match record batch schema");
+  auto builder = flatbuffers::FlatBufferBuilder{initial_buffer_size};
+  // Pack layout.
+  auto flat_layout = std::vector<char>{};
+  caf::binary_serializer source(nullptr, flat_layout);
+  auto error = source(layout);
+  VAST_ASSERT(error == caf::no_error);
+  auto layout_buffer = builder.CreateVector(
+    reinterpret_cast<const unsigned char*>(flat_layout.data()),
+    flat_layout.size());
+  // Pack schema.
+#  if ARROW_VERSION_MAJOR >= 2
+  auto flat_schema
+    = arrow::ipc::SerializeSchema(*record_batch->schema()).ValueOrDie();
+#  else
+  auto flat_schema
+    = arrow::ipc::SerializeSchema(*record_batch->schema(), nullptr).ValueOrDie();
+#  endif
+  auto schema_buffer
+    = builder.CreateVector(flat_schema->data(), flat_schema->size());
+  // Pack record batch.
+  auto flat_record_batch
+    = arrow::ipc::SerializeRecordBatch(*record_batch,
+                                       arrow::ipc::IpcWriteOptions::Defaults())
+        .ValueOrDie();
+  auto record_batch_buffer = builder.CreateVector(flat_record_batch->data(),
+                                                  flat_record_batch->size());
+  // Create Arrow-encoded table slices.
+  auto arrow_table_slice_buffer = fbs::table_slice::arrow::Createv0(
+    builder, layout_buffer, schema_buffer, record_batch_buffer);
+  // Create and finish table slice.
+  auto table_slice_buffer
+    = fbs::CreateTableSlice(builder, fbs::table_slice::TableSlice::arrow_v0,
+                            arrow_table_slice_buffer.Union());
+  fbs::FinishTableSliceBuffer(builder, table_slice_buffer);
+  // Create the table slice from the chunk.
+  auto chunk = fbs::release(builder);
+  return table_slice{std::move(chunk), table_slice::verify::no, layout};
+}
+
 size_t arrow_table_slice_builder::rows() const noexcept {
   return rows_;
 }
@@ -516,22 +638,21 @@ void arrow_table_slice_builder::reserve([[maybe_unused]] size_t num_rows) {
 arrow_table_slice_builder::arrow_table_slice_builder(record_type layout,
                                                      size_t initial_buffer_size)
   : table_slice_builder{std::move(layout)},
-    flat_layout_{flatten(this->layout())},
-    schema_{make_arrow_schema(this->flat_layout_)},
+    schema_{make_arrow_schema(this->layout())},
     builder_{initial_buffer_size} {
   VAST_ASSERT(schema_);
   VAST_ASSERT(schema_->num_fields()
-              == detail::narrow_cast<int>(flat_layout_.fields.size()));
-  column_builders_.reserve(flat_layout_.fields.size());
-  auto pool = arrow::default_memory_pool();
-  for (auto& field : flat_layout_.fields)
-    column_builders_.emplace_back(column_builder::make(field.type, pool));
+              == detail::narrow_cast<int>(this->layout().num_leaves()));
+  column_builders_.reserve(columns());
+  auto* pool = arrow::default_memory_pool();
+  for (const auto& field : record_type::each(this->layout()))
+    column_builders_.emplace_back(column_builder::make(field.type(), pool));
 }
 
 bool arrow_table_slice_builder::add_impl(data_view x) {
   if (!column_builders_[column_]->add(x))
     return false;
-  if (++column_ == flat_layout_.fields.size()) {
+  if (++column_ == columns()) {
     ++rows_;
     column_ = 0;
   }
@@ -543,8 +664,8 @@ bool arrow_table_slice_builder::add_impl(data_view x) {
 std::shared_ptr<arrow::Schema> make_arrow_schema(const record_type& t) {
   std::vector<std::shared_ptr<arrow::Field>> arrow_fields;
   arrow_fields.reserve(t.fields.size());
-  for (auto& field : t.fields) {
-    auto field_ptr = arrow::field(field.name, make_arrow_type(field.type));
+  for (const auto& field : record_type::each(t)) {
+    auto field_ptr = arrow::field(field.key(), make_arrow_type(field.type()));
     arrow_fields.emplace_back(std::move(field_ptr));
   }
   auto metadata = arrow::key_value_metadata({{"name", t.name()}});
