@@ -1,15 +1,10 @@
-/******************************************************************************
- *                    _   _____   __________                                  *
- *                   | | / / _ | / __/_  __/     Visibility                   *
- *                   | |/ / __ |_\ \  / /          Across                     *
- *                   |___/_/ |_/___/ /_/       Space and Time                 *
- *                                                                            *
- * This file is part of VAST. It is subject to the license terms in the       *
- * LICENSE file found in the top-level directory of this distribution and at  *
- * http://vast.io/license. No part of VAST, including this file, may be       *
- * copied, modified, propagated, or distributed except according to the terms *
- * contained in the LICENSE file.                                             *
- ******************************************************************************/
+//    _   _____   __________
+//   | | / / _ | / __/_  __/     Visibility
+//   | |/ / __ |_\ \  / /          Across
+//   |___/_/ |_/___/ /_/       Space and Time
+//
+// SPDX-FileCopyrightText: (c) 2019 The VAST Contributors
+// SPDX-License-Identifier: BSD-3-Clause
 
 #include "vast/format/zeek.hpp"
 
@@ -246,8 +241,8 @@ const char* reader::name() const {
   return "zeek-reader";
 }
 
-caf::error reader::read_impl(size_t max_events, size_t max_slice_size,
-                             consumer& f) {
+caf::error
+reader::read_impl(size_t max_events, size_t max_slice_size, consumer& f) {
   // Sanity checks.
   VAST_ASSERT(max_events > 0);
   VAST_ASSERT(max_slice_size > 0);
@@ -394,13 +389,8 @@ caf::error reader::parse_header() {
   }
   // Retrieve remaining header lines.
   const char* prefixes[] = {
-    "#set_separator",
-    "#empty_field",
-    "#unset_field",
-    "#path",
-    "#open",
-    "#fields",
-    "#types",
+    "#set_separator", "#empty_field", "#unset_field", "#path",
+    "#open",          "#fields",      "#types",
   };
   std::vector<std::string> header(sizeof(prefixes) / sizeof(const char*));
   for (auto i = 0u; i < header.size(); ++i) {
@@ -430,9 +420,8 @@ caf::error reader::parse_header() {
   if (fields.size() != types.size())
     return caf::make_error(ec::format_error, "fields and types have different "
                                              "size");
-  // TODO: Consider unflattening the incoming layout.
   std::vector<record_field> record_fields;
-  proto_field_ = caf::none;
+  proto_field_.reset();
   for (auto i = 0u; i < fields.size(); ++i) {
     auto t = parse_type(types[i]);
     if (!t)
@@ -463,18 +452,17 @@ caf::error reader::parse_header() {
       return caf::make_error(ec::format_error,
                              "the zeek reader expects records for "
                              "the top level types in the schema");
-    auto flat = flatten(*r);
-    for (auto& f : flat.fields) {
+    for (const auto& field : record_type::each(*r)) {
       auto i = std::find_if(layout_.fields.begin(), layout_.fields.end(),
-                            [&](auto& hf) { return hf.name == f.name; });
+                            [&](auto& hf) { return hf.name == field.key(); });
       if (i != layout_.fields.end()) {
-        if (!congruent(i->type, f.type))
+        if (!congruent(i->type, field.type()))
           VAST_WARN("{} encountered a type mismatch between the schema "
-                    "definition ({}) and the input data ({})",
-                    detail::pretty_type_name(this), f, *i);
-        else if (!f.type.attributes().empty()) {
-          i->type.attributes(f.type.attributes());
-        }
+                    "definition ({}: {}) and the input data ({})",
+                    detail::pretty_type_name(this), field.key(), field.type(),
+                    *i);
+        else if (!field.type().attributes().empty())
+          i->type.attributes(field.type().attributes());
       }
     }
   } // We still do attribute inference for the user provided layouts.
@@ -518,15 +506,11 @@ writer::writer(const caf::settings& options) {
   auto output
     = get_or(options, "vast.export.write", vast::defaults::export_::write);
   if (output != "-")
-    dir_ = std::move(output);
+    dir_ = std::filesystem::path{std::move(output)};
   show_timestamp_tags_
     = !caf::get_or(options, "vast.export.zeek.disable-timestamp-tags", false);
 }
 
-writer::~writer() {
-  // nop
-}
-
 namespace {
 
 template <class Iterator>
@@ -625,7 +609,7 @@ public:
       p, slice, {std::string_view{&separator, 1}, "", "", ""});
   }
 
-  const char* name() const override {
+  [[nodiscard]] const char* name() const override {
     return "zeek-writer";
   }
 
@@ -637,7 +621,7 @@ public:
 caf::error writer::write(const table_slice& slice) {
   ostream_writer* child = nullptr;
   auto&& layout = slice.layout();
-  if (dir_.empty()) {
+  if (dir_.string().empty()) {
     if (writers_.empty()) {
       VAST_DEBUG("{} creates a new stream for STDOUT",
                  detail::pretty_type_name(this));
@@ -657,15 +641,26 @@ caf::error writer::write(const table_slice& slice) {
     } else {
       VAST_DEBUG("{} creates new stream for layout {}",
                  detail::pretty_type_name(this), layout.name());
-      if (!exists(dir_)) {
-        if (auto err = mkdir(dir_))
-          return err;
-      } else if (!dir_.is_directory()) {
-        return caf::make_error(ec::format_error,
-                               "got existing non-directory path", dir_);
+      std::error_code err{};
+      const auto exists = std::filesystem::exists(dir_, err);
+      if (err)
+        return caf::make_error(ec::filesystem_error,
+                               fmt::format("failed to check if file {} exists: "
+                                           "{}",
+                                           dir_.string(), err.message()));
+      if (!exists) {
+        std::filesystem::create_directory(dir_, err);
+        if (err)
+          return caf::make_error(ec::filesystem_error,
+                                 fmt::format("failed to create directory {}: "
+                                             "{}",
+                                             dir_.string(), err.message()));
+      } else if (!std::filesystem::is_directory(dir_, err)) {
+        return caf::make_error(
+          ec::format_error, "got existing non-directory path", dir_.string());
       }
       auto filename = dir_ / (layout.name() + ".log");
-      auto fos = std::make_unique<std::ofstream>(filename.str());
+      auto fos = std::make_unique<std::ofstream>(filename.string());
       print_header(layout, *fos, show_timestamp_tags_);
       auto i = writers_.emplace(
         layout.name(),
