@@ -1,25 +1,18 @@
-/******************************************************************************
- *                    _   _____   __________                                  *
- *                   | | / / _ | / __/_  __/     Visibility                   *
- *                   | |/ / __ |_\ \  / /          Across                     *
- *                   |___/_/ |_/___/ /_/       Space and Time                 *
- *                                                                            *
- * This file is part of VAST. It is subject to the license terms in the       *
- * LICENSE file found in the top-level directory of this distribution and at  *
- * http://vast.io/license. No part of VAST, including this file, may be       *
- * copied, modified, propagated, or distributed except according to the terms *
- * contained in the LICENSE file.                                             *
- ******************************************************************************/
+//    _   _____   __________
+//   | | / / _ | / __/_  __/     Visibility
+//   | |/ / __ |_\ \  / /          Across
+//   |___/_/ |_/___/ /_/       Space and Time
+//
+// SPDX-FileCopyrightText: (c) 2018 The VAST Contributors
+// SPDX-License-Identifier: BSD-3-Clause
 
 #include "vast/segment_store.hpp"
 
 #include "vast/bitmap_algorithms.hpp"
 #include "vast/concept/printable/to_string.hpp"
 #include "vast/concept/printable/vast/error.hpp"
-#include "vast/concept/printable/vast/filesystem.hpp"
 #include "vast/concept/printable/vast/uuid.hpp"
 #include "vast/detail/overload.hpp"
-#include "vast/directory.hpp"
 #include "vast/error.hpp"
 #include "vast/fbs/segment.hpp"
 #include "vast/fbs/utils.hpp"
@@ -32,11 +25,59 @@
 #include <caf/dictionary.hpp>
 #include <caf/settings.hpp>
 
+#include <filesystem>
+#include <system_error>
+
 namespace vast {
 
+segment_store::lookup::lookup(const segment_store& store, ids xs,
+                              std::vector<uuid>&& candidates)
+  : store_{store}, xs_{std::move(xs)}, candidates_{std::move(candidates)} {
+  // nop
+}
+
+caf::expected<table_slice> segment_store::lookup::next() {
+  // Update the buffer if it has been consumed or the previous
+  // refresh return an error.
+  while (!buffer_ || it_ == buffer_->end()) {
+    buffer_ = handle_segment();
+    if (!buffer_)
+      // Either an error occurred, or the list of candidates is exhausted.
+      return buffer_.error();
+    it_ = buffer_->begin();
+  }
+  return *it_++;
+}
+
+caf::expected<std::vector<table_slice>>
+segment_store::lookup::handle_segment() {
+  if (first_ == candidates_.end())
+    return caf::no_error;
+  auto& cand = *first_++;
+  if (cand == store_.builder_.id()) {
+    VAST_DEBUG("{} looks into the active segment {}",
+               detail::pretty_type_name(this), cand);
+    return store_.builder_.lookup(xs_);
+  }
+  auto i = store_.cache_.find(cand);
+  if (i != store_.cache_.end()) {
+    VAST_DEBUG("{} got cache hit for segment {}",
+               detail::pretty_type_name(this), cand);
+    return i->second.lookup(xs_);
+  }
+  VAST_DEBUG("{} got cache miss for segment {}", detail::pretty_type_name(this),
+             cand);
+  auto s = store_.load_segment(cand);
+  if (!s)
+    return s.error();
+  store_.cache_.emplace(cand, *s);
+  return s->lookup(xs_);
+}
+
 // TODO: return expected<segment_store_ptr> for better error propagation.
-segment_store_ptr segment_store::make(path dir, size_t max_segment_size,
-                                      size_t in_memory_segments) {
+segment_store_ptr
+segment_store::make(std::filesystem::path dir, size_t max_segment_size,
+                    size_t in_memory_segments) {
   VAST_TRACE_SCOPE("{} {} {}", VAST_ARG(dir), VAST_ARG(max_segment_size),
                    VAST_ARG(in_memory_segments));
   VAST_ASSERT(max_segment_size > 0);
@@ -47,7 +88,8 @@ segment_store_ptr segment_store::make(path dir, size_t max_segment_size,
   return result;
 }
 
-segment_store::segment_store(path dir, uint64_t max_segment_size,
+segment_store::segment_store(std::filesystem::path dir,
+                             uint64_t max_segment_size,
                              size_t in_memory_segments)
   : dir_{std::move(dir)},
     max_segment_size_{max_segment_size},
@@ -58,10 +100,6 @@ segment_store::segment_store(path dir, uint64_t max_segment_size,
   // nop
 }
 
-segment_store::~segment_store() {
-  // nop
-}
-
 caf::error segment_store::put(table_slice xs) {
   VAST_TRACE_SCOPE("{}", VAST_ARG(xs));
   if (!segments_.inject(xs.offset(), xs.offset() + xs.rows(), builder_.id()))
@@ -75,62 +113,8 @@ caf::error segment_store::put(table_slice xs) {
   return flush();
 }
 
-std::unique_ptr<store::lookup> segment_store::extract(const ids& xs) const {
-  class lookup : public store::lookup {
-  public:
-    using uuid_iterator = std::vector<uuid>::iterator;
-
-    lookup(const segment_store& store, ids xs, std::vector<uuid>&& candidates)
-      : store_{store}, xs_{std::move(xs)}, candidates_{std::move(candidates)} {
-      // nop
-    }
-
-    caf::expected<table_slice> next() override {
-      // Update the buffer if it has been consumed or the previous
-      // refresh return an error.
-      while (!buffer_ || it_ == buffer_->end()) {
-        buffer_ = handle_segment();
-        if (!buffer_)
-          // Either an error occurred, or the list of candidates is exhausted.
-          return buffer_.error();
-        it_ = buffer_->begin();
-      }
-      return *it_++;
-    }
-
-  private:
-    caf::expected<std::vector<table_slice>> handle_segment() {
-      if (first_ == candidates_.end())
-        return caf::no_error;
-      auto& cand = *first_++;
-      if (cand == store_.builder_.id()) {
-        VAST_DEBUG("{} looks into the active segment {}",
-                   detail::pretty_type_name(this), cand);
-        return store_.builder_.lookup(xs_);
-      }
-      auto i = store_.cache_.find(cand);
-      if (i != store_.cache_.end()) {
-        VAST_DEBUG("{} got cache hit for segment {}",
-                   detail::pretty_type_name(this), cand);
-        return i->second.lookup(xs_);
-      }
-      VAST_DEBUG("{} got cache miss for segment {}",
-                 detail::pretty_type_name(this), cand);
-      auto s = store_.load_segment(cand);
-      if (!s)
-        return s.error();
-      store_.cache_.emplace(cand, *s);
-      return s->lookup(xs_);
-    }
-
-    const segment_store& store_;
-    ids xs_;
-    std::vector<uuid> candidates_;
-    uuid_iterator first_ = candidates_.begin();
-    caf::expected<std::vector<table_slice>> buffer_{caf::no_error};
-    std::vector<table_slice>::iterator it_;
-  };
-
+std::unique_ptr<segment_store::lookup>
+segment_store::extract(const ids& xs) const {
   VAST_TRACE_SCOPE("{}", VAST_ARG(xs));
   // Collect candidate segments by seeking through the ID set and
   // probing each ID interval.
@@ -259,7 +243,10 @@ caf::error segment_store::erase(const ids& xs) {
                    detail::pretty_type_name(this));
       auto stale_filename = segment_path() / to_string(segment_id);
       // Schedule deletion of the segment file when releasing the chunk.
-      seg.chunk()->add_deletion_step([=]() noexcept { rm(stale_filename); });
+      seg.chunk()->add_deletion_step([=]() noexcept {
+        std::error_code err{};
+        std::filesystem::remove(stale_filename, err);
+      });
     }
     // else: nothing to do, since we can continue filling the active segment.
   };
@@ -347,7 +334,7 @@ caf::error segment_store::flush() {
   // Keep new segment in the cache.
   cache_.emplace(seg.id(), seg);
   VAST_DEBUG("{} wrote new segment to {}", detail::pretty_type_name(this),
-             filename.trim(-3));
+             filename.parent_path());
   return caf::none;
 }
 
@@ -373,17 +360,26 @@ void segment_store::inspect_status(caf::settings& xs,
 }
 
 caf::error segment_store::register_segments() {
-  for (auto filename : directory{segment_path()})
-    if (auto err = register_segment(filename))
-      return err;
+  if (!std::filesystem::exists(segment_path()))
+    return caf::none;
+  std::error_code err{};
+  std::filesystem::directory_iterator dir{segment_path(), err};
+  if (err)
+    return caf::make_error(ec::filesystem_error,
+                           fmt::format("failed to find segment path {} : {}",
+                                       dir->path(), err.message()));
+  for (const auto& entry : dir)
+    if (entry.exists())
+      if (auto err = register_segment(entry.path()))
+        return err;
   return caf::none;
 }
 
-caf::error segment_store::register_segment(const path& filename) {
+caf::error
+segment_store::register_segment(const std::filesystem::path& filename) {
   auto chk = chunk::mmap(filename);
   if (!chk)
-    return caf::make_error(ec::filesystem_error, "failed to mmap chunk",
-                           filename);
+    return std::move(chk.error());
   // We don't verify the segment here, since doing that would access
   // most of the pages of the mapping and effectively cause us to
   // read of the whole archive contents from disk. When the database
@@ -391,7 +387,7 @@ caf::error segment_store::register_segment(const path& filename) {
   // (see also tdhtf/ch1935)
   // TODO: Create a library function that performs verification on a
   // subset of the fields of a flatbuffer table.
-  auto s = fbs::GetSegment(chk->data());
+  auto s = fbs::GetSegment(chk->get()->data());
   if (s == nullptr)
     return caf::make_error(ec::format_error, "segment integrity check failed");
   auto s0 = s->segment_as_v0();
@@ -415,9 +411,8 @@ caf::expected<segment> segment_store::load_segment(uuid id) const {
              filename);
   auto chk = chunk::mmap(filename);
   if (!chk)
-    return caf::make_error(ec::filesystem_error, "failed to mmap chunk",
-                           filename);
-  if (auto segment = segment::make(std::move(chk))) {
+    return std::move(chk.error());
+  if (auto segment = segment::make(std::move(*chk))) {
     return segment;
   } else {
     VAST_ERROR("{} failed to load segment at {} with error: {}",
@@ -460,7 +455,10 @@ uint64_t segment_store::drop(segment& x) {
             segment_id);
   // Schedule deletion of the segment file when releasing the chunk.
   auto filename = segment_path() / to_string(segment_id);
-  x.chunk()->add_deletion_step([=]() noexcept { rm(filename); });
+  x.chunk()->add_deletion_step([=]() noexcept {
+    std::error_code err{};
+    std::filesystem::remove(filename, err);
+  });
   segments_.erase_value(segment_id);
   return erased_events;
 }
