@@ -1,15 +1,10 @@
-/******************************************************************************
- *                    _   _____   __________                                  *
- *                   | | / / _ | / __/_  __/     Visibility                   *
- *                   | |/ / __ |_\ \  / /          Across                     *
- *                   |___/_/ |_/___/ /_/       Space and Time                 *
- *                                                                            *
- * This file is part of VAST. It is subject to the license terms in the       *
- * LICENSE file found in the top-level directory of this distribution and at  *
- * http://vast.io/license. No part of VAST, including this file, may be       *
- * copied, modified, propagated, or distributed except according to the terms *
- * contained in the LICENSE file.                                             *
- ******************************************************************************/
+//    _   _____   __________
+//   | | / / _ | / __/_  __/     Visibility
+//   | |/ / __ |_\ \  / /          Across
+//   |___/_/ |_/___/ /_/       Space and Time
+//
+// SPDX-FileCopyrightText: (c) 2018 The VAST Contributors
+// SPDX-License-Identifier: BSD-3-Clause
 
 #include "vast/system/application.hpp"
 
@@ -46,20 +41,10 @@
 #  include "vast/format/arrow.hpp"
 #endif
 
-#if VAST_ENABLE_PCAP
-#  include "vast/format/pcap.hpp"
-#  include "vast/system/pcap_writer_command.hpp"
-#endif
-
 namespace vast::system {
 
 namespace {
 
-auto make_pcap_options(std::string_view category) {
-  return sink_opts(category).add<size_t>(
-    "flush-interval,f", "flush to disk after this many packets");
-}
-
 command::opts_builder add_index_opts(command::opts_builder ob) {
   return std::move(ob)
     .add<size_t>("max-partition-size", "maximum number of events in a "
@@ -124,43 +109,53 @@ auto make_export_command() {
       .add<bool>("continuous,c", "marks a query as continuous")
       .add<bool>("unified,u", "marks a query as unified")
       .add<bool>("disable-taxonomies", "don't substitute taxonomy identifiers")
+      .add<std::string>("timeout", "timeout to stop the export after")
+      // We don't expose the `preserve-ids` option to the user because it
+      // doesnt' affect the formatted output.
+      //.add<bool>("preserve-ids", "don't substitute taxonomy identifiers")
       .add<size_t>("max-events,n", "maximum number of results")
       .add<std::string>("read,r", "path for reading the query")
       .add<std::string>("write,w", "path to write events to")
       .add<bool>("uds,d", "treat -w as UNIX domain socket to connect to"));
-  // NOTICE: The `sink_opts` are relocated in `command.cpp:fixup_options()`,
-  // That function must be kept in sync with the commands added here.
   export_->add_subcommand("zeek", "exports query results in Zeek format",
                           documentation::vast_export_zeek,
-                          sink_opts("?vast.export.zeek"));
+                          opts("?vast.export.zeek")
+                            .add<bool>("disable-timestamp-tags",
+                                       "whether the output should contain "
+                                       "#open/#close tags"));
   export_->add_subcommand("csv", "exports query results in CSV format",
                           documentation::vast_export_csv,
-                          sink_opts("?vast.export.csv"));
+                          opts("?vast.export.csv"));
   export_->add_subcommand("ascii", "exports query results in ASCII format",
                           documentation::vast_export_ascii,
-                          sink_opts("?vast.export.ascii"));
+                          opts("?vast.export.ascii"));
   export_->add_subcommand("json", "exports query results in JSON format",
                           documentation::vast_export_json,
-                          sink_opts("?vast.export.json")
+                          opts("?vast.export.json")
                             .add<bool>("flatten", "flatten nested objects into "
-                                                  "the top-level"));
+                                                  "the top-level")
+                            .add<bool>("numeric-durations",
+                                       "render durations as numbers as opposed "
+                                       "to human-readable strings"));
   export_->add_subcommand("null",
                           "exports query without printing them (debug option)",
                           documentation::vast_export_null,
-                          sink_opts("?vast.export.null"));
+                          opts("?vast.export.null"));
 #if VAST_ENABLE_ARROW
-  // The Arrow export does not support --write or --uds, so we don't use the
-  // sink_opts here intentionally.
   export_->add_subcommand("arrow", "exports query results in Arrow format",
                           documentation::vast_export_arrow,
                           opts("?vast.export.arrow"));
 
 #endif
-#if VAST_ENABLE_PCAP
-  export_->add_subcommand("pcap", "exports query results in PCAP format",
-                          documentation::vast_export_pcap,
-                          make_pcap_options("?vast.export.pcap"));
-#endif
+  for (const auto& plugin : plugins::get()) {
+    if (const auto* writer = plugin.as<writer_plugin>()) {
+      auto opts_category
+        = fmt::format("?vast.export.{}", writer->writer_format());
+      export_->add_subcommand(writer->writer_format(), writer->writer_help(),
+                              writer->writer_documentation(),
+                              writer->writer_options(opts(opts_category)));
+    }
+  }
   return export_;
 }
 
@@ -198,49 +193,38 @@ auto make_import_command() {
       .add<std::string>("schema-file,s", "path to alternate schema")
       .add<std::string>("type,t", "filter event type based on prefix matching")
       .add<bool>("uds,d", "treat -r as listening UNIX domain socket"));
-  // NOTICE: The `source_opts` are relocated in `command.cpp:fixup_options()`,
-  // That function must be kept in sync with the commands added here.
   import_->add_subcommand("zeek", "imports Zeek TSV logs from STDIN or file",
                           documentation::vast_import_zeek,
-                          source_opts("?vast.import.zeek"));
+                          opts("?vast.import.zeek"));
   import_->add_subcommand("zeek-json",
                           "imports Zeek JSON logs from STDIN or file",
                           documentation::vast_import_zeek,
-                          source_opts("?vast.import.zeek-json"));
+                          opts("?vast.import.zeek-json"));
   import_->add_subcommand("csv", "imports CSV logs from STDIN or file",
                           documentation::vast_import_csv,
-                          source_opts("?vast.import.csv"));
+                          opts("?vast.import.csv"));
   import_->add_subcommand("json", "imports JSON with schema",
                           documentation::vast_import_json,
-                          source_opts("?vast.import.json"));
+                          opts("?vast.import.json"));
   import_->add_subcommand("suricata", "imports suricata eve json",
                           documentation::vast_import_suricata,
-                          source_opts("?vast.import.suricata"));
+                          opts("?vast.import.suricata"));
   import_->add_subcommand("syslog", "imports syslog messages",
                           documentation::vast_import_syslog,
-                          source_opts("?vast.import.syslog"));
+                          opts("?vast.import.syslog"));
   import_->add_subcommand(
     "test", "imports random data for testing or benchmarking",
     documentation::vast_import_test,
-    source_opts("?vast.import.test").add<size_t>("seed", "the PRNG seed"));
-#if VAST_ENABLE_PCAP
-  import_->add_subcommand(
-    "pcap", "imports PCAP logs from STDIN or file",
-    documentation::vast_import_pcap,
-    source_opts("?vast.import.pcap")
-      .add<std::string>("interface,i", "network interface to read packets from")
-      .add<size_t>("cutoff,c", "skip flow packets after this many bytes")
-      .add<size_t>("max-flows,m", "number of concurrent flows to track")
-      .add<size_t>("max-flow-age,a", "max flow lifetime before eviction")
-      .add<size_t>("flow-expiry,e", "flow table expiration interval")
-      .add<size_t>("pseudo-realtime-factor,p", "factor c delaying packets by "
-                                               "1/c")
-      .add<size_t>("snaplen", "snapshot length in bytes")
-      .add<double>("drop-rate-threshold", "drop rate that must be exceeded for "
-                                          "warnings to occur")
-      .add<bool>("disable-community-id", "disable computation of community id "
-                                         "for every packet"));
-#endif
+    opts("?vast.import.test").add<size_t>("seed", "the PRNG seed"));
+  for (const auto& plugin : plugins::get()) {
+    if (const auto* reader = plugin.as<reader_plugin>()) {
+      auto opts_category
+        = fmt::format("?vast.import.{}", reader->reader_format());
+      import_->add_subcommand(reader->reader_format(), reader->reader_help(),
+                              reader->reader_documentation(),
+                              reader->reader_options(opts(opts_category)));
+    }
+  }
   return import_;
 }
 
@@ -258,7 +242,9 @@ auto make_pivot_command() {
   auto pivot = std::make_unique<command>(
     "pivot", "extracts related events of a given type",
     documentation::vast_pivot,
-    make_pcap_options("?vast.pivot")
+    opts("?vast.pivot")
+      .add<size_t>("flush-interval,f", "flush to disk after this many packets "
+                                       "(only with the PCAP plugin)")
       .add<bool>("disable-taxonomies", "don't substitute taxonomy identifiers")
       .add<std::string>("format", "output format "
                                   "(default: JSON)"));
@@ -293,45 +279,37 @@ auto make_spawn_source_command() {
   spawn_source->add_subcommand("csv",
                                "creates a new CSV source inside the node",
                                documentation::vast_spawn_source_csv,
-                               source_opts("?vast.spawn.source.csv"));
+                               opts("?vast.spawn.source.csv"));
   spawn_source->add_subcommand("json",
                                "creates a new JSON source inside the node",
                                documentation::vast_spawn_source_json,
-                               source_opts("?vast.spawn.source.json"));
-#if VAST_ENABLE_PCAP
-  spawn_source->add_subcommand(
-    "pcap", "creates a new PCAP source inside the node",
-    documentation::vast_spawn_source_pcap,
-    source_opts("?vast.spawn.source.pcap")
-      .add<std::string>("interface,i", "network interface to read packets from")
-      .add<size_t>("cutoff,c", "skip flow packets after this many bytes")
-      .add<size_t>("max-flows,m", "number of concurrent flows to track")
-      .add<size_t>("max-flow-age,a", "max flow lifetime before eviction")
-      .add<size_t>("flow-expiry,e", "flow table expiration interval")
-      .add<size_t>("pseudo-realtime-factor,p", "factor c delaying packets by "
-                                               "1/c")
-      .add<size_t>("snaplen", "snapshot length in bytes")
-      .add<double>("drop-rate-threshold", "drop rate that must be exceeded for "
-                                          "warnings to occur")
-      .add<bool>("disable-community-id", "disable computation of community id "
-                                         "for every packet"));
-#endif
+                               opts("?vast.spawn.source.json"));
   spawn_source->add_subcommand("suricata",
                                "creates a new Suricata source inside the node",
                                documentation::vast_spawn_source_suricata,
-                               source_opts("?vast.spawn.source.suricata"));
+                               opts("?vast.spawn.source.suricata"));
   spawn_source->add_subcommand("syslog",
                                "creates a new Syslog source inside the node",
                                documentation::vast_spawn_source_syslog,
-                               source_opts("?vast.spawn.source.syslog"));
+                               opts("?vast.spawn.source.syslog"));
   spawn_source->add_subcommand(
     "test", "creates a new test source inside the node",
     documentation::vast_spawn_source_test,
-    source_opts("?vast.spawn.source.test").add<size_t>("seed", "the PRNG seed"));
+    opts("?vast.spawn.source.test").add<size_t>("seed", "the PRNG seed"));
   spawn_source->add_subcommand("zeek",
                                "creates a new Zeek source inside the node",
                                documentation::vast_spawn_source_zeek,
-                               source_opts("?vast.spawn.source.zeek"));
+                               opts("?vast.spawn.source.zeek"));
+  for (const auto& plugin : plugins::get()) {
+    if (const auto* reader = plugin.as<reader_plugin>()) {
+      auto opts_category
+        = fmt::format("?vast.spawn.source.{}", reader->reader_format());
+      spawn_source->add_subcommand(reader->reader_format(),
+                                   reader->reader_help(),
+                                   reader->reader_documentation(),
+                                   reader->reader_options(opts(opts_category)));
+    }
+  }
   return spawn_source;
 }
 
@@ -342,10 +320,6 @@ auto make_spawn_sink_command() {
       .add<std::string>("write,w", "path to write events to")
       .add<bool>("uds,d", "treat -w as UNIX domain socket"),
     false);
-  spawn_sink->add_subcommand("pcap", "creates a new PCAP sink", "",
-                             opts("?vast.spawn.sink.pcap")
-                               .add<size_t>("flush,f", "flush to disk after "
-                                                       "this many packets"));
   spawn_sink->add_subcommand("zeek", "creates a new Zeek sink", "",
                              opts("?vast.spawn.sink.zeek"));
   spawn_sink->add_subcommand("ascii", "creates a new ASCII sink", "",
@@ -354,6 +328,15 @@ auto make_spawn_sink_command() {
                              opts("?vast.spawn.sink.csv"));
   spawn_sink->add_subcommand("json", "creates a new JSON sink", "",
                              opts("?vast.spawn.sink.json"));
+  for (const auto& plugin : plugins::get()) {
+    if (const auto* writer = plugin.as<writer_plugin>()) {
+      auto opts_category
+        = fmt::format("?vast.spawn.sink.{}", writer->writer_format());
+      spawn_sink->add_subcommand(writer->writer_format(), writer->writer_help(),
+                                 writer->writer_documentation(),
+                                 writer->writer_options(opts(opts_category)));
+    }
+  }
   return spawn_sink;
 }
 
@@ -403,8 +386,12 @@ auto make_start_command() {
       .add<bool>("print-endpoint", "print the client endpoint on stdout")
       .add<size_t>("disk-budget-check-interval", "time between two disk size "
                                                  "scans")
+      .add<std::string>("disk-budget-check-binary",
+                        "binary to run to determine current disk usage")
       .add<std::string>("disk-budget-high", "high-water mark for disk budget")
-      .add<std::string>("disk-budget-low", "low-water mark for disk budget"));
+      .add<std::string>("disk-budget-low", "low-water mark for disk budget")
+      .add<size_t>("disk-budget-step-size", "number of partitions to erase "
+                                            "before re-checking size"));
 }
 
 auto make_stop_command() {
@@ -422,7 +409,7 @@ auto make_command_factory() {
   // When updating this list, remember to update its counterpart in node.cpp as
   // well iff necessary
   // clang-format off
-  return command::factory{
+  auto result = command::factory{
     {"count", count_command},
     {"dump", remote_command},
     {"dump concepts", remote_command},
@@ -434,26 +421,17 @@ auto make_command_factory() {
     {"export null", make_writer_command("null")},
 #if VAST_ENABLE_ARROW
     {"export arrow", make_writer_command("arrow")},
-#endif
-#if VAST_ENABLE_PCAP
-    {"export pcap", pcap_writer_command},
 #endif
     {"export zeek", make_writer_command("zeek")},
     {"get", get_command},
     {"infer", infer_command},
-    {"import csv", import_command<format::csv::reader>},
-    {"import json", import_command<
-      format::json::reader<format::json::default_selector>>},
-#if VAST_ENABLE_PCAP
-    {"import pcap", import_command<format::pcap::reader>},
-#endif
-    {"import suricata", import_command<
-      format::json::reader<format::json::suricata_selector>>},
-    {"import syslog", import_command<format::syslog::reader>},
-    {"import test", import_command<format::test::reader>},
-    {"import zeek", import_command<format::zeek::reader>},
-    {"import zeek-json", import_command<
-      format::json::reader<format::json::zeek_selector>>},
+    {"import csv", import_command},
+    {"import json", import_command},
+    {"import suricata", import_command},
+    {"import syslog", import_command},
+    {"import test", import_command},
+    {"import zeek", import_command},
+    {"import zeek-json", import_command},
     {"kill", remote_command},
     {"peer", remote_command},
     {"pivot", pivot_command},
@@ -469,11 +447,9 @@ auto make_command_factory() {
     {"spawn sink ascii", remote_command},
     {"spawn sink csv", remote_command},
     {"spawn sink json", remote_command},
-    {"spawn sink pcap", remote_command},
     {"spawn sink zeek", remote_command},
     {"spawn source csv", remote_command},
     {"spawn source json", remote_command},
-    {"spawn source pcap", remote_command},
     {"spawn source suricata", remote_command},
     {"spawn source syslog", remote_command},
     {"spawn source test", remote_command},
@@ -485,6 +461,21 @@ auto make_command_factory() {
     {"version", version_command},
   };
   // clang-format on
+  for (auto& plugin : plugins::get()) {
+    if (auto* reader = plugin.as<reader_plugin>()) {
+      result.emplace(fmt::format("import {}", reader->reader_format()),
+                     import_command);
+      result.emplace(fmt::format("spawn source {}", reader->reader_format()),
+                     remote_command);
+    }
+    if (auto* writer = plugin.as<writer_plugin>()) {
+      result.emplace(fmt::format("export {}", writer->writer_format()),
+                     make_writer_command(writer->writer_format()));
+      result.emplace(fmt::format("spawn sink {}", writer->writer_format()),
+                     remote_command);
+    }
+  }
+  return result;
 } // namespace
 
 auto make_root_command(std::string_view path) {
@@ -493,18 +484,22 @@ auto make_root_command(std::string_view path) {
   // interested in "vast".
   path.remove_prefix(std::min(path.find_last_of('/') + 1, path.size()));
   // For documentation, we use the complete man-page formatted as Markdown
-  auto binary = detail::objectpath();
+  const auto binary = detail::objectpath();
   auto schema_desc
     = "list of directories to look for schema files ([/etc/vast/schema"s;
   if (binary) {
-    auto relative_schema_dir
-      = binary->parent().parent() / "share" / "vast" / "schema";
-    schema_desc += ", " + relative_schema_dir.str();
+    const auto relative_schema_dir
+      = binary->parent_path().parent_path() / "share" / "vast" / "schema";
+    schema_desc += ", " + relative_schema_dir.string();
   }
   schema_desc += "])";
   auto ob
     = opts("?vast")
         .add<std::string>("config", "path to a configuration file")
+        .add<bool>("bare-mode",
+                   "disable user and system configuration, schema and plugin "
+                   "directories lookup and static and dynamic plugin "
+                   "autoloading (this may only be used on the command line)")
         .add<caf::atom_value>("verbosity", "output verbosity level on the "
                                            "console")
         .add<std::vector<std::string>>("schema-dirs", schema_desc.c_str())
@@ -516,8 +511,6 @@ auto make_root_command(std::string_view path) {
         .add<std::string>("node-id,i", "the unique ID of this node")
         .add<bool>("node,N", "spawn a node instead of connecting to one")
         .add<bool>("enable-metrics", "keep track of performance metrics")
-        .add<bool>("no-default-schema", "don't load the default schema "
-                                        "definitions")
         .add<std::vector<std::string>>("plugin-dirs", "additional directories "
                                                       "to load plugins from")
         .add<std::vector<std::string>>("plugins", "plugins to load at startup")
@@ -558,8 +551,10 @@ make_application(std::string_view path) {
   auto root_factory = make_command_factory();
   // Add additional commands from plugins.
   for (auto& plugin : plugins::get()) {
-    if (auto cp = plugin.as<command_plugin>()) {
+    if (auto* cp = plugin.as<command_plugin>()) {
       auto&& [cmd, cmd_factory] = cp->make_command();
+      if (!cmd || cmd_factory.empty())
+        continue;
       root->add_subcommand(std::move(cmd));
       root_factory.insert(std::make_move_iterator(cmd_factory.begin()),
                           std::make_move_iterator(cmd_factory.end()));
@@ -596,34 +591,6 @@ void render_error(const command& root, const caf::error& err,
   }
 }
 
-command::opts_builder source_opts(std::string_view category) {
-  // The deprecated options are still accepted but get copied over to their
-  // new location in `command.cpp:fixup_options()`.
-  return command::opts(category)
-    .add<std::string>("listen,l", "deprecated - this option now applies to the "
-                                  "import command")
-    .add<std::string>("read,r", "deprecated - this option now applies to the "
-                                "import command")
-    .add<std::string>("schema-file,s", "deprecated - this option now applies "
-                                       "to the import command")
-    .add<std::string>("schema,S", "deprecated - this option now applies to the "
-                                  "import command")
-    .add<std::string>("type,t", "deprecated - this option now applies to the "
-                                "import command")
-    .add<bool>("uds,d", "deprecated - this option now applies to the import "
-                        "command");
-}
-
-command::opts_builder sink_opts(std::string_view category) {
-  // The deprecated options are still accepted but get copied over to their
-  // new location in `command.cpp:fixup_options()`.
-  return command::opts(category)
-    .add<std::string>("write,w", "deprecated - this option now applies to the "
-                                 "import command")
-    .add<bool>("uds,d", "deprecated - this option now applies to the import "
-                        "command");
-}
-
 command::opts_builder opts(std::string_view category) {
   return command::opts(category);
 }
