@@ -1,15 +1,10 @@
-/******************************************************************************
- *                    _   _____   __________                                  *
- *                   | | / / _ | / __/_  __/     Visibility                   *
- *                   | |/ / __ |_\ \  / /          Across                     *
- *                   |___/_/ |_/___/ /_/       Space and Time                 *
- *                                                                            *
- * This file is part of VAST. It is subject to the license terms in the       *
- * LICENSE file found in the top-level directory of this distribution and at  *
- * http://vast.io/license. No part of VAST, including this file, may be       *
- * copied, modified, propagated, or distributed except according to the terms *
- * contained in the LICENSE file.                                             *
- ******************************************************************************/
+//    _   _____   __________
+//   | | / / _ | / __/_  __/     Visibility
+//   | |/ / __ |_\ \  / /          Across
+//   |___/_/ |_/___/ /_/       Space and Time
+//
+// SPDX-FileCopyrightText: (c) 2016 The VAST Contributors
+// SPDX-License-Identifier: BSD-3-Clause
 
 #include "vast/system/index.hpp"
 
@@ -41,9 +36,9 @@
 #include "vast/io/read.hpp"
 #include "vast/io/save.hpp"
 #include "vast/logger.hpp"
-#include "vast/meta_index.hpp"
 #include "vast/partition_synopsis.hpp"
 #include "vast/system/evaluator.hpp"
+#include "vast/system/meta_index.hpp"
 #include "vast/system/partition.hpp"
 #include "vast/system/query_supervisor.hpp"
 #include "vast/system/shutdown.hpp"
@@ -53,13 +48,15 @@
 #include "vast/value_index.hpp"
 
 #include <caf/error.hpp>
+#include <caf/response_promise.hpp>
 #include <caf/typed_event_based_actor.hpp>
-
+#include <caf/typed_response_promise.hpp>
 #include <flatbuffers/flatbuffers.h>
 
 #include <algorithm>
 #include <chrono>
 #include <ctime>
+#include <filesystem>
 #include <memory>
 #include <unistd.h>
 
@@ -110,20 +107,19 @@ namespace vast::system {
 
 namespace {
 
-caf::error
-extract_partition_synopsis(const vast::path& partition_path,
-                           const vast::path& partition_synopsis_path) {
+caf::error extract_partition_synopsis(
+  const std::filesystem::path& partition_path,
+  const std::filesystem::path& partition_synopsis_path) {
   // Use blocking operations here since this is part of the startup.
   auto chunk = chunk::mmap(partition_path);
   if (!chunk)
-    return caf::make_error(ec::system_error, "could not mmap partition at "
-                                               + partition_path.str());
-  auto partition = fbs::GetPartition(chunk->data());
+    return std::move(chunk.error());
+  const auto* partition = fbs::GetPartition(chunk->get()->data());
   if (partition->partition_type() != fbs::partition::Partition::v0)
     return caf::make_error(ec::format_error, "found unsupported version for "
                                              "partition "
-                                               + partition_path.str());
-  auto partition_v0 = partition->partition_as_v0();
+                                               + partition_path.string());
+  const auto* partition_v0 = partition->partition_as_v0();
   VAST_ASSERT(partition_v0);
   partition_synopsis ps;
   unpack(*partition_v0, ps);
@@ -142,11 +138,12 @@ extract_partition_synopsis(const vast::path& partition_path,
 
 } // namespace
 
-vast::path index_state::partition_path(const uuid& id) const {
+std::filesystem::path index_state::partition_path(const uuid& id) const {
   return dir / to_string(id);
 }
 
-vast::path index_state::partition_synopsis_path(const uuid& id) const {
+std::filesystem::path
+index_state::partition_synopsis_path(const uuid& id) const {
   return synopsisdir / (to_string(id) + ".mdx");
 }
 
@@ -155,9 +152,10 @@ partition_actor partition_factory::operator()(const uuid& id) const {
   VAST_ASSERT(std::find(state_.persisted_partitions.begin(),
                         state_.persisted_partitions.end(), id)
               != state_.persisted_partitions.end());
-  auto path = state_.partition_path(id);
+  const auto path = state_.partition_path(id);
   VAST_DEBUG("{} loads partition {} for path {}", state_.self, id, path);
-  return state_.self->spawn(passive_partition, id, filesystem_, path);
+  return state_.self->spawn(passive_partition, id, filesystem_, path,
+                            state_.store);
 }
 
 filesystem_actor& partition_factory::filesystem() {
@@ -175,11 +173,13 @@ index_state::index_state(index_actor::pointer self)
 caf::error index_state::load_from_disk() {
   // We dont use the filesystem actor here because this function is only
   // called once during startup, when no other actors exist yet.
-  if (!exists(dir)) {
+  std::error_code err{};
+  const auto file_exists = std::filesystem::exists(dir, err);
+  if (!file_exists) {
     VAST_VERBOSE("{} found no prior state, starting with a clean slate", self);
     return caf::none;
   }
-  if (auto fname = index_filename(); exists(fname)) {
+  if (auto fname = index_filename(); std::filesystem::exists(fname, err)) {
     VAST_VERBOSE("{} loads state from {}", self, fname);
     auto buffer = io::read(fname);
     if (!buffer) {
@@ -189,19 +189,20 @@ caf::error index_state::load_from_disk() {
     }
     // TODO: Create a `index_ondisk_state` struct and move this part of the
     // code into an `unpack()` function.
-    auto index = fbs::GetIndex(buffer->data());
+    const auto* index = fbs::GetIndex(buffer->data());
     if (index->index_type() != fbs::index::Index::v0)
       return caf::make_error(ec::format_error, "invalid index version");
-    auto index_v0 = index->index_as_v0();
-    auto partition_uuids = index_v0->partitions();
+    const auto* index_v0 = index->index_as_v0();
+    const auto* partition_uuids = index_v0->partitions();
     VAST_ASSERT(partition_uuids);
-    for (auto uuid_fb : *partition_uuids) {
+    auto synopses = std::make_shared<std::map<uuid, partition_synopsis>>();
+    for (const auto* uuid_fb : *partition_uuids) {
       VAST_ASSERT(uuid_fb);
-      vast::uuid partition_uuid;
+      vast::uuid partition_uuid{};
       unpack(*uuid_fb, partition_uuid);
-      auto partition_path = dir / to_string(partition_uuid);
-      auto partition_synopsis_path = dir / (to_string(partition_uuid) + ".mdx");
-      if (!exists(partition_path)) {
+      auto part_dir = partition_path(partition_uuid);
+      auto synopsis_dir = partition_synopsis_path(partition_uuid);
+      if (!exists(part_dir)) {
         VAST_WARN("{} found partition {}"
                   "in the index state but not on disk; this may have been "
                   "caused by an unclean shutdown",
@@ -209,17 +210,17 @@ caf::error index_state::load_from_disk() {
         continue;
       }
       // Generate external partition synopsis file if it doesn't exist.
-      if (!exists(partition_synopsis_path)) {
-        if (auto error = extract_partition_synopsis(partition_path,
-                                                    partition_synopsis_path))
+      if (!exists(synopsis_dir)) {
+        if (auto error = extract_partition_synopsis(part_dir, synopsis_dir))
           return error;
       }
-      auto chunk = chunk::mmap(partition_synopsis_path);
+      auto chunk = chunk::mmap(synopsis_dir);
       if (!chunk) {
-        VAST_WARN("{} could not mmap partition at {}", self, partition_path);
+        VAST_WARN("{} could not mmap partition at {}", self, part_dir);
         continue;
       }
-      auto ps_flatbuffer = fbs::GetPartitionSynopsis(chunk->data());
+      const auto* ps_flatbuffer
+        = fbs::GetPartitionSynopsis(chunk->get()->data());
       partition_synopsis ps;
       if (ps_flatbuffer->partition_synopsis_type()
           != fbs::partition_synopsis::PartitionSynopsis::v0)
@@ -227,19 +228,43 @@ caf::error index_state::load_from_disk() {
                                                  "version");
       if (auto error = unpack(*ps_flatbuffer->partition_synopsis_as_v0(), ps))
         return error;
-      meta_idx.merge(partition_uuid, std::move(ps));
+      meta_index_bytes += ps.memusage();
       persisted_partitions.insert(partition_uuid);
+      synopses->emplace(partition_uuid, std::move(ps));
     }
-    auto stats = index_v0->stats();
+    // We collect all synopses to send them in bulk, since the `await` interface
+    // doesn't lend itself to a huge number of awaited messages: Only the tip of
+    // the current awaited list is considered, leading to an O(n**2) worst-case
+    // behavior if the responses arrive in the same order to how they were sent.
+    VAST_DEBUG("{} requesting bulk merge of {} partitions", self,
+               synopses->size());
+    this->accept_queries = false;
+    self
+      ->request(meta_index, caf::infinite, atom::merge_v,
+                std::exchange(synopses, {}))
+      .then(
+        [=](atom::ok) {
+          VAST_VERBOSE("{} successfully loaded meta index from disk and will "
+                       "start processing queries",
+                       self);
+          this->accept_queries = true;
+        },
+        [=](caf::error& err) {
+          VAST_ERROR("{} could not load meta index state from disk, shutting "
+                     "down with error {}",
+                     self, err);
+          self->send_exit(self, std::move(err));
+        });
+    const auto* stats = index_v0->stats();
     if (!stats)
       return caf::make_error(ec::format_error, "no stats in persisted index "
                                                "state");
-    for (const auto stat : *stats) {
+    for (const auto* const stat : *stats) {
       this->stats.layouts[stat->name()->str()]
         = layout_statistics{stat->count()};
     }
   } else {
-    VAST_WARN("{} found existing database dir {} without index "
+    VAST_INFO("{} found existing database dir {} without index "
               "statefile, "
               "will start with fresh state",
               self, dir);
@@ -247,7 +272,7 @@ caf::error index_state::load_from_disk() {
   return caf::none;
 }
 
-bool index_state::worker_available() {
+bool index_state::worker_available() const {
   return !idle_workers.empty();
 }
 
@@ -269,11 +294,24 @@ void index_state::add_flush_listener(flush_listener_actor listener) {
   detail::notify_listeners_if_clean(*this, *stage);
 }
 
+// The whole purpose of the `-b` flag is to somehow block until all imported
+// data is available for querying, so we have to layer hack upon hack here to
+// achieve this most of the time. This is only used for integration tests.
+// Note that there's still a race condition here if the call to
+// `notify_flush_listeners()` arrives when there's still data en route to
+// the unpersisted partitions.
+// TODO(ch19583): Rip out the whole 'notifying_stream_manager' and replace it
+// with some kind of ping/pong protocol.
 void index_state::notify_flush_listeners() {
   VAST_DEBUG("{} sends 'flush' messages to {} listeners", self,
              flush_listeners.size());
-  for (auto& listener : flush_listeners)
-    self->send(listener, atom::flush_v);
+  for (auto& listener : flush_listeners) {
+    if (active_partition.actor)
+      self->send(active_partition.actor, atom::subscribe_v, atom::flush_v,
+                 listener);
+    else
+      self->send(listener, atom::flush_v);
+  }
   flush_listeners.clear();
 }
 
@@ -289,7 +327,7 @@ void index_state::create_active_partition() {
   put(synopsis_options, "string-synopsis-fp-rate", meta_index_fp_rate);
   active_partition.actor
     = self->spawn(::vast::system::active_partition, id, filesystem, index_opts,
-                  synopsis_options);
+                  synopsis_options, store);
   active_partition.stream_slot
     = stage->add_outbound_path(active_partition.actor);
   active_partition.capacity = partition_capacity;
@@ -301,31 +339,43 @@ void index_state::decomission_active_partition() {
   auto id = active_partition.id;
   auto actor = std::exchange(active_partition.actor, {});
   unpersisted[id] = actor;
-  // Send buffered batches.
+  // Send buffered batches and remove active partition from the stream.
   stage->out().fan_out_flush();
-  stage->out().force_emit_batches();
-  // Remove active partition from the stream.
   stage->out().close(active_partition.stream_slot);
+  stage->out().force_emit_batches();
   // Persist active partition asynchronously.
-  auto part_dir = dir / to_string(id);
-  auto synopsis_dir = synopsisdir / (to_string(id) + ".mdx");
+  auto part_dir = partition_path(id);
+  auto synopsis_dir = partition_synopsis_path(id);
   VAST_DEBUG("{} persists active partition to {}", self, part_dir);
-  self->request(actor, caf::infinite, atom::persist_v, part_dir, synopsisdir)
+  self->request(actor, caf::infinite, atom::persist_v, part_dir, synopsis_dir)
     .then(
       [=](std::shared_ptr<partition_synopsis>& ps) {
         VAST_DEBUG("{} successfully persisted partition {}", self, id);
         // Semantically ps is a unique_ptr, and the partition releases its
         // copy before sending. We use shared_ptr for the transport because
         // CAF message types must be copy-constructible.
-        VAST_ASSERT(ps.use_count() == 1);
-        meta_idx.merge(id, std::move(*ps));
-        unpersisted.erase(id);
-        persisted_partitions.insert(id);
+        meta_index_bytes += ps->memusage();
+        // TODO: We should skip this continuation if we're currently shutting
+        // down.
+        self
+          ->request(meta_index, caf::infinite, atom::merge_v, id, std::move(ps))
+          .then(
+            [=](atom::ok) {
+              VAST_DEBUG("{} received ok for request to persist partition {}",
+                         self, id);
+              unpersisted.erase(id);
+              persisted_partitions.insert(id);
+            },
+            [=](const caf::error& err) {
+              VAST_DEBUG("{} received error for request to persist partition "
+                         "{}: {}",
+                         self, id, err);
+            });
       },
-      [=](const caf::error& err) {
+      [=](caf::error& err) {
         VAST_ERROR("{} failed to persist partition {} with error: {}", self, id,
-                   render(err));
-        self->quit(err);
+                   err);
+        self->quit(std::move(err));
       });
 }
 
@@ -355,7 +405,7 @@ index_state::status(status_verbosity v) const {
   if (v >= status_verbosity::detailed) {
     auto& stats_object = put_dictionary(index_status, "statistics");
     auto& layout_object = put_dictionary(stats_object, "layouts");
-    for (auto& [name, layout_stats] : stats.layouts) {
+    for (const auto& [name, layout_stats] : stats.layouts) {
       auto xs = caf::dictionary<caf::config_value>{};
       xs.emplace("count", layout_stats.count);
       // We cannot use put_dictionary(layout_object, name) here, because this
@@ -363,7 +413,7 @@ index_state::status(status_verbosity v) const {
       // Hence the fallback to low-level primitives.
       layout_object.insert_or_assign(name, std::move(xs));
     }
-    put(index_status, "meta-index-bytes", meta_idx.memusage());
+    put(index_status, "meta-index-bytes", meta_index_bytes);
     put(index_status, "num-active-partitions",
         active_partition.actor == nullptr ? 0 : 1);
     put(index_status, "num-cached-partitions", inmem_partitions.size());
@@ -406,11 +456,11 @@ index_state::status(status_verbosity v) const {
       partition_status(active_partition.id, active_partition.actor, active);
     auto& cached = put_list(partitions, "cached");
     cached.reserve(inmem_partitions.size());
-    for (auto& [id, actor] : inmem_partitions)
+    for (const auto& [id, actor] : inmem_partitions)
       partition_status(id, actor, cached);
     auto& unpersisted = put_list(partitions, "unpersisted");
     unpersisted.reserve(this->unpersisted.size());
-    for (auto& [id, actor] : this->unpersisted)
+    for (const auto& [id, actor] : this->unpersisted)
       partition_status(id, actor, unpersisted);
     // General state such as open streams.
     detail::fill_status_map(index_status, self);
@@ -431,7 +481,7 @@ index_state::collect_query_actors(query_state& lookup,
   auto partition_is_loaded = [&](const uuid& candidate) {
     return (active_partition.actor != nullptr
             && active_partition.id == candidate)
-           || unpersisted.count(candidate)
+           || (unpersisted.count(candidate) != 0u)
            || inmem_partitions.contains(candidate);
   };
   std::partition(lookup.partitions.begin(), lookup.partitions.end(),
@@ -462,7 +512,7 @@ index_state::collect_query_actors(query_state& lookup,
   while (it != last && result.size() < num_partitions) {
     auto partition_id = *it++;
     if (auto partition_actor = spin_up(partition_id))
-      result.push_back(std::make_pair(partition_id, partition_actor));
+      result.emplace_back(partition_id, partition_actor);
   }
   lookup.partitions.erase(lookup.partitions.begin(), it);
   VAST_DEBUG("{} launched {} partition actors to evaluate query", self,
@@ -470,16 +520,16 @@ index_state::collect_query_actors(query_state& lookup,
   return result;
 }
 
-path index_state::index_filename(path basename) const {
+std::filesystem::path
+index_state::index_filename(const std::filesystem::path& basename) const {
   return basename / dir / "index.bin";
 }
 
 caf::expected<flatbuffers::Offset<fbs::Index>>
 pack(flatbuffers::FlatBufferBuilder& builder, const index_state& state) {
-  VAST_DEBUG("{} persists {} uuids of definitely persisted and {}"
+  VAST_DEBUG("index persists {} uuids of definitely persisted and {}"
              "uuids of maybe persisted partitions",
-             state.self, state.persisted_partitions.size(),
-             state.unpersisted.size());
+             state.persisted_partitions.size(), state.unpersisted.size());
   std::vector<flatbuffers::Offset<fbs::uuid::v0>> partition_offsets;
   for (auto uuid : state.persisted_partitions) {
     if (auto uuid_fb = pack(builder, uuid))
@@ -491,7 +541,7 @@ pack(flatbuffers::FlatBufferBuilder& builder, const index_state& state) {
   // of the system is shut down (in case of a hard/dirty shutdown), so we just
   // store everything and throw out the missing partitions when loading the
   // index.
-  for (auto& kv : state.unpersisted) {
+  for (const auto& kv : state.unpersisted) {
     if (auto uuid_fb = pack(builder, kv.first))
       partition_offsets.push_back(*uuid_fb);
     else
@@ -499,7 +549,7 @@ pack(flatbuffers::FlatBufferBuilder& builder, const index_state& state) {
   }
   auto partitions = builder.CreateVector(partition_offsets);
   std::vector<flatbuffers::Offset<fbs::layout_statistics::v0>> stats_offsets;
-  for (auto& [name, layout_stats] : state.stats.layouts) {
+  for (const auto& [name, layout_stats] : state.stats.layouts) {
     auto name_fb = builder.CreateString(name);
     fbs::layout_statistics::v0Builder stats_builder(builder);
     stats_builder.add_name(name_fb);
@@ -525,13 +575,12 @@ void index_state::flush_to_disk() {
   auto builder = flatbuffers::FlatBufferBuilder{};
   auto index = pack(builder, *this);
   if (!index) {
-    VAST_WARN("{} failed to pack index: {}", self, render(index.error()));
+    VAST_WARN("{} failed to pack index: {}", self, index.error());
     return;
   }
   auto chunk = fbs::release(builder);
   self
-    ->request(caf::actor_cast<caf::actor>(filesystem), caf::infinite,
-              atom::write_v, index_filename(), chunk)
+    ->request(filesystem, caf::infinite, atom::write_v, index_filename(), chunk)
     .then(
       [=](atom::ok) {
         VAST_DEBUG("{} successfully persisted index state", self);
@@ -542,10 +591,11 @@ void index_state::flush_to_disk() {
 }
 
 index_actor::behavior_type
-index(index_actor::stateful_pointer<index_state> self,
-      filesystem_actor filesystem, path dir, size_t partition_capacity,
-      size_t max_inmem_partitions, size_t taste_partitions, size_t num_workers,
-      path meta_index_dir, double meta_index_fp_rate) {
+index(index_actor::stateful_pointer<index_state> self, store_actor store,
+      filesystem_actor filesystem, const std::filesystem::path& dir,
+      size_t partition_capacity, size_t max_inmem_partitions,
+      size_t taste_partitions, size_t num_workers,
+      const std::filesystem::path& meta_index_dir, double meta_index_fp_rate) {
   VAST_TRACE_SCOPE("{} {} {} {} {} {} {}", VAST_ARG(filesystem), VAST_ARG(dir),
                    VAST_ARG(partition_capacity), VAST_ARG(max_inmem_partitions),
                    VAST_ARG(taste_partitions), VAST_ARG(num_workers),
@@ -557,7 +607,10 @@ index(index_actor::stateful_pointer<index_state> self,
     VAST_VERBOSE("{} uses {} for meta index data", self, meta_index_dir);
   // Set members.
   self->state.self = self;
+  self->state.accept_queries = true;
+  self->state.store = std::move(store);
   self->state.filesystem = std::move(filesystem);
+  self->state.meta_index = self->spawn<caf::lazy_init>(meta_index);
   self->state.dir = dir;
   self->state.synopsisdir = meta_index_dir;
   self->state.partition_capacity = partition_capacity;
@@ -565,6 +618,7 @@ index(index_actor::stateful_pointer<index_state> self,
   self->state.inmem_partitions.factory().filesystem() = self->state.filesystem;
   self->state.inmem_partitions.resize(max_inmem_partitions);
   self->state.meta_index_fp_rate = meta_index_fp_rate;
+  self->state.meta_index_bytes = 0;
   // Read persistent state.
   if (auto err = self->state.load_from_disk()) {
     VAST_ERROR("{} failed to load index state from disk: {}", self,
@@ -606,9 +660,13 @@ index(index_actor::stateful_pointer<index_state> self,
       }
     },
     [self](caf::unit_t&, const caf::error& err) {
+      // During "normal" shutdown, the node will send an exit message to
+      // the importer which then cuts the stream to the index, and the
+      // index exits afterwards.
       // We get an 'unreachable' error when the stream becomes unreachable
-      // because the actor was destroyed; in this case we can't use `self`
-      // anymore.
+      // during actor destruction; in this case we can't use `self->state`
+      // anymore since it will already be destroyed.
+      VAST_DEBUG("index finalized streaming with error {}", render(err));
       if (err && err != caf::exit_reason::unreachable) {
         if (err != caf::exit_reason::user_shutdown)
           VAST_ERROR("{} got a stream error: {}", self, render(err));
@@ -618,16 +676,15 @@ index(index_actor::stateful_pointer<index_state> self,
         // importer.
         self->send_exit(self, err);
       }
-      VAST_DEBUG("index finalized streaming");
     });
   self->set_exit_handler([self](const caf::exit_msg& msg) {
     VAST_DEBUG("{} received EXIT from {} with reason: {}", self, msg.source,
                msg.reason);
     // Flush buffered batches and end stream.
+    self->state.stage->shutdown(); // closes inbound paths
     self->state.stage->out().fan_out_flush();
+    self->state.stage->out().close(); // closes outbound paths
     self->state.stage->out().force_emit_batches();
-    self->state.stage->out().close();
-    self->state.stage->shutdown();
     // Bring down active partition.
     if (self->state.active_partition.actor)
       self->state.decomission_active_partition();
@@ -667,12 +724,18 @@ index(index_actor::stateful_pointer<index_state> self,
       self->state.accountant = std::move(accountant);
     },
     [self](atom::subscribe, atom::flush, flush_listener_actor listener) {
+      VAST_DEBUG("{} adds flush listener", self);
       self->state.add_flush_listener(std::move(listener));
     },
-    [self](vast::expression expr) -> caf::result<void> {
+    [self](vast::query query) -> caf::result<void> {
+      if (!self->state.accept_queries) {
+        VAST_VERBOSE("{} delays query {} because it is still starting up", self,
+                     query);
+        return caf::skip;
+      }
       // TODO: This check is not required technically, but we use the query
-      // supervisor availability to rate-limit meta-index lookups. Do we really
-      // need this?
+      // supervisor availability to rate-limit meta-index lookups. Do we
+      // really need this?
       if (!self->state.worker_available())
         return caf::skip;
       // Query handling
@@ -698,36 +761,61 @@ index(index_actor::stateful_pointer<index_state> self,
         respond(caf::sec::invalid_argument);
         return {};
       }
-      // Get all potentially matching partitions.
-      auto candidates = self->state.meta_idx.lookup(expr);
+      std::vector<uuid> candidates;
       if (self->state.active_partition.actor)
         candidates.push_back(self->state.active_partition.id);
       for (const auto& [id, _] : self->state.unpersisted)
         candidates.push_back(id);
-      if (candidates.empty()) {
-        VAST_DEBUG("{} returns without result: no partitions qualify", self);
-        no_result();
-        return {};
-      }
-      // Allows the client to query further results after initial taste.
-      auto query_id = uuid::random();
-      // Ensure the query id is unique.
-      while (self->state.pending.find(query_id) != self->state.pending.end()
-             || query_id == uuid::nil())
-        query_id = uuid::random();
-      auto total = candidates.size();
-      auto scheduled = detail::narrow<uint32_t>(
-        std::min(candidates.size(), self->state.taste_partitions));
-      auto lookup = query_state{query_id, expr, std::move(candidates)};
-      auto result = self->state.pending.emplace(query_id, std::move(lookup));
-      VAST_ASSERT(result.second);
-      respond(query_id, detail::narrow<uint32_t>(total), scheduled);
-      self->delegate(caf::actor_cast<caf::actor>(self), query_id, scheduled);
-      return {};
+      auto rp = self->make_response_promise<void>();
+      // Get all potentially matching partitions.
+      self->request(self->state.meta_index, caf::infinite, query.expr)
+        .then(
+          [=, candidates = std::move(candidates)](
+            std::vector<uuid> midx_candidates) mutable {
+            VAST_DEBUG("{} got initial candidates {} and from meta-index {}",
+                       self, candidates, midx_candidates);
+            candidates.insert(candidates.end(), midx_candidates.begin(),
+                              midx_candidates.end());
+            std::sort(candidates.begin(), candidates.end());
+            candidates.erase(std::unique(candidates.begin(), candidates.end()),
+                             candidates.end());
+            if (candidates.empty()) {
+              VAST_DEBUG("{} returns without result: no partitions qualify",
+                         self);
+              no_result();
+              // TODO: When updating to CAF 0.18, remove the use of the
+              // untyped response promise and call deliver without arguments.
+              auto& untyped_rp = static_cast<caf::response_promise&>(rp);
+              untyped_rp.deliver(caf::unit);
+              return;
+            }
+            // Allows the client to query further results after initial taste.
+            auto query_id = uuid::random();
+            // Ensure the query id is unique.
+            while (self->state.pending.find(query_id)
+                     != self->state.pending.end()
+                   || query_id == uuid::nil())
+              query_id = uuid::random();
+            auto total = candidates.size();
+            auto scheduled = detail::narrow<uint32_t>(
+              std::min(candidates.size(), self->state.taste_partitions));
+            auto lookup = query_state{query_id, query, std::move(candidates)};
+            auto result
+              = self->state.pending.emplace(query_id, std::move(lookup));
+            VAST_ASSERT(result.second);
+            respond(query_id, detail::narrow<uint32_t>(total), scheduled);
+            rp.delegate(caf::actor_cast<caf::actor>(self), query_id, scheduled);
+          },
+          [=](caf::error err) mutable {
+            VAST_ERROR("{} failed to receive candidates from meta-index: {}",
+                       self, render(err));
+            rp.deliver(std::move(err));
+          });
+      return rp;
     },
     [self](const uuid& query_id, uint32_t num_partitions) -> caf::result<void> {
       auto sender = self->current_sender();
-      auto client = caf::actor_cast<index_client_actor>(sender);
+      auto client = caf::actor_cast<receiver_actor<atom::done>>(sender);
       // Sanity checks.
       if (!sender) {
         VAST_ERROR("{} ignores an anonymous query", self);
@@ -758,7 +846,7 @@ index(index_actor::stateful_pointer<index_state> self,
       VAST_DEBUG("{} schedules {} more partition(s) for query id {}"
                  "with {} partitions remaining",
                  self, actors.size(), query_id, query_state.partitions.size());
-      self->send(*worker, query_state.expression, std::move(actors), client);
+      self->send(*worker, query_state.query, std::move(actors), client);
       // Cleanup if we exhausted all candidates.
       if (query_state.partitions.empty())
         self->state.pending.erase(iter);
@@ -770,9 +858,13 @@ index(index_actor::stateful_pointer<index_state> self,
       auto path = self->state.partition_path(partition_id);
       auto synopsis_path = self->state.partition_synopsis_path(partition_id);
       bool adjust_stats = true;
-      if (!self->state.persisted_partitions.count(partition_id)) {
-        if (!exists(path)) {
-          rp.deliver(caf::make_error(ec::logic_error, "unknown partition"));
+      if (self->state.persisted_partitions.count(partition_id) == 0u) {
+        std::error_code err{};
+        const auto file_exists = std::filesystem::exists(path, err);
+        if (!file_exists) {
+          rp.deliver(caf::make_error(
+            ec::logic_error, fmt::format("unknown partition for path {}: {}",
+                                         path, err.message())));
           return rp;
         }
         // As a special case, if the partition exists on disk we just continue
@@ -782,21 +874,31 @@ index(index_actor::stateful_pointer<index_state> self,
       }
       self->state.inmem_partitions.drop(partition_id);
       self->state.persisted_partitions.erase(partition_id);
+      self
+        ->request(self->state.meta_index, caf::infinite, atom::erase_v,
+                  partition_id)
+        .then([](atom::ok) { /* nop */ },
+              [partition_id](const caf::error& err) {
+                VAST_WARN("index encountered an error trying to erase "
+                          "partition {} from the meta index: {}",
+                          partition_id, err);
+              });
       self->request(self->state.filesystem, caf::infinite, atom::mmap_v, path)
         .then(
-          [=](chunk_ptr chunk) mutable {
+          [=](const chunk_ptr& chunk) mutable {
             // Adjust layout stats by subtracting the events of the removed
             // partition.
-            auto partition = fbs::GetPartition(chunk->data());
+            const auto* partition = fbs::GetPartition(chunk->data());
             if (partition->partition_type() != fbs::partition::Partition::v0) {
-              rp.deliver(caf::make_error(ec::format_error, "unexpected format "
+              rp.deliver(caf::make_error(ec::format_error, "unexpected "
+                                                           "format "
                                                            "version"));
               return;
             }
             vast::ids all_ids;
-            auto partition_v0 = partition->partition_as_v0();
-            for (auto partition_stats : *partition_v0->type_ids()) {
-              auto name = partition_stats->name();
+            const auto* partition_v0 = partition->partition_as_v0();
+            for (const auto* partition_stats : *partition_v0->type_ids()) {
+              const auto* name = partition_stats->name();
               vast::ids ids;
               if (auto error
                   = fbs::deserialize_bytes(partition_stats->ids(), ids)) {
@@ -810,17 +912,25 @@ index(index_actor::stateful_pointer<index_state> self,
               if (adjust_stats)
                 self->state.stats.layouts[name->str()].count -= rank(ids);
             }
-            // Note that mmap's will increase the reference count of a file, so
-            // unlinking should not affect indexers that are currently loaded
-            // and answering a query.
-            if (!rm(path))
+            // Note that mmap's will increase the reference count of a file,
+            // so unlinking should not affect indexers that are currently
+            // loaded and answering a query.
+            auto try_remove_all = [](const auto& p, const auto& error_fn) {
+              std::error_code err{};
+              std::filesystem::remove_all(p, err);
+              if (err)
+                error_fn();
+            };
+            try_remove_all(path, [&] {
               VAST_WARN("{} could not unlink partition at {}", self, path);
-            if (!rm(synopsis_path))
+            });
+            try_remove_all(synopsis_path, [&] {
               VAST_WARN("{} could not unlink partition synopsis at", self,
-                        path);
+                        synopsis_path);
+            });
             rp.deliver(std::move(all_ids));
           },
-          [=](caf::error e) mutable { rp.deliver(e); });
+          [=](caf::error& err) mutable { rp.deliver(std::move(err)); });
       return rp;
     },
     // -- query_supervisor_master_actor ----------------------------------------
