@@ -1,25 +1,22 @@
-/******************************************************************************
- *                    _   _____   __________                                  *
- *                   | | / / _ | / __/_  __/     Visibility                   *
- *                   | |/ / __ |_\ \  / /          Across                     *
- *                   |___/_/ |_/___/ /_/       Space and Time                 *
- *                                                                            *
- * This file is part of VAST. It is subject to the license terms in the       *
- * LICENSE file found in the top-level directory of this distribution and at  *
- * http://vast.io/license. No part of VAST, including this file, may be       *
- * copied, modified, propagated, or distributed except according to the terms *
- * contained in the LICENSE file.                                             *
- ******************************************************************************/
+//    _   _____   __________
+//   | | / / _ | / __/_  __/     Visibility
+//   | |/ / __ |_\ \  / /          Across
+//   |___/_/ |_/___/ /_/       Space and Time
+//
+// SPDX-FileCopyrightText: (c) 2016 The VAST Contributors
+// SPDX-License-Identifier: BSD-3-Clause
 
 #include "vast/system/node.hpp"
 
 #include "vast/fwd.hpp"
 
 #include "vast/accountant/config.hpp"
+#include "vast/concept/convertible/to.hpp"
 #include "vast/concept/parseable/to.hpp"
 #include "vast/concept/parseable/vast/endpoint.hpp"
 #include "vast/concept/printable/stream.hpp"
 #include "vast/concept/printable/to_string.hpp"
+#include "vast/concept/printable/vast/expression.hpp"
 #include "vast/concept/printable/vast/json.hpp"
 #include "vast/config.hpp"
 #include "vast/data.hpp"
@@ -36,6 +33,7 @@
 #include "vast/format/test.hpp"
 #include "vast/format/zeek.hpp"
 #include "vast/logger.hpp"
+#include "vast/plugin.hpp"
 #include "vast/system/accountant.hpp"
 #include "vast/system/node.hpp"
 #include "vast/system/posix_filesystem.hpp"
@@ -55,14 +53,12 @@
 #include "vast/system/spawn_sink.hpp"
 #include "vast/system/spawn_source.hpp"
 #include "vast/system/spawn_type_registry.hpp"
+#include "vast/system/status_verbosity.hpp"
 #include "vast/system/terminate.hpp"
+#include "vast/system/version_command.hpp"
 #include "vast/table_slice.hpp"
 #include "vast/taxonomies.hpp"
 
-#if VAST_ENABLE_PCAP
-#  include "vast/format/pcap.hpp"
-#endif
-
 #include <caf/function_view.hpp>
 #include <caf/io/middleman.hpp>
 #include <caf/settings.hpp>
@@ -86,6 +82,17 @@ auto make_error_msg(ec code, std::string msg) {
   return caf::make_message(caf::make_error(code, std::move(msg)));
 }
 
+/// A list of components that are essential for importing and exporting data
+/// from the node.
+std::set<const char*> core_components
+  = {"archive", "filesystem", "importer", "index"};
+
+bool is_core_component(std::string_view type) {
+  auto pred = [&](const char* x) { return x == type; };
+  return std::any_of(std::begin(core_components), std::end(core_components),
+                     pred);
+}
+
 /// Helper function to determine whether a component can be spawned at most
 /// once.
 bool is_singleton(std::string_view type) {
@@ -98,8 +105,8 @@ bool is_singleton(std::string_view type) {
   // refactoring will be much easier once the NODE itself is a typed actor, so
   // let's hold off until then.
   const char* singletons[]
-    = {"accountant", "archive", "eraser",       "filesystem",
-       "importer",   "index",   "type-registry"};
+    = {"accountant", "archive",  "disk-monitor", "eraser",
+       "filesystem", "importer", "index",        "type-registry"};
   auto pred = [&](const char* x) { return x == type; };
   return std::any_of(std::begin(singletons), std::end(singletons), pred);
 }
@@ -137,12 +144,25 @@ void collect_component_status(node_actor::stateful_pointer<node_state> self,
   };
   auto req_state = std::make_shared<req_state_t>();
   req_state->rp = std::move(status_promise);
+  // Pre-fill the version information. Note that we must remove all nil values
+  // first, as the conversion to a caf::settings object fails otherwise. We can
+  // remove the cleansing step once we use record instead of caf::settings for
+  // retrieving the status.
+  auto version_data = retrieve_versions();
+  for (auto& [_, value] : version_data)
+    if (value == caf::none)
+      value = record{};
+  if (auto version = to<caf::settings>(version_data))
+    put(req_state->content, "version", *version);
+  else
+    VAST_WARN("{} failed to add remote version to status: {}", self,
+              version.error());
   // Pre-fill our result with system stats.
   auto& sys = self->system();
   auto& system = put_dictionary(req_state->content, "system");
   if (v >= status_verbosity::info) {
     put(system, "in-memory-table-slices", table_slice::instances());
-    put(system, "database-path", self->state.dir.str());
+    put(system, "database-path", self->state.dir.string());
     detail::merge_settings(detail::get_status(), system);
   }
   if (v >= status_verbosity::debug) {
@@ -187,6 +207,36 @@ void collect_component_status(node_actor::stateful_pointer<node_state> self,
   }
 }
 
+/// Registers (and monitors) a component through the node.
+caf::error
+register_component(node_actor::stateful_pointer<node_state> self,
+                   const caf::actor& component, const std::string& type,
+                   const std::string& label = {}) {
+  if (!self->state.registry.add(component, type, label)) {
+    auto msg // separate variable for clang-format only
+      = fmt::format("{} failed to add component to registry: {}", self,
+                    label.empty() ? type : label);
+    return caf::make_error(ec::unspecified, std::move(msg));
+  }
+  self->monitor(component);
+  return caf::none;
+}
+
+/// Deregisters (and demonitors) a component through the node.
+caf::expected<caf::actor>
+deregister_component(node_actor::stateful_pointer<node_state> self,
+                     const std::string& label) {
+  auto component = self->state.registry.remove(label);
+  if (!component) {
+    auto msg // separate variable for clang-format only
+      = fmt::format("{} failed to deregister non-existant component: {}", self,
+                    label);
+    return caf::make_error(ec::unspecified, std::move(msg));
+  }
+  self->demonitor(component->actor);
+  return component->actor;
+}
+
 } // namespace
 
 caf::message dump_command(const invocation& inv, caf::actor_system&) {
@@ -308,19 +358,20 @@ caf::message kill_command(const invocation& inv, caf::actor_system&) {
                                             "argument");
   auto rp = self->make_response_promise();
   auto& label = *first;
-  auto component = self->state.registry.find_by_label(label);
+  auto component = deregister_component(self, label);
   if (!component) {
-    rp.deliver(caf::make_error(ec::unspecified, "no such component: " + label));
+    rp.deliver(caf::make_error(ec::unspecified,
+                               fmt::format("no such component: {}", label)));
   } else {
-    self->demonitor(component);
-    terminate<policy::parallel>(self, component)
+    terminate<policy::parallel>(self, std::move(*component))
       .then(
         [=](atom::done) mutable {
           VAST_DEBUG("{} terminated component {}", self, label);
           rp.deliver(atom::ok_v);
         },
         [=](const caf::error& err) mutable {
-          VAST_DEBUG("{} terminated component {}", self, label);
+          VAST_WARN("{} failed to terminate component {}: {}", self, label,
+                    err);
           rp.deliver(err);
         });
   }
@@ -345,51 +396,48 @@ node_state::component_factory_fun lift_component_factory() {
 }
 
 auto make_component_factory() {
-  return node_state::named_component_factory {
+  auto result = node_state::named_component_factory{
     {"spawn accountant", lift_component_factory<spawn_accountant>()},
-      {"spawn archive", lift_component_factory<spawn_archive>()},
-      {"spawn counter", lift_component_factory<spawn_counter>()},
-      {"spawn disk_monitor", lift_component_factory<spawn_disk_monitor>()},
-      {"spawn eraser", lift_component_factory<spawn_eraser>()},
-      {"spawn exporter", lift_component_factory<spawn_exporter>()},
-      {"spawn explorer", lift_component_factory<spawn_explorer>()},
-      {"spawn importer", lift_component_factory<spawn_importer>()},
-      {"spawn type-registry", lift_component_factory<spawn_type_registry>()},
-      {"spawn index", lift_component_factory<spawn_index>()},
-      {"spawn pivoter", lift_component_factory<spawn_pivoter>()},
-      {"spawn source csv",
-       lift_component_factory<spawn_source<format::csv::reader>>()},
-      {"spawn source json",
-       lift_component_factory<
-         spawn_source<format::json::reader<format::json::default_selector>>>()},
-#if VAST_ENABLE_PCAP
-      {"spawn source pcap",
-       lift_component_factory<spawn_source<format::pcap::reader>>()},
-#endif
-      {"spawn source suricata",
-       lift_component_factory<
-         spawn_source<format::json::reader<format::json::suricata_selector>>>()},
-      {"spawn source syslog",
-       lift_component_factory<spawn_source<format::syslog::reader>>()},
-      {"spawn source test",
-       lift_component_factory<spawn_source<format::test::reader>>()},
-      {"spawn source zeek",
-       lift_component_factory<spawn_source<format::zeek::reader>>()},
-      {"spawn source zeek-json",
-       lift_component_factory<
-         spawn_source<format::json::reader<format::json::zeek_selector>>>()},
-      {"spawn sink pcap", lift_component_factory<spawn_pcap_sink>()},
-      {"spawn sink zeek", lift_component_factory<spawn_zeek_sink>()},
-      {"spawn sink csv", lift_component_factory<spawn_csv_sink>()},
-      {"spawn sink ascii", lift_component_factory<spawn_ascii_sink>()},
-      {"spawn sink json", lift_component_factory<spawn_json_sink>()},
+    {"spawn archive", lift_component_factory<spawn_archive>()},
+    {"spawn counter", lift_component_factory<spawn_counter>()},
+    {"spawn disk-monitor", lift_component_factory<spawn_disk_monitor>()},
+    {"spawn eraser", lift_component_factory<spawn_eraser>()},
+    {"spawn exporter", lift_component_factory<spawn_exporter>()},
+    {"spawn explorer", lift_component_factory<spawn_explorer>()},
+    {"spawn importer", lift_component_factory<spawn_importer>()},
+    {"spawn type-registry", lift_component_factory<spawn_type_registry>()},
+    {"spawn index", lift_component_factory<spawn_index>()},
+    {"spawn pivoter", lift_component_factory<spawn_pivoter>()},
+    {"spawn source", lift_component_factory<spawn_source>()},
+    {"spawn source csv", lift_component_factory<spawn_source>()},
+    {"spawn source json", lift_component_factory<spawn_source>()},
+    {"spawn source suricata", lift_component_factory<spawn_source>()},
+    {"spawn source syslog", lift_component_factory<spawn_source>()},
+    {"spawn source test", lift_component_factory<spawn_source>()},
+    {"spawn source zeek", lift_component_factory<spawn_source>()},
+    {"spawn source zeek-json", lift_component_factory<spawn_source>()},
+    {"spawn sink zeek", lift_component_factory<spawn_sink>()},
+    {"spawn sink csv", lift_component_factory<spawn_sink>()},
+    {"spawn sink ascii", lift_component_factory<spawn_sink>()},
+    {"spawn sink json", lift_component_factory<spawn_sink>()},
   };
+  for (const auto& plugin : plugins::get()) {
+    if (const auto* reader = plugin.as<reader_plugin>()) {
+      auto command = fmt::format("spawn source {}", reader->reader_format());
+      result.emplace(command, lift_component_factory<spawn_source>());
+    }
+    if (const auto* writer = plugin.as<writer_plugin>()) {
+      auto command = fmt::format("spawn sink {}", writer->writer_format());
+      result.emplace(command, lift_component_factory<spawn_sink>());
+    }
+  }
+  return result;
 }
 
 auto make_command_factory() {
   // When updating this list, remember to update its counterpart in
   // application.cpp as well iff necessary
-  return command::factory{
+  auto result = command::factory{
     {"dump", dump_command},
     {"dump concepts", dump_command},
     {"dump models", dump_command},
@@ -398,7 +446,7 @@ auto make_command_factory() {
     {"spawn accountant", node_state::spawn_command},
     {"spawn archive", node_state::spawn_command},
     {"spawn counter", node_state::spawn_command},
-    {"spawn disk_monitor", node_state::spawn_command},
+    {"spawn disk-monitor", node_state::spawn_command},
     {"spawn eraser", node_state::spawn_command},
     {"spawn explorer", node_state::spawn_command},
     {"spawn exporter", node_state::spawn_command},
@@ -409,11 +457,9 @@ auto make_command_factory() {
     {"spawn sink ascii", node_state::spawn_command},
     {"spawn sink csv", node_state::spawn_command},
     {"spawn sink json", node_state::spawn_command},
-    {"spawn sink pcap", node_state::spawn_command},
     {"spawn sink zeek", node_state::spawn_command},
     {"spawn source csv", node_state::spawn_command},
     {"spawn source json", node_state::spawn_command},
-    {"spawn source pcap", node_state::spawn_command},
     {"spawn source suricata", node_state::spawn_command},
     {"spawn source syslog", node_state::spawn_command},
     {"spawn source test", node_state::spawn_command},
@@ -421,6 +467,17 @@ auto make_command_factory() {
     {"spawn source zeek-json", node_state::spawn_command},
     {"status", status_command},
   };
+  for (const auto& plugin : plugins::get()) {
+    if (const auto* reader = plugin.as<reader_plugin>()) {
+      auto command = fmt::format("spawn source {}", reader->reader_format());
+      result.emplace(command, node_state::spawn_command);
+    }
+    if (const auto* writer = plugin.as<writer_plugin>()) {
+      auto command = fmt::format("spawn sink {}", writer->writer_format());
+      result.emplace(command, node_state::spawn_command);
+    }
+  }
+  return result;
 }
 
 std::string generate_label(node_actor::stateful_pointer<node_state> self,
@@ -436,7 +493,10 @@ node_state::spawn_command(const invocation& inv,
   VAST_TRACE_SCOPE("{}", inv);
   using std::begin;
   using std::end;
-  auto self = this_node;
+  auto* self = this_node;
+  if (self->state.tearing_down)
+    return caf::make_message(caf::make_error( //
+      ec::no_error, "can't spawn a component while tearing down"));
   auto rp = self->make_response_promise();
   // We configured the command to have the name of the component.
   auto inv_name_split = detail::split(inv.full_name, " ");
@@ -447,6 +507,11 @@ node_state::spawn_command(const invocation& inv,
   if (auto label_ptr = caf::get_if<std::string>(&inv.options, "vast.spawn."
                                                               "label")) {
     label = *label_ptr;
+    if (label.empty()) {
+      auto err = caf::make_error(ec::unspecified, "empty component label");
+      rp.deliver(err);
+      return caf::make_message(std::move(err));
+    }
     if (self->state.registry.find_by_label(label)) {
       auto err = caf::make_error(ec::unspecified, "duplicate component label");
       rp.deliver(err);
@@ -481,10 +546,11 @@ node_state::spawn_command(const invocation& inv,
       rp.deliver(component.error());
       return caf::make_message(std::move(component.error()));
     }
-    self->monitor(*component);
-    auto okay = self->state.registry.add(*component, std::move(comp_type),
-                                         std::move(label));
-    VAST_ASSERT(okay);
+    if (auto err = register_component(self, *component, comp_type, label)) {
+      rp.deliver(err);
+      return caf::make_message(std::move(err));
+    }
+    VAST_DEBUG("{} registered {} as {}", self, comp_type, label);
     rp.deliver(*component);
     return caf::make_message(*component);
   };
@@ -522,7 +588,8 @@ node_state::spawn_command(const invocation& inv,
 }
 
 node_actor::behavior_type
-node(node_actor::stateful_pointer<node_state> self, std::string name, path dir,
+node(node_actor::stateful_pointer<node_state> self, std::string name,
+     std::filesystem::path dir,
      std::chrono::milliseconds shutdown_grace_period) {
   self->state.name = std::move(name);
   self->state.dir = std::move(dir);
@@ -530,71 +597,88 @@ node(node_actor::stateful_pointer<node_state> self, std::string name, path dir,
   node_state::component_factory = make_component_factory();
   node_state::command_factory = make_command_factory();
   // Initialize the file system with the node directory as root.
-  auto fs = self->spawn<caf::linked + caf::detached>(posix_filesystem,
-                                                     self->state.dir);
-  self->state.registry.add(caf::actor_cast<caf::actor>(fs), "filesystem");
+  auto fs = self->spawn<caf::detached>(posix_filesystem, self->state.dir);
+  auto err
+    = register_component(self, caf::actor_cast<caf::actor>(fs), "filesystem");
+  VAST_ASSERT(err == caf::none); // Registration cannot fail; empty registry.
   // Remove monitored components.
   self->set_down_handler([=](const caf::down_msg& msg) {
     VAST_DEBUG("{} got DOWN from {}", self, msg.source);
-    auto component = caf::actor_cast<caf::actor>(msg.source);
-    auto type = self->state.registry.find_type_for(component);
-    // All monitored components are in the registry.
-    VAST_ASSERT(type != nullptr);
-    if (is_singleton(*type)) {
-      auto label = self->state.registry.find_label_for(component);
-      VAST_ASSERT(label != nullptr); // Per the above assertion.
-      VAST_ERROR("{} got DOWN from {} ; initiating shutdown", self, *label);
-      self->send_exit(self, caf::exit_reason::user_shutdown);
+    if (!self->state.tearing_down) {
+      auto actor = caf::actor_cast<caf::actor>(msg.source);
+      auto component = self->state.registry.remove(actor);
+      VAST_ASSERT(component);
+      // Terminate if a singleton dies.
+      if (is_core_component(component->type)) {
+        VAST_ERROR("{} terminates after DOWN from {}", self, component->type);
+        self->send_exit(self, caf::exit_reason::user_shutdown);
+      }
     }
-    self->state.registry.remove(component);
   });
   // Terminate deterministically on shutdown.
   self->set_exit_handler([=](const caf::exit_msg& msg) {
     VAST_DEBUG("{} got EXIT from {}", self, msg.source);
-    auto& registry = self->state.registry;
-    std::vector<caf::actor> actors;
-    auto schedule_teardown = [&](caf::actor actor) {
-      self->demonitor(actor);
-      registry.remove(actor);
-      actors.push_back(std::move(actor));
-    };
-    // Terminate the accountant first because it acts like a source and may
-    // hold buffered data.
-    if (auto [accountant] = registry.find<accountant_actor>(); accountant)
-      schedule_teardown(caf::actor_cast<caf::actor>(std::move(accountant)));
-    // Take out the filesystem, which we terminate at the very end.
-    auto [filesystem] = registry.find<filesystem_actor>();
-    VAST_ASSERT(filesystem);
-    self->unlink_from(filesystem); // avoid receiving an unneeded EXIT
-    registry.remove(caf::actor_cast<caf::actor>(filesystem));
-    // Tear down the ingestion pipeline from source to sink.
-    auto pipeline = {"source", "importer", "index", "archive", "exporter"};
-    for (auto component : pipeline)
-      for (auto actor : registry.find_by_type(component)) {
-        schedule_teardown(std::move(actor));
+    self->state.tearing_down = true;
+    // Ignore duplicate EXIT messages except for hard kills.
+    self->set_exit_handler([=](const caf::exit_msg& msg) {
+      if (msg.reason == caf::exit_reason::kill) {
+        VAST_WARN("{} received hard kill and terminates immediately", self);
+        self->quit(msg.reason);
+      } else {
+        VAST_DEBUG("{} ignores duplicate EXIT message from {}", self,
+                   msg.source);
       }
-    // Now terminate everything else.
-    std::vector<caf::actor> remaining;
-    for ([[maybe_unused]] auto& [label, comp] : registry.components()) {
-      remaining.push_back(comp.actor);
+    });
+    auto& registry = self->state.registry;
+    // Core components are terminated in a second stage, we remove them from the
+    // registry upfront and deal with them later.
+    std::vector<caf::actor> core_shutdown_handles;
+    // The components listed here need to be terminated in sequential order.
+    // The importer needs to shut down first because it might still have
+    // buffered data. The index uses the archive for querying. The filesystem
+    // is needed by all others for the persisting logic.
+    auto shutdown_sequence = std::initializer_list<const char*>{
+      "importer", "index", "archive", "filesystem"};
+    // Make sure that these remain in sync.
+    VAST_ASSERT(std::set<const char*>{shutdown_sequence} == core_components);
+    for (const char* name : shutdown_sequence) {
+      if (auto comp = registry.remove(name))
+        core_shutdown_handles.push_back(comp->actor);
     }
-    for (auto& actor : remaining)
-      schedule_teardown(actor);
-    // Finally, bring down the filesystem.
-    // FIXME: there's a super-annoying bug that makes it impossible to receive a
-    // DOWN message from the filesystem during shutdown, but *only* when the
-    // filesystem is detached! This might be related to a bug we experienced
-    // earlier: https://github.com/actor-framework/actor-framework/issues/1110.
-    // Until it gets fixed, we cannot add the filesystem to the set of
-    // sequentially terminated actors but instead let it implicitly terminate
-    // after the node exits when the filesystem ref count goes to 0. (A
-    // shutdown after the node won't be an issue because the filesystem is
-    // currently stateless, but this needs to be reconsidered when it changes.)
-    // // TODO: uncomment when we get a DOWN from detached actors.
-    // actors.push_back(std::move(filesystem));
+    std::vector<caf::actor> aux_components;
+    for (const auto& [_, comp] : registry.components()) {
+      self->demonitor(comp.actor);
+      // Ignore remote actors.
+      if (comp.actor->node() != self->node())
+        continue;
+      aux_components.push_back(comp.actor);
+    }
+    // Drop everything.
+    registry.clear();
     auto shutdown_kill_timeout = shutdown_grace_period / 5;
-    shutdown<policy::sequential>(self, std::move(actors), shutdown_grace_period,
-                                 shutdown_kill_timeout);
+    auto core_shutdown_sequence =
+      [=, core_shutdown_handles = std::move(core_shutdown_handles)]() mutable {
+        for (const auto& comp : core_shutdown_handles)
+          self->demonitor(comp);
+        shutdown<policy::sequential>(self, std::move(core_shutdown_handles),
+                                     shutdown_grace_period,
+                                     shutdown_kill_timeout);
+      };
+    terminate<policy::parallel>(self, std::move(aux_components),
+                                shutdown_grace_period, shutdown_kill_timeout)
+      .then(
+        [self, core_shutdown_sequence](atom::done) mutable {
+          VAST_DEBUG("{} terminated auxiliary actors, commencing core shutdown "
+                     "sequence...",
+                     self);
+          core_shutdown_sequence();
+        },
+        [self, core_shutdown_sequence](const caf::error& err) mutable {
+          VAST_ERROR("{} failed to cleanly terminate auxiliary actors {}, "
+                     "shutting down core components",
+                     self, err);
+          core_shutdown_sequence();
+        });
   });
   // Define the node behavior.
   return {
@@ -605,6 +689,28 @@ node(node_actor::stateful_pointer<node_state> self, std::string name, path dir,
       this_node = self;
       return run(inv, self->system(), node_state::command_factory);
     },
+    [self](atom::internal, atom::spawn, atom::plugin) -> caf::result<void> {
+      // Add all plugins to the component registry.
+      for (const auto& plugin : plugins::get()) {
+        if (const auto* component = plugin.as<component_plugin>()) {
+          if (auto handle = component->make_component(self); !handle)
+            return caf::make_error( //
+              ec::unspecified, fmt::format("{} failed to spawn component "
+                                           "plugin {}",
+                                           self, component->name()));
+          else if (auto err = register_component(
+                     self, caf::actor_cast<caf::actor>(handle),
+                     component->name());
+                   err && err != caf::no_error)
+            return caf::make_error( //
+              ec::unspecified, fmt::format("{} failed to register component "
+                                           "plugin {} in component registry: "
+                                           "{}",
+                                           self, component->name(), err));
+        }
+      }
+      return {};
+    },
     [self](atom::spawn, const invocation& inv) {
       VAST_DEBUG("{} got spawn command {} with options {} and arguments {}",
                  self, inv.full_name, inv.options, inv.arguments);
@@ -634,6 +740,8 @@ node(node_actor::stateful_pointer<node_state> self, std::string name, path dir,
     [self](atom::put, const caf::actor& component,
            const std::string& type) -> caf::result<atom::ok> {
       VAST_DEBUG("{} got new {}", self, type);
+      if (type.empty())
+        return caf::make_error(ec::unspecified, "empty component type");
       // Check if the new component is a singleton.
       auto& registry = self->state.registry;
       if (is_singleton(type) && registry.find_by_label(type))
@@ -641,9 +749,8 @@ node(node_actor::stateful_pointer<node_state> self, std::string name, path dir,
       // Generate label
       auto label = generate_label(self, type);
       VAST_DEBUG("{} generated new component label {}", self, label);
-      if (!registry.add(component, type, label))
-        return caf::make_error(ec::unspecified, "failed to add component");
-      self->monitor(component);
+      if (auto err = register_component(self, component, type, label))
+        return err;
       return atom::ok_v;
     },
     [self](atom::get, atom::type, const std::string& type) {
@@ -670,8 +777,8 @@ node(node_actor::stateful_pointer<node_state> self, std::string name, path dir,
                  result);
       return result;
     },
-    [](atom::get, atom::version) -> std::string { //
-      return VAST_VERSION;
+    [](atom::get, atom::version) { //
+      return retrieve_versions();
     },
     [self](atom::signal, int signal) {
       VAST_WARN("{} got signal {}", self, ::strsignal(signal));
