@@ -1,15 +1,10 @@
-/******************************************************************************
- *                    _   _____   __________                                  *
- *                   | | / / _ | / __/_  __/     Visibility                   *
- *                   | |/ / __ |_\ \  / /          Across                     *
- *                   |___/_/ |_/___/ /_/       Space and Time                 *
- *                                                                            *
- * This file is part of VAST. It is subject to the license terms in the       *
- * LICENSE file found in the top-level directory of this distribution and at  *
- * http://vast.io/license. No part of VAST, including this file, may be       *
- * copied, modified, propagated, or distributed except according to the terms *
- * contained in the LICENSE file.                                             *
- ******************************************************************************/
+//    _   _____   __________
+//   | | / / _ | / __/_  __/     Visibility
+//   | |/ / __ |_\ \  / /          Across
+//   |___/_/ |_/___/ /_/       Space and Time
+//
+// SPDX-FileCopyrightText: (c) 2016 The VAST Contributors
+// SPDX-License-Identifier: BSD-3-Clause
 
 #include "vast/system/partition.hpp"
 
@@ -24,8 +19,8 @@
 #include "vast/concept/printable/vast/table_slice.hpp"
 #include "vast/concept/printable/vast/uuid.hpp"
 #include "vast/detail/assert.hpp"
+#include "vast/detail/notifying_stream_manager.hpp"
 #include "vast/detail/settings.hpp"
-#include "vast/expression.hpp"
 #include "vast/expression_visitors.hpp"
 #include "vast/fbs/partition.hpp"
 #include "vast/fbs/utils.hpp"
@@ -49,17 +44,12 @@
 #include <caf/deserializer.hpp>
 #include <caf/error.hpp>
 #include <caf/sec.hpp>
-
 #include <flatbuffers/base.h> // FLATBUFFERS_MAX_BUFFER_SIZE
 #include <flatbuffers/flatbuffers.h>
 
+#include <filesystem>
 #include <memory>
 
-using namespace std::chrono;
-using namespace caf;
-
-CAF_ALLOW_UNSAFE_MESSAGE_TYPE(std::shared_ptr<vast::partition_synopsis>)
-
 namespace vast::system {
 
 /// Gets the ACTIVE INDEXER at a certain position.
@@ -68,6 +58,20 @@ active_indexer_actor active_partition_state::indexer_at(size_t position) const {
   return as_vector(indexers)[position].second;
 }
 
+void active_partition_state::add_flush_listener(flush_listener_actor listener) {
+  VAST_DEBUG("{} adds a new 'flush' subscriber: {}", self, listener);
+  flush_listeners.emplace_back(std::move(listener));
+  detail::notify_listeners_if_clean(*this, *stage);
+}
+
+void active_partition_state::notify_flush_listeners() {
+  VAST_DEBUG("{} sends 'flush' messages to {} listeners", self,
+             flush_listeners.size());
+  for (auto& listener : flush_listeners)
+    self->send(listener, atom::flush_v);
+  flush_listeners.clear();
+}
+
 /// Gets the INDEXER at a certain position.
 indexer_actor passive_partition_state::indexer_at(size_t position) const {
   VAST_ASSERT(position < indexers.size());
@@ -121,18 +125,18 @@ fetch_indexer(const PartitionState& state, const data_extractor& dx,
 /// @relates passive_partition_state
 template <typename PartitionState>
 indexer_actor
-fetch_indexer(const PartitionState& state, const attribute_extractor& ex,
+fetch_indexer(const PartitionState& state, const meta_extractor& ex,
               relational_operator op, const data& x) {
   VAST_TRACE_SCOPE("{} {} {}", VAST_ARG(ex), VAST_ARG(op), VAST_ARG(x));
   ids row_ids;
-  if (ex.attr == atom::type_v) {
+  if (ex.kind == meta_extractor::type) {
     // We know the answer immediately: all IDs that are part of the table.
     // However, we still have to "lift" this result into an actor for the
     // EVALUATOR.
     for (auto& [name, ids] : state.type_ids)
       if (evaluate(name, op, x))
         row_ids |= ids;
-  } else if (ex.attr == atom::field_v) {
+  } else if (ex.kind == meta_extractor::field) {
     auto s = caf::get_if<std::string>(&x);
     if (!s) {
       VAST_WARN("{} #field meta queries only support string "
@@ -161,7 +165,7 @@ fetch_indexer(const PartitionState& state, const attribute_extractor& ex,
       row_ids = partition_ids ^ row_ids;
     }
   } else {
-    VAST_WARN("{} got unsupported attribute: {}", state.self, ex.attr);
+    VAST_WARN("{} got unsupported attribute: {}", state.self, ex.kind);
     return {};
   }
   // TODO: Spawning a one-shot actor is quite expensive. Maybe the
@@ -194,7 +198,7 @@ evaluate(const PartitionState& state, const expression& expr) {
       return fetch_indexer(state, ext, pred.op, x);
     };
     auto v = detail::overload{
-      [&](const attribute_extractor& ex, const data& x) {
+      [&](const meta_extractor& ex, const data& x) {
         return get_indexer_handle(ex, x);
       },
       [&](const data_extractor& dx, const data& x) {
@@ -328,7 +332,7 @@ unpack(const fbs::partition::v0& partition, passive_partition_state& state) {
   if (state.combined_layout.fields.size() < indexes->size()) {
     VAST_ERROR("{} found incoherent number of indexers in deserialized "
                "state; {} fields for {} indexes",
-               state.self, state.combined_layout.fields.size(),
+               state.name, state.combined_layout.fields.size(),
                indexes->size());
     return caf::make_error(ec::format_error, "incoherent number of indexers");
   }
@@ -336,7 +340,7 @@ unpack(const fbs::partition::v0& partition, passive_partition_state& state) {
   // vector must be the same as in `combined_layout`. The actual indexers are
   // deserialized and spawned lazily on demand.
   state.indexers.resize(indexes->size());
-  VAST_DEBUG("{} found {} indexers for partition {}", state.self,
+  VAST_DEBUG("{} found {} indexers for partition {}", state.name,
              indexes->size(), state.id);
   auto type_ids = partition.type_ids();
   for (size_t i = 0; i < type_ids->size(); ++i) {
@@ -347,7 +351,7 @@ unpack(const fbs::partition::v0& partition, passive_partition_state& state) {
     if (auto error = fbs::deserialize_bytes(ids_data, ids))
       return error;
   }
-  VAST_DEBUG("{} restored {} type-to-ids mapping for partition {}", state.self,
+  VAST_DEBUG("{} restored {} type-to-ids mapping for partition {}", state.name,
              state.type_ids.size(), state.id);
   return caf::none;
 }
@@ -363,13 +367,14 @@ caf::error unpack(const fbs::partition::v0& x, partition_synopsis& ps) {
 active_partition_actor::behavior_type active_partition(
   active_partition_actor::stateful_pointer<active_partition_state> self,
   uuid id, filesystem_actor filesystem, caf::settings index_opts,
-  caf::settings synopsis_opts) {
+  caf::settings synopsis_opts, store_actor store) {
   self->state.self = self;
   self->state.name = "partition-" + to_string(id);
   self->state.id = id;
   self->state.offset = invalid_id;
   self->state.events = 0;
   self->state.filesystem = std::move(filesystem);
+  self->state.store = std::move(store);
   self->state.streaming_initiated = false;
   self->state.synopsis = std::make_shared<partition_synopsis>();
   self->state.synopsis_opts = std::move(synopsis_opts);
@@ -377,13 +382,14 @@ active_partition_actor::behavior_type active_partition(
   // The active partition stage is a caf stream stage that takes
   // a stream of `table_slice` as input and produces several
   // streams of `table_slice_column` as output.
-  self->state.stage = caf::attach_continuous_stream_stage(
-    self,
+  self->state.stage = detail::attach_notifying_stream_stage(
+    self, true,
     [=](caf::unit_t&) {
       // nop
     },
     [=](caf::unit_t&, caf::downstream<table_slice_column>& out, table_slice x) {
-      VAST_TRACE_SCOPE("{} {}", VAST_ARG(out), VAST_ARG(x));
+      VAST_TRACE_SCOPE("partition {} got table slice {} {}", self->state.id,
+                       VAST_ARG(out), VAST_ARG(x));
       // We rely on `invalid_id` actually being the highest possible id
       // when using `min()` below.
       static_assert(invalid_id == std::numeric_limits<vast::id>::max());
@@ -416,16 +422,14 @@ active_partition_actor::behavior_type active_partition(
       }
     },
     [=](caf::unit_t&, const caf::error& err) {
+      VAST_DEBUG("active partition {} finalized streaming {}", id, render(err));
       // We get an 'unreachable' error when the stream becomes unreachable
-      // because the actor was destroyed; in this case we can't use `self`
-      // anymore.
+      // because the actor was destroyed; in this case the state was already
+      // destroyed during `local_actor::on_exit()`.
       if (err && err != caf::exit_reason::unreachable) {
         VAST_ERROR("{} aborts with error: {}", self, render(err));
-        // We don't exit here, since there might be outstanding evaluators who
-        // still need our indexers.
         return;
       }
-      VAST_DEBUG("partition {} finalized streaming", id);
     },
     // Every "outbound path" has a path_state, which consists of a "Filter"
     // and a vector of "T", the output buffer. In the case of a partition,
@@ -439,19 +443,16 @@ active_partition_actor::behavior_type active_partition(
     // indexers, and compute the qualified record field name for each. A
     // specialized downstream manager could optimize this by using e.g. a map
     // from qualified record fields to downstream indexers.
-    caf::policy::arg<broadcast_downstream_manager<
+    caf::policy::arg<caf::broadcast_downstream_manager<
       table_slice_column, vast::qualified_record_field, partition_selector>>{});
   self->set_exit_handler([=](const caf::exit_msg& msg) {
     VAST_DEBUG("{} received EXIT from {} with reason: {}", self, msg.source,
                msg.reason);
-    if (self->state.stage->idle()
-        && !self->state.stage->out().paths().empty()) {
-      VAST_DEBUG("{} closes outbound path to {}/{} indexers", self,
-                 self->state.stage->out().paths().size(),
-                 self->state.stage->out().path_slots().size());
+    if (self->state.streaming_initiated
+        && self->state.stage->inbound_paths().empty()) {
       self->state.stage->out().fan_out_flush();
-      self->state.stage->out().force_emit_batches();
       self->state.stage->out().close();
+      self->state.stage->out().force_emit_batches();
     }
     // Delay shutdown if we're currently in the process of persisting.
     if (self->state.persistence_promise.pending()) {
@@ -486,7 +487,12 @@ active_partition_actor::behavior_type active_partition(
       self->state.streaming_initiated = true;
       return self->state.stage->add_inbound_path(in);
     },
-    [self](atom::persist, const path& part_dir, const path& synopsis_dir) {
+    [self](atom::subscribe, atom::flush, const flush_listener_actor& listener) {
+      self->state.add_flush_listener(listener);
+    },
+    [self](atom::persist, const std::filesystem::path& part_dir,
+           const std::filesystem::path& synopsis_dir) {
+      VAST_DEBUG("{} got persist atom", self);
       // Ensure that the response promise has not already been initialized.
       VAST_ASSERT(
         !static_cast<caf::response_promise&>(self->state.persistence_promise)
@@ -496,25 +502,23 @@ active_partition_actor::behavior_type active_partition(
       self->state.persisted_indexers = 0;
       self->state.persistence_promise
         = self->make_response_promise<std::shared_ptr<partition_synopsis>>();
-      // We use a high message priority here because we want to start persisting
-      // as soon as possible in order to avoid shutdown delay.
-      self->send<caf::message_priority::high>(self, atom::internal_v,
-                                              atom::persist_v, atom::resume_v);
+      self->send(self, atom::internal_v, atom::persist_v, atom::resume_v);
       return self->state.persistence_promise;
     },
     [self](atom::internal, atom::persist, atom::resume) {
-      // Wait for outstanding data to avoid data loss.
-      if (!self->state.streaming_initiated
-          || !self->state.stage->inbound_paths().empty()
-          || !self->state.stage->idle()) {
-        VAST_DEBUG("{} waits for stream before persisting", self);
+      VAST_TRACE("{} resumes persist atom {}", self,
+                 self->state.indexers.size());
+      if (self->state.streaming_initiated
+          && self->state.stage->inbound_paths().empty()) {
+        self->state.stage->out().fan_out_flush();
+        self->state.stage->out().close();
+        self->state.stage->out().force_emit_batches();
+      } else {
+        using namespace std::chrono_literals;
         self->delayed_send(self, 50ms, atom::internal_v, atom::persist_v,
                            atom::resume_v);
         return;
       }
-      self->state.stage->out().fan_out_flush();
-      self->state.stage->out().force_emit_batches();
-      self->state.stage->out().close();
       if (self->state.indexers.empty()) {
         self->state.persistence_promise.deliver(
           caf::make_error(ec::logic_error, "partition has no indexers"));
@@ -622,15 +626,29 @@ active_partition_actor::behavior_type active_partition(
             });
       }
     },
-    [self](const expression& expr,
-           partition_client_actor client) -> caf::result<atom::done> {
+    [self](vast::query query) -> caf::result<atom::done> {
       // TODO: We should do a candidate check using `self->state.synopsis` and
       // return early if that doesn't yield any results.
-      auto triples = evaluate(self->state, expr);
+      auto triples = evaluate(self->state, query.expr);
       if (triples.empty())
         return atom::done_v;
-      auto eval = self->spawn(evaluator, expr, triples);
-      return self->delegate(eval, client);
+      auto eval = self->spawn(evaluator, query.expr, triples);
+      auto rp = self->make_response_promise<atom::done>();
+      self->request(eval, caf::infinite, atom::run_v)
+        .then(
+          [self, rp, query = std::move(query)](const ids& hits) mutable {
+            // TODO: Use the first path if the expression can be evaluated
+            // exactly.
+            auto* count = caf::get_if<query::count>(&query.cmd);
+            if (count && count->mode == query::count::estimate) {
+              self->send(count->sink, rank(hits));
+              rp.deliver(atom::done_v);
+            } else {
+              rp.delegate(self->state.store, std::move(query), hits);
+            }
+          },
+          [rp](caf::error& err) mutable { rp.deliver(std::move(err)); });
+      return rp;
     },
     [self](atom::status,
            status_verbosity v) -> caf::typed_response_promise<caf::settings> {
@@ -687,8 +705,10 @@ active_partition_actor::behavior_type active_partition(
 
 partition_actor::behavior_type passive_partition(
   partition_actor::stateful_pointer<passive_partition_state> self, uuid id,
-  filesystem_actor filesystem, class path path) {
+  filesystem_actor filesystem, const std::filesystem::path& path,
+  store_actor store) {
   self->state.self = self;
+  self->state.store = std::move(store);
   self->set_exit_handler([=](const caf::exit_msg& msg) {
     VAST_DEBUG("{} received EXIT from {} with reason: {}", self, msg.source,
                msg.reason);
@@ -769,15 +789,14 @@ partition_actor::behavior_type passive_partition(
         // Delegate all deferred evaluations now that we have the partition chunk.
         VAST_DEBUG("{} delegates {} deferred evaluations", self,
                    self->state.deferred_evaluations.size());
-        for (auto&& [expr, client, rp] :
+        for (auto&& [expr, rp] :
              std::exchange(self->state.deferred_evaluations, {}))
-          rp.delegate(static_cast<partition_actor>(self), std::move(expr),
-                      client);
+          rp.delegate(static_cast<partition_actor>(self), std::move(expr));
       },
       [=](caf::error err) {
         VAST_ERROR("{} failed to load partition: {}", self, render(err));
         // Deliver the error for all deferred evaluations.
-        for (auto&& [expr, client, rp] :
+        for (auto&& [expr, rp] :
              std::exchange(self->state.deferred_evaluations, {})) {
           // Because of a deficiency in the typed_response_promise API, we must
           // access the underlying response_promise to deliver the error.
@@ -788,20 +807,40 @@ partition_actor::behavior_type passive_partition(
         self->quit(std::move(err));
       });
   return {
-    [self](const expression& expr,
-           partition_client_actor client) -> caf::result<atom::done> {
-      VAST_TRACE_SCOPE("{} {}", self, VAST_ARG(expr));
+    [self](vast::query query) -> caf::result<atom::done> {
+      VAST_TRACE_SCOPE("{} {}", self, VAST_ARG(query));
       if (!self->state.partition_chunk)
-        return get<2>(self->state.deferred_evaluations.emplace_back(
-          expr, client, self->make_response_promise<atom::done>()));
+        return std::get<1>(self->state.deferred_evaluations.emplace_back(
+          std::move(query), self->make_response_promise<atom::done>()));
       // We can safely assert that if we have the partition chunk already, all
       // deferred evaluations were taken care of.
       VAST_ASSERT(self->state.deferred_evaluations.empty());
-      auto triples = evaluate(self->state, expr);
+      // Don't handle queries after we already received an exit message, while
+      // the terminator is running. Since we require every partition to have at
+      // least one indexer, we can use this to check.
+      if (self->state.indexers.empty())
+        return caf::make_error(ec::system_error, "can not handle query because "
+                                                 "shutdown was requested");
+      auto triples = evaluate(self->state, query.expr);
       if (triples.empty())
         return atom::done_v;
-      auto eval = self->spawn(evaluator, expr, triples);
-      return self->delegate(eval, client);
+      auto eval = self->spawn(evaluator, query.expr, triples);
+      auto rp = self->make_response_promise<atom::done>();
+      self->request(eval, caf::infinite, atom::run_v)
+        .then(
+          [self, rp, query = std::move(query)](const ids& hits) mutable {
+            // TODO: Use the first path if the expression can be evaluated
+            // exactly.
+            auto* count = caf::get_if<query::count>(&query.cmd);
+            if (count && count->mode == query::count::estimate) {
+              self->send(count->sink, rank(hits));
+              rp.deliver(atom::done_v);
+            } else {
+              rp.delegate(self->state.store, std::move(query), hits);
+            }
+          },
+          [rp](caf::error& err) mutable { rp.deliver(std::move(err)); });
+      return rp;
     },
     [self](atom::status,
            status_verbosity /*v*/) -> caf::config_value::dictionary {
