@@ -1,15 +1,10 @@
-/******************************************************************************
- *                    _   _____   __________                                  *
- *                   | | / / _ | / __/_  __/     Visibility                   *
- *                   | |/ / __ |_\ \  / /          Across                     *
- *                   |___/_/ |_/___/ /_/       Space and Time                 *
- *                                                                            *
- * This file is part of VAST. It is subject to the license terms in the       *
- * LICENSE file found in the top-level directory of this distribution and at  *
- * http://vast.io/license. No part of VAST, including this file, may be       *
- * copied, modified, propagated, or distributed except according to the terms *
- * contained in the LICENSE file.                                             *
- ******************************************************************************/
+//    _   _____   __________
+//   | | / / _ | / __/_  __/     Visibility
+//   | |/ / __ |_\ \  / /          Across
+//   |___/_/ |_/___/ /_/       Space and Time
+//
+// SPDX-FileCopyrightText: (c) 2018 The VAST Contributors
+// SPDX-License-Identifier: BSD-3-Clause
 
 #include "vast/table_slice.hpp"
 
@@ -33,6 +28,7 @@
 
 #if VAST_ENABLE_ARROW
 #  include "vast/arrow_table_slice.hpp"
+#  include "vast/arrow_table_slice_builder.hpp"
 #endif // VAST_ENABLE_ARROW
 
 namespace vast {
@@ -103,7 +99,7 @@ const fbs::TableSlice* as_flatbuffer(const chunk_ptr& chunk) noexcept {
 chunk_ptr
 verified_or_none(chunk_ptr&& chunk, enum table_slice::verify verify) noexcept {
   if (verify == table_slice::verify::yes && chunk) {
-    const auto data = reinterpret_cast<const uint8_t*>(chunk->data());
+    const auto* const data = reinterpret_cast<const uint8_t*>(chunk->data());
     auto verifier = flatbuffers::Verifier{data, chunk->size()};
     if (!verifier.template VerifyBuffer<fbs::TableSlice>())
       chunk = {};
@@ -140,9 +136,7 @@ state([[maybe_unused]] Slice&& encoded, State&& state) noexcept {
 
 // -- constructors, destructors, and assignment operators ----------------------
 
-table_slice::table_slice() noexcept {
-  // nop
-}
+table_slice::table_slice() noexcept = default;
 
 table_slice::table_slice(chunk_ptr&& chunk, enum verify verify) noexcept
   : chunk_{verified_or_none(std::move(chunk), verify)} {
@@ -186,24 +180,32 @@ table_slice::table_slice(chunk_ptr&& chunk, enum verify verify,
 table_slice::table_slice(const fbs::FlatTableSlice& flat_slice,
                          const chunk_ptr& parent_chunk,
                          enum verify verify) noexcept {
-  const auto flat_slice_begin
+  const auto* const flat_slice_begin
     = reinterpret_cast<const std::byte*>(flat_slice.data()->data());
   const auto flat_slice_size = flat_slice.data()->size();
   VAST_ASSERT(flat_slice_begin >= parent_chunk->data());
-  VAST_ASSERT(flat_slice_begin + flat_slice_size
-              <= parent_chunk->data() + parent_chunk->size());
+  VAST_ASSERT(std::next(flat_slice_begin, flat_slice_size)
+              <= std::next(parent_chunk->data(), parent_chunk->size()));
   auto chunk = parent_chunk->slice(flat_slice_begin - parent_chunk->data(),
                                    flat_slice_size);
   // Delegate the sliced chunk to the constructor.
   *this = table_slice{std::move(chunk), verify};
 }
 
-table_slice::table_slice(const table_slice& other) noexcept
-  : chunk_{other.chunk_}, offset_{other.offset_}, state_{other.state_} {
-  // nop
+#if VAST_ENABLE_ARROW
+
+table_slice::table_slice(const std::shared_ptr<arrow::RecordBatch>& record_batch,
+                         const record_type& layout) {
+  *this = arrow_table_slice_builder::create(record_batch, layout);
 }
 
+#endif // VAST_ENABLE_ARROW
+
+table_slice::table_slice(const table_slice& other) noexcept = default;
+
 table_slice& table_slice::operator=(const table_slice& rhs) noexcept {
+  if (this == &rhs)
+    return *this;
   chunk_ = rhs.chunk_;
   offset_ = rhs.offset_;
   state_ = rhs.state_;
@@ -224,9 +226,7 @@ table_slice& table_slice::operator=(table_slice&& rhs) noexcept {
   return *this;
 }
 
-table_slice::~table_slice() noexcept {
-  // nop
-}
+table_slice::~table_slice() noexcept = default;
 
 // -- operators ----------------------------------------------------------------
 
@@ -344,6 +344,8 @@ data_view table_slice::at(table_slice::size_type row,
 
 data_view table_slice::at(table_slice::size_type row,
                           table_slice::size_type column, const type& t) const {
+  if (const auto* alias = caf::get_if<alias_type>(&t))
+    return at(row, column, alias->value_type);
   VAST_ASSERT(row < rows());
   VAST_ASSERT(column < columns());
   auto f = detail::overload{
@@ -411,24 +413,22 @@ rebuild(table_slice slice, enum table_slice_encoding encoding) noexcept {
     [&]() noexcept -> table_slice { return {}; },
     [&](const auto& encoded) noexcept -> table_slice {
       if (encoding == state(encoded, slice.state_)->encoding
-          && state(encoded, slice.state_)->is_latest_version) {
+          && state(encoded, slice.state_)->is_latest_version)
         return std::move(slice);
-      } else {
-        auto builder = factory<table_slice_builder>::make(builder_id(encoding),
-                                                          slice.layout());
-        if (!builder)
-          return table_slice{};
-        auto flat_layout = flatten(slice.layout());
-        for (table_slice::size_type row = 0; row < slice.rows(); ++row)
-          for (table_slice::size_type column = 0;
-               column < flat_layout.fields.size(); ++column)
-            if (!builder->add(
-                  slice.at(row, column, flat_layout.fields[column].type)))
-              return {};
-        auto result = builder->finish();
-        result.offset(slice.offset());
-        return result;
-      }
+      auto builder = factory<table_slice_builder>::make(builder_id(encoding),
+                                                        slice.layout());
+      if (!builder)
+        return table_slice{};
+      auto flat_layout = flatten(slice.layout());
+      for (table_slice::size_type row = 0; row < slice.rows(); ++row)
+        for (table_slice::size_type column = 0;
+             column < flat_layout.fields.size(); ++column)
+          if (!builder->add(
+                slice.at(row, column, flat_layout.fields[column].type)))
+            return {};
+      auto result = builder->finish();
+      result.offset(slice.offset());
+      return result;
     },
   };
   return visit(f, as_flatbuffer(slice.chunk_));
@@ -518,24 +518,34 @@ select(const table_slice& slice, const ids& selection) {
   return result;
 }
 
-table_slice truncate(const table_slice& slice, size_t num_rows) {
+table_slice truncate(table_slice slice, size_t num_rows) {
   VAST_ASSERT(slice.encoding() != table_slice_encoding::none);
   VAST_ASSERT(num_rows > 0);
   if (slice.rows() <= num_rows)
     return slice;
+  // make_ids fails if the offset is an invalid_id.
+  auto offset = slice.offset();
+  if (offset == invalid_id)
+    slice.offset(0u);
   auto selection = make_ids({{slice.offset(), slice.offset() + num_rows}});
   auto xs = select(slice, selection);
   VAST_ASSERT(xs.size() == 1);
+  if (offset == invalid_id)
+    xs.back().offset(invalid_id);
   return std::move(xs.back());
 }
 
 std::pair<table_slice, table_slice>
-split(const table_slice& slice, size_t partition_point) {
+split(table_slice slice, size_t partition_point) {
   VAST_ASSERT(slice.encoding() != table_slice_encoding::none);
   if (partition_point == 0)
     return {{}, slice};
   if (partition_point >= slice.rows())
     return {slice, {}};
+  // make_ids fails if the offset is an invalid_id.
+  auto offset = slice.offset();
+  if (offset == invalid_id)
+    slice.offset(0u);
   auto first = slice.offset();
   auto mid = first + partition_point;
   auto last = first + slice.rows();
@@ -545,12 +555,16 @@ split(const table_slice& slice, size_t partition_point) {
   // Create second table slice.
   select(xs, slice, make_ids({{mid, last}}));
   VAST_ASSERT(xs.size() == 2);
+  if (offset == invalid_id) {
+    xs.front().offset(invalid_id);
+    xs.back().offset(invalid_id);
+  }
   return {std::move(xs.front()), std::move(xs.back())};
 }
 
 uint64_t rows(const std::vector<table_slice>& slices) {
   auto result = uint64_t{0};
-  for (auto& slice : slices)
+  for (const auto& slice : slices)
     result += slice.rows();
   return result;
 }
@@ -578,17 +592,13 @@ struct row_evaluator {
   }
 
   bool operator()(const conjunction& c) {
-    for (auto& op : c)
-      if (!caf::visit(*this, op))
-        return false;
-    return true;
+    return std::all_of(c.begin(), c.end(),
+                       [&](const auto& op) { return caf::visit(*this, op); });
   }
 
   bool operator()(const disjunction& d) {
-    for (auto& op : d)
-      if (caf::visit(*this, op))
-        return true;
-    return false;
+    return std::any_of(d.begin(), d.end(),
+                       [&](const auto& op) { return caf::visit(*this, op); });
   }
 
   bool operator()(const negation& n) {
@@ -600,17 +610,17 @@ struct row_evaluator {
     return caf::visit(*this, p.lhs, p.rhs);
   }
 
-  bool operator()(const attribute_extractor& e, const data& d) {
+  bool operator()(const meta_extractor& e, const data& d) {
     // TODO: Transform this AST node into a constant-time lookup node (e.g.,
     // data_extractor). It's not necessary to iterate over the schema for
     // every row; this should happen upfront.
     auto&& layout = slice_.layout();
     // TODO: type and field queries don't produce false positives in the
     // partition. Is there actually any reason to do the check here?
-    if (e.attr == atom::type_v)
+    if (e.kind == meta_extractor::type)
       return evaluate(layout.name(), op_, d);
-    if (e.attr == atom::field_v) {
-      auto s = caf::get_if<std::string>(&d);
+    if (e.kind == meta_extractor::field) {
+      const auto* s = caf::get_if<std::string>(&d);
       if (!s) {
         VAST_WARN("#field can only compare with string");
         return false;
@@ -618,7 +628,7 @@ struct row_evaluator {
       auto result = false;
       auto neg = is_negated(op_);
       // auto abs_op = neg ? negate(op_) : op_;
-      for (auto& field : record_type::each{layout}) {
+      for (const auto& field : record_type::each{layout}) {
         auto fqn = layout.name() + "." + field.key();
         if (detail::ends_with(fqn, *s)) {
           result = true;
@@ -648,7 +658,7 @@ struct row_evaluator {
 
   const table_slice& slice_;
   size_t row_;
-  relational_operator op_;
+  relational_operator op_ = {};
 };
 
 } // namespace
@@ -664,4 +674,126 @@ ids evaluate(const expression& expr, const table_slice& slice) {
   return result;
 }
 
+std::optional<table_slice>
+filter(const table_slice& slice, const expression& expr, const ids& hints) {
+  VAST_ASSERT(slice.encoding() != table_slice_encoding::none);
+  const auto offset = slice.offset();
+  auto slice_ids = make_ids({{offset, offset + slice.rows()}});
+  auto selection = slice_ids;
+  if (!hints.empty())
+    selection &= hints;
+  // Do no rows qualify?
+  auto selection_rank = rank(selection);
+  if (selection_rank == 0)
+    return std::nullopt;
+  if (expr == expression{}) {
+    // Do all rows qualify?
+    if (rank(slice_ids) == selection_rank)
+      return slice;
+  }
+  // Get the desired encoding, and the already serialized layout.
+  auto f = detail::overload{
+    []() noexcept -> std::pair<table_slice_encoding, span<const std::byte>> {
+      die("cannot filter an invalid table slice");
+    },
+    [&](const auto& encoded) noexcept {
+      return std::pair{
+        builder_id(state(encoded, slice.state_)->encoding),
+        span{reinterpret_cast<const std::byte*>(encoded.layout()->data()),
+             encoded.layout()->size()}};
+    },
+  };
+  table_slice_encoding implementation_id;
+  span<const std::byte> serialized_layout = {};
+  std::tie(implementation_id, serialized_layout)
+    = visit(f, as_flatbuffer(slice.chunk_));
+  // Start slicing and dicing.
+  auto builder
+    = factory<table_slice_builder>::make(implementation_id, slice.layout());
+  VAST_ASSERT(builder);
+  auto flat_layout = flatten(slice.layout());
+  auto check = [&](row_evaluator eval) {
+    if (expr == expression{})
+      return true;
+    return caf::visit(eval, expr);
+  };
+  for (auto id : select(selection)) {
+    VAST_ASSERT(id >= offset);
+    auto row = id - offset;
+    VAST_ASSERT(row < slice.rows());
+    if (check(row_evaluator{slice, row})) {
+      for (size_t column = 0; column < flat_layout.fields.size(); ++column) {
+        auto cell_value
+          = slice.at(row, column, flat_layout.fields[column].type);
+        auto ret = builder->add(cell_value);
+        VAST_ASSERT(ret);
+      }
+    }
+  }
+  if (builder->rows() == 0)
+    return std::nullopt;
+  if (builder->rows() == slice.rows())
+    return slice;
+  auto new_slice = builder->finish(serialized_layout);
+  VAST_ASSERT(new_slice.encoding() != table_slice_encoding::none);
+  return new_slice;
+}
+
+std::optional<table_slice>
+filter(const table_slice& slice, const expression& expr) {
+  return filter(slice, expr, ids{});
+}
+
+std::optional<table_slice> filter(const table_slice& slice, const ids& hints) {
+  return filter(slice, expression{}, hints);
+}
+
+uint64_t count_matching(const table_slice& slice, const expression& expr,
+                        const ids& hints) {
+  VAST_ASSERT(slice.encoding() != table_slice_encoding::none);
+  const auto offset = slice.offset();
+  auto slice_ids = make_ids({{offset, offset + slice.rows()}});
+  auto selection = slice_ids;
+  if (!hints.empty())
+    selection &= hints;
+  // Do no rows qualify?
+  auto selection_rank = rank(selection);
+  if (selection_rank == 0)
+    return 0;
+  if (expr == expression{}) {
+    // Do all rows qualify?
+    if (rank(slice_ids) == selection_rank)
+      return slice.rows();
+  }
+  // Get the desired encoding, and the already serialized layout.
+  auto f = detail::overload{
+    []() noexcept -> std::pair<table_slice_encoding, span<const std::byte>> {
+      die("cannot filter an invalid table slice");
+    },
+    [&](const auto& encoded) noexcept {
+      return std::pair{
+        builder_id(state(encoded, slice.state_)->encoding),
+        span{reinterpret_cast<const std::byte*>(encoded.layout()->data()),
+             encoded.layout()->size()}};
+    },
+  };
+  table_slice_encoding implementation_id;
+  span<const std::byte> serialized_layout = {};
+  std::tie(implementation_id, serialized_layout)
+    = visit(f, as_flatbuffer(slice.chunk_));
+  auto check = [&](row_evaluator eval) -> uint64_t {
+    if (expr == expression{})
+      return 1u;
+    return static_cast<uint64_t>(caf::visit(eval, expr));
+  };
+  uint64_t cnt = 0u;
+  for (auto id : select(selection)) {
+    VAST_ASSERT(id >= offset);
+    auto row = id - offset;
+    VAST_ASSERT(row < slice.rows());
+    cnt += check(row_evaluator{slice, row});
+  }
+  return cnt;
+}
+
 } // namespace vast
