@@ -1,15 +1,10 @@
-/******************************************************************************
- *                    _   _____   __________                                  *
- *                   | | / / _ | / __/_  __/     Visibility                   *
- *                   | |/ / __ |_\ \  / /          Across                     *
- *                   |___/_/ |_/___/ /_/       Space and Time                 *
- *                                                                            *
- * This file is part of VAST. It is subject to the license terms in the       *
- * LICENSE file found in the top-level directory of this distribution and at  *
- * http://vast.io/license. No part of VAST, including this file, may be       *
- * copied, modified, propagated, or distributed except according to the terms *
- * contained in the LICENSE file.                                             *
- ******************************************************************************/
+//    _   _____   __________
+//   | | / / _ | / __/_  __/     Visibility
+//   | |/ / __ |_\ \  / /          Across
+//   |___/_/ |_/___/ /_/       Space and Time
+//
+// SPDX-FileCopyrightText: (c) 2020 The VAST Contributors
+// SPDX-License-Identifier: BSD-3-Clause
 
 #include "vast/taxonomies.hpp"
 
@@ -322,7 +317,7 @@ resolve_impl(const taxonomies& ts, const expression& e,
         auto make_meta_field_predicate
           = [&](relational_operator op, const vast::data&) {
               return [&, op](std::string item) {
-                return predicate{attribute_extractor{atom::field_v}, op,
+                return predicate{meta_extractor{meta_extractor::field}, op,
                                  vast::data{item}};
               };
             };
