@@ -1,15 +1,10 @@
-/******************************************************************************
- *                    _   _____   __________                                  *
- *                   | | / / _ | / __/_  __/     Visibility                   *
- *                   | |/ / __ |_\ \  / /          Across                     *
- *                   |___/_/ |_/___/ /_/       Space and Time                 *
- *                                                                            *
- * This file is part of VAST. It is subject to the license terms in the       *
- * LICENSE file found in the top-level directory of this distribution and at  *
- * http://vast.io/license. No part of VAST, including this file, may be       *
- * copied, modified, propagated, or distributed except according to the terms *
- * contained in the LICENSE file.                                             *
- ******************************************************************************/
+//    _   _____   __________
+//   | | / / _ | / __/_  __/     Visibility
+//   | |/ / __ |_\ \  / /          Across
+//   |___/_/ |_/___/ /_/       Space and Time
+//
+// SPDX-FileCopyrightText: (c) 2016 The VAST Contributors
+// SPDX-License-Identifier: BSD-3-Clause
 
 #include "vast/type.hpp"
 
@@ -24,6 +19,7 @@
 #include "vast/pattern.hpp"
 #include "vast/schema.hpp"
 
+#include <optional>
 #include <tuple>
 #include <typeindex>
 #include <utility>
@@ -258,10 +254,10 @@ size_t record_type::num_leaves() const {
   return count;
 }
 
-caf::optional<offset> record_type::resolve(std::string_view key) const {
+std::optional<offset> record_type::resolve(std::string_view key) const {
   offset result;
   if (key.empty())
-    return caf::none;
+    return {};
   auto offset = offset::size_type{0};
   for (auto i = fields.begin(); i != fields.end(); ++i, ++offset) {
     auto& field = *i;
@@ -278,33 +274,33 @@ caf::optional<offset> record_type::resolve(std::string_view key) const {
       auto remainder = key.substr(1 + name.size());
       auto rec = get_if<record_type>(&field.type);
       if (!rec)
-        return caf::none;
+        return {};
       auto sub_result = rec->resolve(remainder);
       if (!sub_result)
-        return caf::none;
+        return {};
       result.insert(result.end(), sub_result->begin(), sub_result->end());
       return result;
     }
   }
-  return caf::none;
+  return {};
 }
 
-caf::optional<std::string> record_type::resolve(const offset& o) const {
+std::optional<std::string> record_type::resolve(const offset& o) const {
   if (o.empty())
-    return caf::none;
+    return {};
   std::string result;
   auto r = this;
   for (size_t i = 0; i < o.size(); ++i) {
     auto x = o[i];
     if (x >= r->fields.size())
-      return caf::none;
+      return {};
     if (!result.empty())
       result += '.';
     result += r->fields[x].name;
     if (i != o.size() - 1) {
       r = get_if<record_type>(&r->fields[x].type);
       if (!r)
-        return caf::none;
+        return {};
     }
   }
   return result;
@@ -322,10 +318,11 @@ struct offset_map_builder {
   using result_type = std::vector<std::pair<offset, std::string>>;
 
   offset_map_builder(const record_type& r, result_type& result)
-    : r_{r}, result_{result} {
+    : r_{r}, result_{result}, trace_{r.name()} {
     run(r_);
   }
 
+private:
   void run(const record_type& r) {
     off_.push_back(0);
     auto prev_trace_size = trace_.size();
@@ -342,7 +339,7 @@ struct offset_map_builder {
 
   const record_type& r_;
   result_type& result_;
-  std::string trace_ = r_.name();
+  std::string trace_;
   offset off_;
 };
 
@@ -357,9 +354,9 @@ std::vector<std::pair<offset, std::string>> offset_map(const record_type& r) {
 std::vector<offset> record_type::find_suffix(std::string_view key) const {
   std::vector<offset> result;
   auto om = offset_map(*this);
-  auto rx_ = ".*" + pattern::glob(key) + "$";
+  auto rx_ = ".*\\." + pattern::glob(key) + "$";
   for (auto& [off, name] : om)
-    if (rx_.match(name))
+    if (key == name || rx_.match(name))
       result.emplace_back(off);
   return result;
 }
@@ -391,6 +388,34 @@ const record_field* record_type::at(const offset& o) const& {
   return nullptr;
 }
 
+// TODO: Make an inplace version of this function after type flatbuffers allow
+// us to modify the fields in place.
+caf::expected<record_type>
+record_type::assign(const offset& o, const record_field& field) const {
+  std::vector<std::string> names;
+  const auto* r = this;
+  for (size_t i = 0; i < o.size() - 1; ++i) {
+    const auto& idx = o[i];
+    if (idx >= r->fields.size())
+      return caf::make_error(ec::invalid_argument, "invalid offset");
+    const auto* f = &r->fields[idx];
+    names.push_back(f->name);
+    r = get_if<record_type>(&f->type);
+    if (!r)
+      return caf::make_error(ec::invalid_argument, "invalid offset");
+  }
+  vast::record_type update_type;
+  update_type.fields.push_back(field);
+  for (auto it = names.rbegin(); it != names.rend(); ++it) {
+    vast::record_type container;
+    container.fields.emplace_back(*it, std::move(update_type));
+    update_type = std::move(container);
+  }
+  vast::record_type result = *this;
+  priority_merge(result, update_type, merge_policy::prefer_right);
+  return result;
+}
+
 bool record_type::equals(const abstract_type& other) const {
   return super::equals(other) && fields == downcast(other).fields;
 }
@@ -399,7 +424,7 @@ bool record_type::less_than(const abstract_type& other) const {
   return super::less_than(other) || fields < downcast(other).fields;
 }
 
-caf::optional<record_field> record_type::flat_field_at(offset o) const {
+std::optional<record_field> record_type::flat_field_at(offset o) const {
   record_field result;
   auto r = this;
   size_t x = 0;
@@ -409,7 +434,7 @@ caf::optional<record_field> record_type::flat_field_at(offset o) const {
     if (auto next_record = get_if<record_type>(&next.type))
       r = next_record;
     else
-      return caf::none;
+      return {};
   }
   auto next = r->fields[o[x]];
   result.name += next.name + '.';
@@ -417,13 +442,13 @@ caf::optional<record_field> record_type::flat_field_at(offset o) const {
   return result;
 }
 
-caf::optional<size_t> record_type::flat_index_at(offset o) const {
+std::optional<size_t> record_type::flat_index_at(offset o) const {
   // Empty offsets are invalid.
   if (o.empty())
-    return caf::none;
+    return {};
   // Bounds check.
   if (o[0] >= fields.size())
-    return caf::none;
+    return {};
   // Example: o = [1] picks the second element. However, we still need the
   // total amount of nested elements of the first element.
   size_t flat_index = 0;
@@ -431,39 +456,156 @@ caf::optional<size_t> record_type::flat_index_at(offset o) const {
     flat_index += flat_size(fields[i].type);
   // Now, we know how many fields are on the left. We're done if the offset
   // points to a non-record field in this record.
-  auto record_field = get_if<record_type>(&fields[o[0]].type);
+  auto record_field = caf::get_if<record_type>(&fields[o[0]].type);
   if (o.size() == 1) {
     // Sanity check: the offset is invalid if it points to a record type.
     if (record_field != nullptr)
-      return caf::none;
+      return {};
     return flat_index;
   }
   // The offset points into the field, therefore it must be a record type.
   if (record_field == nullptr)
-    return caf::none;
+    return {};
   // Drop index of the first dimension and dispatch to field recursively.
   o.erase(o.begin());
   auto sub_result = record_field->flat_index_at(o);
   if (!sub_result)
-    return caf::none;
+    return {};
   return flat_index + *sub_result;
 }
 
-caf::optional<offset> record_type::offset_from_index(size_t i) const {
+std::optional<offset> record_type::offset_from_index(size_t i) const {
   auto e = each(*this);
   auto it = e.begin();
   auto end = e.end();
   for (size_t j = 0; j < i; ++j, ++it) {
     if (it == end)
-      return caf::none;
+      return {};
   }
   if (it == end)
-    return caf::none;
+    return {};
   return it->offset;
 }
 
+caf::expected<record_type>
+merge(const record_type& lhs, const record_type& rhs) {
+  record_type result = lhs;
+  auto in_lhs = [&](std::string_view name) {
+    return std::find_if(result.fields.begin(),
+                        result.fields.begin() + lhs.fields.size(),
+                        [&](const auto& field) { return field.name == name; });
+  };
+  for (const auto& rfield : rhs.fields) {
+    if (auto it = in_lhs(rfield.name);
+        it != result.fields.begin() + lhs.fields.size()) {
+      if (it->type == rfield.type)
+        continue;
+      const auto* lrec = caf::get_if<record_type>(&it->type);
+      const auto* rrec = caf::get_if<record_type>(&rfield.type);
+      if (!(rrec && lrec))
+        return caf::make_error(ec::convert_error, //
+                               fmt::format("failed to merge {} and {} because "
+                                           "of duplicate field {}",
+                                           lhs, rhs, rfield.name));
+      auto x = merge(*lrec, *rrec);
+      if (!x)
+        return x.error();
+      it->type = type{std::move(*x)};
+    } else {
+      result.fields.push_back(rfield);
+    }
+  }
+  return result.name("");
+}
+
+record_type
+priority_merge(const record_type& lhs, const record_type& rhs, merge_policy p) {
+  record_type result = lhs;
+  auto in_lhs = [&](std::string_view name) {
+    return std::find_if(result.fields.begin(),
+                        result.fields.begin() + lhs.fields.size(),
+                        [&](const auto& field) { return field.name == name; });
+  };
+  for (const auto& rfield : rhs.fields) {
+    if (auto it = in_lhs(rfield.name);
+        it != result.fields.begin() + lhs.fields.size()) {
+      if (it->type == rfield.type)
+        continue;
+      const auto* lrec = caf::get_if<record_type>(&it->type);
+      const auto* rrec = caf::get_if<record_type>(&rfield.type);
+      if (rrec && lrec)
+        it->type = priority_merge(*lrec, *rrec, p);
+      else if (p == merge_policy::prefer_right)
+        it->type = rfield.type;
+      // else policy_left: continue
+    } else {
+      result.fields.push_back(rfield);
+    }
+  }
+  if (p == merge_policy::prefer_left) {
+    result.attributes(rhs.attributes());
+    result.update_attributes(lhs.attributes());
+  } else {
+    result.update_attributes(rhs.attributes());
+  }
+  return result.name("");
+}
+
+std::optional<record_type>
+remove_field(const record_type& r, std::vector<std::string_view> path) {
+  VAST_ASSERT(!path.empty());
+  auto result = record_type{}.name(r.name()).attributes(r.attributes());
+  for (const auto& f : r.fields) {
+    if (f.name == path.front()) {
+      if (path.size() > 1) {
+        path.erase(path.begin());
+        const auto* field_rec = caf::get_if<record_type>(&f.type);
+        if (!field_rec)
+          return std::nullopt;
+        auto new_rec = remove_field(*field_rec, path);
+        if (!new_rec)
+          return std::nullopt;
+        // TODO: Remove this condition if empty records get allowed.
+        if (!new_rec->fields.empty())
+          result.fields.emplace_back(f.name, *new_rec);
+      }
+      // else skips this field. It is the leaf to remove!
+    } else {
+      result.fields.push_back(f);
+    }
+  }
+  return result;
+}
+
+std::optional<record_type> remove_field(const record_type& r, offset o) {
+  VAST_ASSERT(!o.empty());
+  auto result = record_type{}.name(r.name()).attributes(r.attributes());
+  if (o.front() >= r.fields.size())
+    return {};
+  const auto& field = r.fields[o.front()];
+  for (const auto& f : r.fields) {
+    if (&f == &field) {
+      if (o.size() > 1) {
+        o.erase(o.begin());
+        const auto* field_rec = caf::get_if<record_type>(&field.type);
+        if (!field_rec)
+          return {};
+        auto new_rec = remove_field(*field_rec, std::move(o));
+        if (!new_rec)
+          return {};
+        // TODO: Remove this condition if empty records get allowed.
+        if (!new_rec->fields.empty())
+          result.fields.emplace_back(f.name, *new_rec);
+      }
+    } else {
+      result.fields.push_back(f);
+    }
+  }
+  return result;
+}
+
 record_type flatten(const record_type& rec) {
-  /// Make a copy of the original to keep name and attributes.
+  // Make a copy of the original to keep name and attributes.
   record_type result = rec;
   result.fields.clear();
   for (auto& outer : rec.fields)
@@ -505,41 +647,6 @@ size_t flat_size(const type& t) {
   return 1;
 }
 
-record_type unflatten(const record_type& rec) {
-  record_type result;
-  for (auto& f : rec.fields) {
-    auto names = detail::split(f.name, ".");
-    VAST_ASSERT(!names.empty());
-    record_type* r = &result;
-    for (size_t i = 0; i < names.size() - 1; ++i) {
-      if (r->fields.empty() || r->fields.back().name != names[i])
-        r->fields.emplace_back(std::string(names[i]), record_type{});
-      r = const_cast<record_type*>(get_if<record_type>(&r->fields.back().type));
-    }
-    r->fields.emplace_back(std::string{names.back()}, f.type);
-  }
-  std::vector<std::vector<const record_type*>> rs(1);
-  rs.back().push_back(&result);
-  auto more = true;
-  while (more) {
-    std::vector<const record_type*> next;
-    for (auto& current : rs.back())
-      for (auto& f : current->fields)
-        if (auto r = get_if<record_type>(&f.type))
-          next.push_back(r);
-    if (next.empty())
-      more = false;
-    else
-      rs.push_back(std::move(next));
-  }
-  return result;
-}
-
-type unflatten(const type& t) {
-  auto r = get_if<record_type>(&t);
-  return r ? unflatten(*r) : t;
-}
-
 bool is_basic(const type& x) {
   return x && is<type_flags::basic>(x->flags());
 }
@@ -559,14 +666,9 @@ bool is_container(const type& x) {
 namespace {
 
 struct type_congruence_checker {
-  template <class T>
-  bool operator()(const T&, const T&) const {
-    return true;
-  }
-
-  template <class T, class U>
-  bool operator()(const T&, const U&) const {
-    return false;
+  template <class T, class U = T>
+  constexpr bool operator()(const T&, const U&) const {
+    return std::is_same_v<std::decay_t<T>, std::decay_t<U>>;
   }
 
   template <class T>
@@ -938,6 +1040,10 @@ std::string kind(const type& x) {
   return kind_tbl[x->index()];
 }
 
+bool has_skip_attribute(const type& t) {
+  return has_attribute(t, "skip");
+}
+
 bool convert(const type& t, data& d) {
   record o;
   o["name"] = t.name();
