@@ -1,15 +1,10 @@
-/******************************************************************************
- *                    _   _____   __________                                  *
- *                   | | / / _ | / __/_  __/     Visibility                   *
- *                   | |/ / __ |_\ \  / /          Across                     *
- *                   |___/_/ |_/___/ /_/       Space and Time                 *
- *                                                                            *
- * This file is part of VAST. It is subject to the license terms in the       *
- * LICENSE file found in the top-level directory of this distribution and at  *
- * http://vast.io/license. No part of VAST, including this file, may be       *
- * copied, modified, propagated, or distributed except according to the terms *
- * contained in the LICENSE file.                                             *
- ******************************************************************************/
+//    _   _____   __________
+//   | | / / _ | / __/_  __/     Visibility
+//   | |/ / __ |_\ \  / /          Across
+//   |___/_/ |_/___/ /_/       Space and Time
+//
+// SPDX-FileCopyrightText: (c) 2019 The VAST Contributors
+// SPDX-License-Identifier: BSD-3-Clause
 
 #include "vast/format/zeek.hpp"
 
@@ -22,15 +17,14 @@
 
 #define SUITE format
 
-#include "vast/test/fixtures/actor_system.hpp"
-#include "vast/test/fixtures/events.hpp"
-#include "vast/test/fixtures/filesystem.hpp"
-#include "vast/test/test.hpp"
-
 #include "vast/concept/parseable/to.hpp"
 #include "vast/concept/parseable/vast/schema.hpp"
 #include "vast/concept/parseable/vast/type.hpp"
 #include "vast/detail/fdinbuf.hpp"
+#include "vast/test/fixtures/actor_system.hpp"
+#include "vast/test/fixtures/events.hpp"
+#include "vast/test/fixtures/filesystem.hpp"
+#include "vast/test/test.hpp"
 
 using namespace vast;
 using namespace std::string_literals;
@@ -233,7 +227,7 @@ struct fixture : fixtures::deterministic_actor_system {
   }
 };
 
-} // namspace <anonymous>
+} // namespace
 
 FIXTURE_SCOPE(zeek_reader_tests, fixture)
 
@@ -258,7 +252,7 @@ TEST(zeek data parsing) {
   CHECK(zeek_parse(subnet_type{}, "10.0.0.0/24", d));
   CHECK(d == *to<subnet>("10.0.0.0/24"));
   CHECK(zeek_parse(list_type{integer_type{}}, "49329", d));
-  CHECK(d == list{49329});
+  CHECK(d == list{integer{49329}});
   CHECK(zeek_parse(list_type{string_type{}}, "49329,42", d));
   CHECK(d == list{"49329", "42"});
 }
@@ -398,7 +392,7 @@ FIXTURE_SCOPE(zeek_writer_tests, writer_fixture)
 TEST(zeek writer) {
   // Perform the writing.
   caf::settings options;
-  caf::put(options, "vast.export.write", directory.str());
+  caf::put(options, "vast.export.write", directory.string());
   format::zeek::writer writer{options};
   for (auto& slice : zeek_conn_log)
     if (auto err = writer.write(slice))
@@ -407,9 +401,13 @@ TEST(zeek writer) {
     if (auto err = writer.write(slice))
       FAIL("failed to write HTTP log");
   auto conn_layout = zeek_conn_log[0].layout();
-  CHECK(exists(directory / conn_layout.name() + ".log"));
+  const auto conn_layout_log
+    = std::filesystem::path{conn_layout.name() + ".log"};
+  CHECK(std::filesystem::exists(directory / conn_layout_log));
   auto http_layout = zeek_http_log[0].layout();
-  CHECK(exists(directory / http_layout.name() + ".log"));
+  const auto http_layout_log
+    = std::filesystem::path{http_layout.name() + ".log"};
+  CHECK(std::filesystem::exists(directory / http_layout_log));
   // TODO: these tests should verify content as well.
 }
 
