@@ -1,20 +1,13 @@
-/******************************************************************************
- *                    _   _____   __________                                  *
- *                   | | / / _ | / __/_  __/     Visibility                   *
- *                   | |/ / __ |_\ \  / /          Across                     *
- *                   |___/_/ |_/___/ /_/       Space and Time                 *
- *                                                                            *
- * This file is part of VAST. It is subject to the license terms in the       *
- * LICENSE file found in the top-level directory of this distribution and at  *
- * http://vast.io/license. No part of VAST, including this file, may be       *
- * copied, modified, propagated, or distributed except according to the terms *
- * contained in the LICENSE file.                                             *
- ******************************************************************************/
+//    _   _____   __________
+//   | | / / _ | / __/_  __/     Visibility
+//   | |/ / __ |_\ \  / /          Across
+//   |___/_/ |_/___/ /_/       Space and Time
+//
+// SPDX-FileCopyrightText: (c) 2016 The VAST Contributors
+// SPDX-License-Identifier: BSD-3-Clause
 
 #define SUITE parseable
 
-#include "vast/test/test.hpp"
-
 #include "vast/address.hpp"
 #include "vast/concept/parseable/core.hpp"
 #include "vast/concept/parseable/numeric.hpp"
@@ -22,11 +15,13 @@
 #include "vast/concept/parseable/string.hpp"
 #include "vast/concept/parseable/to.hpp"
 #include "vast/concept/parseable/vast/address.hpp"
+#include "vast/concept/parseable/vast/integer.hpp"
 #include "vast/concept/parseable/vast/offset.hpp"
 #include "vast/concept/parseable/vast/si.hpp"
 #include "vast/concept/parseable/vast/time.hpp"
 #include "vast/detail/narrow.hpp"
 #include "vast/si_literals.hpp"
+#include "vast/test/test.hpp"
 
 #include <caf/test/dsl.hpp>
 
@@ -82,9 +77,7 @@ TEST(choice - unused LHS) {
 TEST(choice triple) {
   using namespace parsers;
   auto fired = false;
-  auto p = chr{'x'}
-         | i32
-         | eps ->* [&] { fired = true; };
+  auto p = chr{'x'} | i32 | eps->*[&] { fired = true; };
   caf::variant<char, int32_t> x;
   CHECK(skip_to_eoi(p)("foobar", x));
   CHECK(fired);
@@ -114,7 +107,7 @@ TEST(maybe) {
   f = &c;
   result = '\0';
   CHECK(maybe_x(f, l, result));
-  CHECK(f == &c); // Iterator not advanced.
+  CHECK(f == &c);        // Iterator not advanced.
   CHECK(result == '\0'); // Result not modified.
 }
 
@@ -128,7 +121,7 @@ TEST(container attribute folding) {
 TEST(action) {
   using namespace parsers;
   auto make_v4 = [](uint32_t a) { return address::v4(&a); };
-  auto ipv4_addr = b32be ->* make_v4;
+  auto ipv4_addr = b32be->*make_v4;
   address x;
   CHECK(ipv4_addr("\x0A\x00\x00\x01", x));
   CHECK_EQUAL(x, unbox(to<address>("10.0.0.1")));
@@ -785,13 +778,18 @@ TEST(si count) {
   CHECK_EQUAL(to_count("8Gi"), 8_Gi);
   CHECK_EQUAL(to_count("9Ti"), 9_Ti);
   CHECK_EQUAL(to_count("10Ei"), 10_Ei);
+  MESSAGE("spaces before unit");
+  CHECK_EQUAL(to_count("1 Mi"), 1_Mi);
+  CHECK_EQUAL(to_count("1  Mi"), 1_Mi);
 }
 
 TEST(si int) {
   auto to_int = to_si<integer>;
-  auto as_int = [](auto x) { return detail::narrow_cast<integer>(x); };
+  auto as_int = [](auto x) {
+    return integer{detail::narrow_cast<integer::value_type>(x)};
+  };
   using namespace si_literals;
-  CHECK_EQUAL(to_int("-42"), -42);
+  CHECK_EQUAL(to_int("-42"), -as_int(42));
   CHECK_EQUAL(to_int("-1k"), -as_int(1_k));
   CHECK_EQUAL(to_int("-2M"), -as_int(2_M));
   CHECK_EQUAL(to_int("-3G"), -as_int(3_G));
