@@ -1,25 +1,18 @@
-/******************************************************************************
- *                    _   _____   __________                                  *
- *                   | | / / _ | / __/_  __/     Visibility                   *
- *                   | |/ / __ |_\ \  / /          Across                     *
- *                   |___/_/ |_/___/ /_/       Space and Time                 *
- *                                                                            *
- * This file is part of VAST. It is subject to the license terms in the       *
- * LICENSE file found in the top-level directory of this distribution and at  *
- * http://vast.io/license. No part of VAST, including this file, may be       *
- * copied, modified, propagated, or distributed except according to the terms *
- * contained in the LICENSE file.                                             *
- ******************************************************************************/
+//    _   _____   __________
+//   | | / / _ | / __/_  __/     Visibility
+//   | |/ / __ |_\ \  / /          Across
+//   |___/_/ |_/___/ /_/       Space and Time
+//
+// SPDX-FileCopyrightText: (c) 2016 The VAST Contributors
+// SPDX-License-Identifier: BSD-3-Clause
 
 #define SUITE type
 
 #include "vast/type.hpp"
 
-#include "vast/test/fixtures/actor_system.hpp"
-#include "vast/test/test.hpp"
-
 #include "vast/concept/hashable/uhash.hpp"
 #include "vast/concept/hashable/xxhash.hpp"
+#include "vast/concept/parseable/to.hpp"
 #include "vast/concept/parseable/vast/type.hpp"
 #include "vast/concept/printable/stream.hpp"
 #include "vast/concept/printable/to_string.hpp"
@@ -27,6 +20,8 @@
 #include "vast/concept/printable/vast/offset.hpp"
 #include "vast/concept/printable/vast/type.hpp"
 #include "vast/data.hpp"
+#include "vast/test/fixtures/actor_system.hpp"
+#include "vast/test/test.hpp"
 
 #include <string_view>
 
@@ -270,7 +265,7 @@ TEST(record resolving) {
   CHECK_EQUAL(*k, "c.x");
 }
 
-TEST(record flattening / unflattening) {
+TEST(record flattening) {
   // clang-format off
   auto x = record_type{
     {"x", record_type{
@@ -299,8 +294,6 @@ TEST(record flattening / unflattening) {
   // clang-format on
   auto f = flatten(x);
   CHECK(f == y);
-  auto u = unflatten(f);
-  CHECK(u == x);
 }
 
 TEST(record flat index computation) {
@@ -324,7 +317,7 @@ TEST(record flat index computation) {
     }}
   };
   // clang-format on
-  using os = caf::optional<size_t>;
+  using os = std::optional<size_t>;
   static const os invalid;
   CHECK_EQUAL(flat_size(x), 6u);
   CHECK_EQUAL(x.flat_index_at(offset({0, 0, 0})), os(0u));
@@ -406,7 +399,20 @@ TEST(record symbol finding - suffix) {
   CHECK_EQUAL(f.find_suffix("b"), (offset_keys{{2}}));
   MESSAGE("record name is part of query");
   CHECK_EQUAL(r.find_suffix("foo.a"), (offset_keys{{0}}));
-  CHECK_EQUAL(f.find_suffix("oo.b.c.y"), (offset_keys{{4}}));
+  CHECK_EQUAL(f.find_suffix("oo.b.c.y"), std::vector<offset>{});
+}
+
+TEST(different fields with same suffix) {
+  auto r = record_type{{"zeek.client", string_type{}},
+                       {"suricata.alert.flow.bytes_toclient", count_type{}}};
+  auto suffixes = r.find_suffix("client");
+  CHECK_EQUAL(suffixes.size(), 1ull);
+}
+
+TEST(same fields with different type) {
+  auto r = record_type{{"client", string_type{}}, {"client", count_type{}}};
+  auto suffixes = r.find_suffix("client");
+  CHECK_EQUAL(suffixes.size(), 2ull);
 }
 
 TEST(congruence) {
@@ -446,6 +452,10 @@ TEST(congruence) {
   a = a.name("r0");
   CHECK(type{a} != type{r0});
   CHECK(congruent(a, r0));
+  MESSAGE("unspecified type");
+  CHECK(congruent(type{}, type{}));
+  CHECK(!congruent(type{string_type{}}, type{}));
+  CHECK(!congruent(type{}, type{string_type{}}));
 }
 
 TEST(subset) {
@@ -492,7 +502,7 @@ TEST(type_check) {
   MESSAGE("basic types");
   TYPE_CHECK(none_type{}, caf::none);
   TYPE_CHECK(bool_type{}, false);
-  TYPE_CHECK(integer_type{}, 42);
+  TYPE_CHECK(integer_type{}, integer{42});
   TYPE_CHECK(count_type{}, 42u);
   TYPE_CHECK(real_type{}, 4.2);
   TYPE_CHECK(duration_type{}, duration{0});
@@ -505,11 +515,12 @@ TEST(type_check) {
   TYPE_CHECK(enumeration_type{{"foo"}}, enumeration{0});
   TYPE_CHECK_FAIL(enumeration_type{{"foo"}}, enumeration{1});
   MESSAGE("containers");
-  TYPE_CHECK(list_type{integer_type{}}, list({1, 2, 3}));
-  TYPE_CHECK(list_type{}, list({1, 2, 3}));
+  TYPE_CHECK(list_type{integer_type{}},
+             list({integer{1}, integer{2}, integer{3}}));
+  TYPE_CHECK(list_type{}, list({integer{1}, integer{2}, integer{3}}));
   TYPE_CHECK(list_type{}, list{});
   TYPE_CHECK(list_type{string_type{}}, list{});
-  auto xs = map{{1, true}, {2, false}};
+  auto xs = map{{integer{1}, true}, {integer{2}, false}};
   TYPE_CHECK((map_type{integer_type{}, bool_type{}}), xs);
   TYPE_CHECK(map_type{}, xs);
   TYPE_CHECK(map_type{}, map{});
@@ -520,7 +531,7 @@ TEST(type_check - nested record) {
   data x = record{
     {"x", "foo"},
     {"r", record{
-      {"i", -42},
+      {"i", integer{-42}},
       {"r", record{
         {"u", 1001u}
       }},
@@ -640,6 +651,26 @@ TEST(parseable) {
   };
   // clang-format on
   CHECK_EQUAL(t, r);
+  MESSAGE("record algebra");
+  // clang-format off
+  r = record_type{
+    {"", none_type{}.name("foo")},
+    {"+", none_type{}.name("bar")}
+  }.attributes({{"$algebra"}});
+  // clang-format on
+  CHECK_EQUAL(unbox(to<type>("foo+bar")), r);
+  CHECK_EQUAL(unbox(to<type>("foo + bar")), r);
+  r.fields[1] = record_field{"-", record_type{{"bar", bool_type{}}}};
+  CHECK_EQUAL(unbox(to<type>("foo-bar")), r);
+  CHECK_EQUAL(unbox(to<type>("foo - bar")), r);
+  str = "record{a: real} + bar"sv;
+  // clang-format off
+  r = record_type{
+    {"", record_type{{"a", real_type{}}}},
+    {"+", none_type{}.name("bar")}
+  }.attributes({{"$algebra"}});
+  // clang-format on
+  CHECK_EQUAL(unbox(to<type>(str)), r);
 }
 
 TEST(hashable) {
