@@ -1,32 +1,27 @@
-/******************************************************************************
- *                    _   _____   __________                                  *
- *                   | | / / _ | / __/_  __/     Visibility                   *
- *                   | |/ / __ |_\ \  / /          Across                     *
- *                   |___/_/ |_/___/ /_/       Space and Time                 *
- *                                                                            *
- * This file is part of VAST. It is subject to the license terms in the       *
- * LICENSE file found in the top-level directory of this distribution and at  *
- * http://vast.io/license. No part of VAST, including this file, may be       *
- * copied, modified, propagated, or distributed except according to the terms *
- * contained in the LICENSE file.                                             *
- ******************************************************************************/
+//    _   _____   __________
+//   | | / / _ | / __/_  __/     Visibility
+//   | |/ / __ |_\ \  / /          Across
+//   |___/_/ |_/___/ /_/       Space and Time
+//
+// SPDX-FileCopyrightText: (c) 2016 The VAST Contributors
+// SPDX-License-Identifier: BSD-3-Clause
 
 #pragma once
 
+#include "vast/aliases.hpp"
+#include "vast/bits.hpp"
+#include "vast/detail/assert.hpp"
+#include "vast/detail/range.hpp"
+#include "vast/detail/type_traits.hpp"
+#include "vast/optional.hpp"
+
+#include <caf/error.hpp>
+
 #include <algorithm>
 #include <iterator>
 #include <queue>
 #include <type_traits>
 
-#include <caf/error.hpp>
-
-#include "vast/aliases.hpp"
-#include "vast/bits.hpp"
-#include "vast/optional.hpp"
-#include "vast/detail/assert.hpp"
-#include "vast/detail/range.hpp"
-#include "vast/detail/type_traits.hpp"
-
 namespace vast {
 
 class bitmap;
@@ -312,21 +307,21 @@ public:
 
   /// @returns The current bit sequence.
   /// @pre `!done()`
-  const bits_type& bits() const {
+  [[nodiscard]] const bits_type& bits() const {
     VAST_ASSERT(!done());
     return rng_.get();
   }
 
   /// @returns The current position in the range.
   /// @pre `!done()`
-  id offset() const {
+  [[nodiscard]] id offset() const {
     VAST_ASSERT(!done());
     return n_ + i_;
   }
 
   /// @returns The bit value at the current position.
   /// @pre `!done()`
-  bool value() const {
+  [[nodiscard]] bool value() const {
     VAST_ASSERT(!done());
     return bits()[i_];
   }
@@ -335,13 +330,13 @@ public:
 
   /// Retrieves the current position in the range.
   /// @pre `!done()`
-  size_type get() const {
+  [[nodiscard]] size_type get() const {
     VAST_ASSERT(!done());
     return offset();
   }
 
   /// @returns `true` if the range is done.
-  bool done() const {
+  [[nodiscard]] bool done() const {
     return rng_.done() && i_ == npos;
   }
 
@@ -466,11 +461,11 @@ public:
       next();
   }
 
-  auto get() const {
+  [[nodiscard]] auto get() const {
     return rng_.get();
   }
 
-  bool done() const {
+  [[nodiscard]] bool done() const {
     return rng_.done();
   }
 
