@@ -1,31 +1,26 @@
-/******************************************************************************
- *                    _   _____   __________                                  *
- *                   | | / / _ | / __/_  __/     Visibility                   *
- *                   | |/ / __ |_\ \  / /          Across                     *
- *                   |___/_/ |_/___/ /_/       Space and Time                 *
- *                                                                            *
- * This file is part of VAST. It is subject to the license terms in the       *
- * LICENSE file found in the top-level directory of this distribution and at  *
- * http://vast.io/license. No part of VAST, including this file, may be       *
- * copied, modified, propagated, or distributed except according to the terms *
- * contained in the LICENSE file.                                             *
- ******************************************************************************/
+//    _   _____   __________
+//   | | / / _ | / __/_  __/     Visibility
+//   | |/ / __ |_\ \  / /          Across
+//   |___/_/ |_/___/ /_/       Space and Time
+//
+// SPDX-FileCopyrightText: (c) 2016 The VAST Contributors
+// SPDX-License-Identifier: BSD-3-Clause
 
 #pragma once
 
-#include <algorithm>
-#include <array>
-#include <limits>
-#include <vector>
-#include <type_traits>
+#include "vast/base.hpp"
+#include "vast/detail/assert.hpp"
+#include "vast/detail/operators.hpp"
+#include "vast/operator.hpp"
 
 #include <caf/meta/load_callback.hpp>
 #include <caf/meta/save_callback.hpp>
 
-#include "vast/base.hpp"
-#include "vast/operator.hpp"
-#include "vast/detail/assert.hpp"
-#include "vast/detail/operators.hpp"
+#include <algorithm>
+#include <array>
+#include <limits>
+#include <type_traits>
+#include <vector>
 
 namespace vast {
 
@@ -83,7 +78,7 @@ struct coder {
 
   /// Retrieves the amout of memory that is occupied by the coder.
   /// @returns The size of the coder measured in heap bytes used.
-  size_t memusage() const;
+  [[nodiscard]] size_t memusage() const;
 
   /// Retrieves the coder-specific bitmap storage.
   auto& storage() const;
@@ -97,7 +92,7 @@ public:
   using size_type = typename Bitmap::size_type;
   using value_type = bool;
 
-  size_t bitmap_count() const noexcept {
+  [[nodiscard]] size_t bitmap_count() const noexcept {
     return 1;
   }
 
@@ -106,7 +101,7 @@ public:
     return bitmap_;
   }
 
-  const bitmap_type& bitmap_at(size_t index) const {
+  [[nodiscard]] const bitmap_type& bitmap_at(size_t index) const {
     VAST_ASSERT(index == 0);
     return bitmap_;
   }
@@ -116,7 +111,7 @@ public:
     bitmap_.append_bits(x, n);
   }
 
-  Bitmap decode(relational_operator op, value_type x) const {
+  [[nodiscard]] Bitmap decode(relational_operator op, value_type x) const {
     VAST_ASSERT(op == relational_operator::equal
                 || op == relational_operator::not_equal);
     auto result = bitmap_;
@@ -135,15 +130,15 @@ public:
     bitmap_.append(other.bitmap_);
   }
 
-  size_type size() const {
+  [[nodiscard]] size_type size() const {
     return bitmap_.size();
   }
 
-  size_t memusage() const {
+  [[nodiscard]] size_t memusage() const {
     return bitmap_.memusage();
   }
 
-  const Bitmap& storage() const {
+  [[nodiscard]] const Bitmap& storage() const {
     return bitmap_;
   }
 
