@@ -1,15 +1,10 @@
-/******************************************************************************
- *                    _   _____   __________                                  *
- *                   | | / / _ | / __/_  __/     Visibility                   *
- *                   | |/ / __ |_\ \  / /          Across                     *
- *                   |___/_/ |_/___/ /_/       Space and Time                 *
- *                                                                            *
- * This file is part of VAST. It is subject to the license terms in the       *
- * LICENSE file found in the top-level directory of this distribution and at  *
- * http://vast.io/license. No part of VAST, including this file, may be       *
- * copied, modified, propagated, or distributed except according to the terms *
- * contained in the LICENSE file.                                             *
- ******************************************************************************/
+//    _   _____   __________
+//   | | / / _ | / __/_  __/     Visibility
+//   | |/ / __ |_\ \  / /          Across
+//   |___/_/ |_/___/ /_/       Space and Time
+//
+// SPDX-FileCopyrightText: (c) 2016 The VAST Contributors
+// SPDX-License-Identifier: BSD-3-Clause
 
 #pragma once
 
@@ -27,7 +22,7 @@ class line_range : range_facade<line_range> {
 public:
   explicit line_range(std::istream& input);
 
-  const std::string& get() const;
+  [[nodiscard]] const std::string& get() const;
 
   void next_impl();
   void next();
@@ -43,11 +38,11 @@ public:
       std::move(timeout)));
   }
 
-  bool done() const;
+  [[nodiscard]] bool done() const;
 
   std::string& line();
 
-  size_t line_number() const;
+  [[nodiscard]] size_t line_number() const;
 
 private:
   std::istream& input_;
@@ -57,4 +52,3 @@ private:
 };
 
 } // namespace vast::detail
-
