@@ -1,28 +1,23 @@
-/******************************************************************************
- *                    _   _____   __________                                  *
- *                   | | / / _ | / __/_  __/     Visibility                   *
- *                   | |/ / __ |_\ \  / /          Across                     *
- *                   |___/_/ |_/___/ /_/       Space and Time                 *
- *                                                                            *
- * This file is part of VAST. It is subject to the license terms in the       *
- * LICENSE file found in the top-level directory of this distribution and at  *
- * http://vast.io/license. No part of VAST, including this file, may be       *
- * copied, modified, propagated, or distributed except according to the terms *
- * contained in the LICENSE file.                                             *
- ******************************************************************************/
+//    _   _____   __________
+//   | | / / _ | / __/_  __/     Visibility
+//   | |/ / __ |_\ \  / /          Across
+//   |___/_/ |_/___/ /_/       Space and Time
+//
+// SPDX-FileCopyrightText: (c) 2019 The VAST Contributors
+// SPDX-License-Identifier: BSD-3-Clause
 
 #pragma once
 
+#include "vast/fwd.hpp"
+
 #include "vast/defaults.hpp"
 #include "vast/format/writer.hpp"
-#include "vast/fwd.hpp"
 #include "vast/type.hpp"
 
-#include <caf/error.hpp>
-#include <caf/expected.hpp>
-
 #include <arrow/io/api.h>
 #include <arrow/ipc/writer.h>
+#include <caf/error.hpp>
+#include <caf/expected.hpp>
 
 #include <memory>
 #include <vector>
@@ -51,7 +46,7 @@ public:
     out_ = std::move(ptr);
   }
 
-  bool layout(const record_type& t);
+  bool layout(const record_type& layout);
 
 private:
   output_stream_ptr out_;
