@@ -1,22 +1,18 @@
-/******************************************************************************
- *                    _   _____   __________                                  *
- *                   | | / / _ | / __/_  __/     Visibility                   *
- *                   | |/ / __ |_\ \  / /          Across                     *
- *                   |___/_/ |_/___/ /_/       Space and Time                 *
- *                                                                            *
- * This file is part of VAST. It is subject to the license terms in the       *
- * LICENSE file found in the top-level directory of this distribution and at  *
- * http://vast.io/license. No part of VAST, including this file, may be       *
- * copied, modified, propagated, or distributed except according to the terms *
- * contained in the LICENSE file.                                             *
- ******************************************************************************/
+//    _   _____   __________
+//   | | / / _ | / __/_  __/     Visibility
+//   | |/ / __ |_\ \  / /          Across
+//   |___/_/ |_/___/ /_/       Space and Time
+//
+// SPDX-FileCopyrightText: (c) 2019 The VAST Contributors
+// SPDX-License-Identifier: BSD-3-Clause
 
 #pragma once
 
-#include <cstdint>
-
 #include <caf/fwd.hpp>
 
+#include <cstdint>
+#include <optional>
+
 namespace vast {
 
 /// ICMP message type.
@@ -55,12 +51,12 @@ enum class icmp6_type : uint8_t {
 /// @param x The ICMP type.
 /// @returns For a request type, the response type - and vice versa.
 /// @relates icmp_type
-caf::optional<icmp_type> dual(icmp_type x);
+std::optional<icmp_type> dual(icmp_type x);
 
 /// Computes the dual to a given ICMP6 type.
 /// @param x The ICMP6 type.
 /// @returns For a request type, the response type - and vice versa.
 /// @relates icmp6_type
-caf::optional<icmp6_type> dual(icmp6_type x);
+std::optional<icmp6_type> dual(icmp6_type x);
 
 } // namespace vast
