@@ -1,15 +1,10 @@
-/******************************************************************************
- *                    _   _____   __________                                  *
- *                   | | / / _ | / __/_  __/     Visibility                   *
- *                   | |/ / __ |_\ \  / /          Across                     *
- *                   |___/_/ |_/___/ /_/       Space and Time                 *
- *                                                                            *
- * This file is part of VAST. It is subject to the license terms in the       *
- * LICENSE file found in the top-level directory of this distribution and at  *
- * http://vast.io/license. No part of VAST, including this file, may be       *
- * copied, modified, propagated, or distributed except according to the terms *
- * contained in the LICENSE file.                                             *
- ******************************************************************************/
+//    _   _____   __________
+//   | | / / _ | / __/_  __/     Visibility
+//   | |/ / __ |_\ \  / /          Across
+//   |___/_/ |_/___/ /_/       Space and Time
+//
+// SPDX-FileCopyrightText: (c) 2020 The VAST Contributors
+// SPDX-License-Identifier: BSD-3-Clause
 
 #pragma once
 
@@ -24,7 +19,6 @@
 #include "vast/time.hpp"
 
 #include <caf/none.hpp>
-#include <caf/optional.hpp>
 #include <caf/variant.hpp>
 
 #include <array>
@@ -32,6 +26,7 @@
 #include <cstddef>
 #include <cstdint>
 #include <cstring>
+#include <optional>
 
 /// The [MessagePack](https://github.com/msgpack/msgpack/blob/master/spec.md)
 /// object serialization specification.
@@ -328,11 +323,11 @@ public:
     // nop
   }
 
-  msgpack::format format() const {
+  [[nodiscard]] msgpack::format format() const {
     return format_;
   }
 
-  auto data() const {
+  [[nodiscard]] auto data() const {
     return data_;
   }
 
@@ -360,17 +355,17 @@ public:
   }
 
   /// @returns The container format.
-  msgpack::format format() const {
+  [[nodiscard]] msgpack::format format() const {
     return format_;
   }
 
   /// @returns The number of elements in the array.
-  size_t size() const {
+  [[nodiscard]] size_t size() const {
     return size_;
   }
 
   /// @returns A pointer to the beginning of the array data.
-  overlay data() const;
+  [[nodiscard]] overlay data() const;
 
   template <class Inspector>
   friend auto inspect(Inspector& f, array_view& x) {
@@ -395,15 +390,15 @@ public:
                 || fmt == ext16 || fmt == ext32);
   }
 
-  msgpack::format format() const {
+  [[nodiscard]] msgpack::format format() const {
     return format_;
   }
 
-  int8_t type() const {
+  [[nodiscard]] int8_t type() const {
     return type_;
   }
 
-  auto data() const {
+  [[nodiscard]] auto data() const {
     return data_;
   }
 
@@ -602,10 +597,10 @@ decltype(auto) visit(Visitor&& f, const object& x) {
 
 /// @relates object
 template <class T>
-caf::optional<T> get(const object& o) {
+std::optional<T> get(const object& o) {
   return visit(vast::detail::overload{
-                 [](auto&&) { return caf::optional<T>{}; },
-                 [](T x) { return caf::optional<T>{std::move(x)}; },
+                 [](auto&&) { return std::optional<T>{}; },
+                 [](T x) { return std::optional<T>{std::move(x)}; },
                },
                o);
 }
@@ -622,7 +617,7 @@ public:
   /// @pre The underlying buffer must represent a sequence of at least one
   /// well-formed msgpack object.
   /// @returns The object at the current position.
-  object get() const;
+  [[nodiscard]] object get() const;
 
   /// Advances to the next object.
   /// @returns The number of bytes skipped or 0 on failure.
@@ -642,7 +637,7 @@ public:
   }
 
 private:
-  const std::byte* at(size_t i) const {
+  [[nodiscard]] const std::byte* at(size_t i) const {
     return buffer_.data() + position_ + i;
   }
 
