@@ -1,29 +1,52 @@
-/******************************************************************************
- *                    _   _____   __________                                  *
- *                   | | / / _ | / __/_  __/     Visibility                   *
- *                   | |/ / __ |_\ \  / /          Across                     *
- *                   |___/_/ |_/___/ /_/       Space and Time                 *
- *                                                                            *
- * This file is part of VAST. It is subject to the license terms in the       *
- * LICENSE file found in the top-level directory of this distribution and at  *
- * http://vast.io/license. No part of VAST, including this file, may be       *
- * copied, modified, propagated, or distributed except according to the terms *
- * contained in the LICENSE file.                                             *
- ******************************************************************************/
+//    _   _____   __________
+//   | | / / _ | / __/_  __/     Visibility
+//   | |/ / __ |_\ \  / /          Across
+//   |___/_/ |_/___/ /_/       Space and Time
+//
+// SPDX-FileCopyrightText: (c) 2021 The VAST Contributors
+// SPDX-License-Identifier: BSD-3-Clause
 
 #pragma once
 
 #include "vast/fwd.hpp"
 
+#include "vast/aliases.hpp"
 #include "vast/atoms.hpp"
 
+#include <caf/io/fwd.hpp>
 #include <caf/replies_to.hpp>
 
+#include <filesystem>
+
 #define VAST_ADD_TYPE_ID(type) CAF_ADD_TYPE_ID(vast_actors, type)
 
+// NOLINTNEXTLINE(cert-dcl58-cpp)
+namespace std::filesystem {
+
+// TODO: With CAF 0.18, drop this inspector that is—strictly speaking—undefined
+// behavior and replace it with a specialization of caf::inspector_access.
+template <class Inspector>
+typename Inspector::result_type
+inspect(Inspector& f, ::std::filesystem::path& x) {
+  auto str = x.string();
+  if constexpr (std::is_same_v<typename Inspector::result_type, void>) {
+    f(str);
+    if constexpr (Inspector::reads_state)
+      x = {str};
+    return;
+  } else {
+    auto result = f(str);
+    if constexpr (Inspector::reads_state)
+      x = {str};
+    return result;
+  }
+}
+
+} // namespace std::filesystem
+
 namespace vast::system {
 
-/// Helper utility that enables extending typed actor forwartd declarations
+/// Helper utility that enables extending typed actor forward declarations
 /// without including <caf/typed_actor.hpp>.
 template <class... Fs>
 struct typed_actor_fwd;
@@ -42,6 +65,7 @@ struct typed_actor_fwd {
   using extend_with = typename extend_with_helper<Handle>::type;
 
   using unwrap = caf::typed_actor<Fs...>;
+  using unwrap_as_broker = caf::io::typed_broker<Fs...>;
 };
 
 /// The STREAM SINK actor interface.
@@ -59,28 +83,16 @@ using flush_listener_actor = typed_actor_fwd<
   // Reacts to the requested flush message.
   caf::reacts_to<atom::flush>>::unwrap;
 
-/// The ARCHIVE CLIENT actor interface.
-using archive_client_actor = typed_actor_fwd<
-  // An ARCHIVE CLIENT receives table slices from the ARCHIVE for partial
-  // query hits.
-  caf::reacts_to<table_slice>,
-  // An ARCHIVE CLIENT receives (done, error) when the query finished.
-  caf::reacts_to<atom::done, caf::error>>::unwrap;
-
-/// The PARTITION CLIENT actor interface.
-using partition_client_actor = typed_actor_fwd<
-  // The client sends an expression to the partition and receives several sets
-  // of ids followed by a final `atom::done` which as sent as response to the
-  // expression. This interface provides the callback for the middle part of
-  // this sequence.
-  caf::reacts_to<ids>>::unwrap;
-
-/// The INDEX CLIENT actor interface.
-using index_client_actor = typed_actor_fwd<
-  // Receives done from the INDEX when the query finished.
-  caf::reacts_to<atom::done>>
-  // Receives ids from the INDEX for partial query hits.
-  ::extend_with<partition_client_actor>::unwrap;
+/// The RECEIVER SINK actor interface.
+/// This can be used to avoid defining an opaque alias for a single-handler
+/// interface.
+/// @tparam T The type of first parameter of the message handler the the actor
+///           handle must implement.
+/// @tparam Ts... The types of additional parameters for the message handler.
+template <class T, class... Ts>
+using receiver_actor = typename typed_actor_fwd<
+  // Add a new source.
+  typename caf::reacts_to<T, Ts...>>::unwrap;
 
 /// The STATUS CLIENT actor interface.
 using status_client_actor = typed_actor_fwd<
@@ -88,12 +100,25 @@ using status_client_actor = typed_actor_fwd<
   caf::replies_to<atom::status, status_verbosity>::with< //
     caf::dictionary<caf::config_value>>>::unwrap;
 
+/// The STORE actor interface.
+using store_actor = typed_actor_fwd<
+  // Handles an extraction for the given expression, optionally optimized by a
+  // set of ids to pre-select the events to evaluate.
+  caf::replies_to<query, ids>::with<atom::done>,
+  // Erase the events with the given ids.
+  caf::replies_to<atom::erase, ids>::with<atom::done>>::unwrap;
+
+/// The STORE BUILDER actor interface.
+using store_builder_actor = typed_actor_fwd<>::extend_with<store_actor>
+  // Conform to the protocol of the STREAM SINK actor for table slices.
+  ::extend_with<stream_sink_actor<table_slice>>
+  // Conform to the protocol of the STATUS CLIENT actor.
+  ::extend_with<status_client_actor>::unwrap;
+
 /// The PARTITION actor interface.
 using partition_actor = typed_actor_fwd<
-  // Evaluate the given expression, returning the relevant evaluation triples.
-  // TODO: Passing the `partition_client_actor` here is an historical artifact,
-  // a cleaner API would be to just return the evaluated `vast::ids`.
-  caf::replies_to<expression, partition_client_actor>::with<atom::done>>
+  // Evaluate the given expression and send the matching events to the receiver.
+  caf::replies_to<query>::with<atom::done>>
   // Conform to the procol of the STATUS CLIENT actor.
   ::extend_with<status_client_actor>::unwrap;
 
@@ -103,15 +128,14 @@ using query_map = std::vector<std::pair<uuid, partition_actor>>;
 
 /// The QUERY SUPERVISOR actor interface.
 using query_supervisor_actor = typed_actor_fwd<
-  /// Reacts to an expression and a set of relevant partitions by
-  /// sending several `vast::ids` to the index_client_actor, followed
-  /// by a final `atom::done`.
-  caf::reacts_to<expression, query_map, index_client_actor>>::unwrap;
+  /// Reacts to a query and a set of relevant partitions by sending several
+  /// `vast::ids` to the index_client_actor, followed by a final `atom::done`.
+  caf::reacts_to<query, query_map, receiver_actor<atom::done>>>::unwrap;
 
 /// The EVALUATOR actor interface.
 using evaluator_actor = typed_actor_fwd<
-  // Re-evaluates the expression and relays new hits to the PARTITION CLIENT.
-  caf::replies_to<partition_client_actor>::with<atom::done>>::unwrap;
+  // Evaluates the expression and responds with matching ids.
+  caf::replies_to<atom::run>::with<ids>>::unwrap;
 
 /// The INDEXER actor interface.
 using indexer_actor = typed_actor_fwd<
@@ -163,16 +187,30 @@ using query_supervisor_master_actor = typed_actor_fwd<
   // Enlist the QUERY SUPERVISOR as an available worker.
   caf::reacts_to<atom::worker, query_supervisor_actor>>::unwrap;
 
+/// The META INDEX actor interface.
+using meta_index_actor = typed_actor_fwd<
+  // Bulk import a set of partition synopses.
+  caf::replies_to<atom::merge, std::shared_ptr<std::map<
+                                 uuid, partition_synopsis>>>::with<atom::ok>,
+  // Merge a single partition synopsis.
+  caf::replies_to<atom::merge, uuid, std::shared_ptr<partition_synopsis>>::with< //
+    atom::ok>,
+  // Erase a single partition synopsis.
+  caf::replies_to<atom::erase, uuid>::with<atom::ok>,
+  // Evaluate the expression.
+  caf::replies_to<expression>::with< //
+    std::vector<uuid>>>::unwrap;
+
 /// The INDEX actor interface.
 using index_actor = typed_actor_fwd<
   // Triggered when the INDEX finished querying a PARTITION.
   caf::reacts_to<atom::done, uuid>,
-  // Registers the ARCHIVE with the ACCOUNTANT.
+  // Registers the INDEX with the ACCOUNTANT.
   caf::reacts_to<accountant_actor>,
   // Subscribes a FLUSH LISTENER to the INDEX.
   caf::reacts_to<atom::subscribe, atom::flush, flush_listener_actor>,
-  // Evaluatates an expression.
-  caf::reacts_to<expression>,
+  // Evaluatates an query.
+  caf::reacts_to<query>,
   // Queries PARTITION actors for a given query id.
   caf::reacts_to<uuid, uint32_t>,
   // Erases the given events from the INDEX, and returns their ids.
@@ -186,34 +224,20 @@ using index_actor = typed_actor_fwd<
 
 /// The ARCHIVE actor interface.
 using archive_actor = typed_actor_fwd<
-  // Register an exporter actor.
-  // TODO: This should probably take an archive_client_actor.
-  caf::reacts_to<atom::exporter, caf::actor>,
   // Registers the ARCHIVE with the ACCOUNTANT.
   caf::reacts_to<accountant_actor>,
-  // Starts handling a query for the given ids.
-  caf::reacts_to<ids, archive_client_actor>,
   // INTERNAL: Handles a query for the given ids, and sends the table slices
-  // back to the ARCHIVE CLIENT.
-  caf::reacts_to<atom::internal, ids, archive_client_actor, uint64_t>,
+  // back to the client.
+  caf::reacts_to<atom::internal, atom::resume>,
   // The internal telemetry loop of the ARCHIVE.
-  caf::reacts_to<atom::telemetry>,
-  // Erase the events with the given ids.
-  caf::replies_to<atom::erase, ids>::with< //
-    atom::done>>
-  // Conform to the protocol of the STREAM SINK actor for table slices.
-  ::extend_with<stream_sink_actor<table_slice>>
-  // Conform to the procotol of the STATUS CLIENT actor.
-  ::extend_with<status_client_actor>::unwrap;
+  caf::reacts_to<atom::telemetry>>
+  // Conform to the protocol of the STORE BUILDER actor.
+  ::extend_with<store_builder_actor>::unwrap;
 
 /// The TYPE REGISTRY actor interface.
 using type_registry_actor = typed_actor_fwd<
   // The internal telemetry loop of the TYPE REGISTRY.
   caf::reacts_to<atom::telemetry>,
-  // Registers the given type.
-  caf::reacts_to<atom::put, vast::type>,
-  // Registers all types in the given schema.
-  caf::reacts_to<atom::put, vast::schema>,
   // Retrieves all known types.
   caf::replies_to<atom::get>::with<type_set>,
   // Registers the given taxonomies.
@@ -249,21 +273,23 @@ using disk_monitor_actor = typed_actor_fwd<
 using filesystem_actor = typed_actor_fwd<
   // Writes a chunk of data to a given path. Creates intermediate directories
   // if needed.
-  caf::replies_to<atom::write, path, chunk_ptr>::with< //
+  caf::replies_to<atom::write, std::filesystem::path, chunk_ptr>::with< //
     atom::ok>,
   // Reads a chunk of data from a given path and returns the chunk.
-  caf::replies_to<atom::read, path>::with< //
+  caf::replies_to<atom::read, std::filesystem::path>::with< //
     chunk_ptr>,
   // Memory-maps a file.
-  caf::replies_to<atom::mmap, path>::with< //
+  caf::replies_to<atom::mmap, std::filesystem::path>::with< //
     chunk_ptr>>
   // Conform to the procotol of the STATUS CLIENT actor.
   ::extend_with<status_client_actor>::unwrap;
 
 /// The interface of an ACTIVE PARTITION actor.
 using active_partition_actor = typed_actor_fwd<
+  caf::reacts_to<atom::subscribe, atom::flush, flush_listener_actor>,
   // Persists the active partition at the specified path.
-  caf::replies_to<atom::persist, path, path>::with< //
+  caf::replies_to<atom::persist, std::filesystem::path,
+                  std::filesystem::path>::with< //
     std::shared_ptr<partition_synopsis>>,
   // INTERNAL: A repeatedly called continuation of the persist request.
   caf::reacts_to<atom::internal, atom::persist, atom::resume>>
@@ -280,31 +306,34 @@ using exporter_actor = typed_actor_fwd<
   caf::reacts_to<atom::extract, uint64_t>,
   // Register the ACCOUNTANT actor.
   caf::reacts_to<accountant_actor>,
-  // Register the ARCHIVE actor.
-  caf::reacts_to<archive_actor>,
   // Register the INDEX actor.
   caf::reacts_to<index_actor>,
   // Register the SINK actor.
   caf::reacts_to<atom::sink, caf::actor>,
   // Execute previously registered query.
   caf::reacts_to<atom::run>,
+  // Execute previously registered query.
+  caf::reacts_to<atom::done>,
+  // Execute previously registered query.
+  caf::reacts_to<table_slice>,
   // Register a STATISTICS SUBSCRIBER actor.
   caf::reacts_to<atom::statistics, caf::actor>>
   // Conform to the protocol of the STREAM SINK actor for table slices.
   ::extend_with<stream_sink_actor<table_slice>>
   // Conform to the protocol of the STATUS CLIENT actor.
-  ::extend_with<status_client_actor>
-  // Conform to the protocol of the ARCHIVE CLIENT actor.
-  ::extend_with<archive_client_actor>
-  // Conform to the protocol of the INDEX CLIENT actor.
-  ::extend_with<index_client_actor>::unwrap;
+  ::extend_with<status_client_actor>::unwrap;
 
-/// The interface of an ANALYZER PLUGIN actors.
+/// The interface of a COMPONENT PLUGIN actor.
+using component_plugin_actor = typed_actor_fwd<>
+  // Conform to the protocol of the STATUS CLIENT actor.
+  ::extend_with<status_client_actor>::unwrap;
+
+/// The interface of an ANALYZER PLUGIN actor.
 using analyzer_plugin_actor = typed_actor_fwd<>
   // Conform to the protocol of the STREAM SINK actor for table slices.
   ::extend_with<stream_sink_actor<table_slice>>
-  // Conform to the protocol of the STATUS CLIENT actor.
-  ::extend_with<status_client_actor>::unwrap;
+  // Conform to the protocol of the COMPONENT PLUGIN actor.
+  ::extend_with<component_plugin_actor>::unwrap;
 
 /// The interface of an IMPORTER actor.
 using importer_actor = typed_actor_fwd<
@@ -325,11 +354,50 @@ using importer_actor = typed_actor_fwd<
   // Conform to the protocol of the STATUS CLIENT actor.
   ::extend_with<status_client_actor>::unwrap;
 
+/// The interface of a SOURCE actor.
+using source_actor = typed_actor_fwd<
+  // Retrieve the currently used schema of the SOURCE.
+  caf::replies_to<atom::get, atom::schema>::with<schema>,
+  // Update the currently used schema of the SOURCE.
+  caf::reacts_to<atom::put, schema>,
+  // Update the expression used for filtering data in the SOURCE.
+  caf::reacts_to<expression>,
+  // Set up a new stream sink for the generated data.
+  caf::reacts_to<stream_sink_actor<table_slice, std::string>>,
+  // INTERNAL: Cause the source to wake up.
+  caf::reacts_to<atom::wakeup>,
+  // INTERNAL: Telemetry loop handler.
+  caf::reacts_to<atom::telemetry>>
+  // Conform to the protocol of the STATUS CLIENT actor.
+  ::extend_with<status_client_actor>::unwrap;
+
+/// The interface of a DATAGRAM SOURCE actor.
+using datagram_source_actor
+  // Reacts to datagram messages.
+  = typed_actor_fwd<caf::reacts_to<caf::io::new_datagram_msg>>
+  // Conform to the protocol of the SOURCE actor.
+  ::extend_with<source_actor>::unwrap_as_broker;
+
+/// The interface of an TRANSFORMER actor.
+using transformer_actor = typed_actor_fwd<
+  // Send transformed slices to this sink.
+  caf::replies_to<stream_sink_actor<table_slice>>::with< //
+    caf::outbound_stream_slot<table_slice>>,
+  // Send transformed slices to this sink; pass the string through along with
+  // the stream handshake.
+  caf::reacts_to<stream_sink_actor<table_slice, std::string>, std::string>>
+  // Conform to the protocol of the STREAM SINK actor for framed table slices
+  ::extend_with<stream_sink_actor<detail::framed<table_slice>>>
+  // Conform to the protocol of the STATUS CLIENT actor.
+  ::extend_with<status_client_actor>::unwrap;
+
 /// The interface of the NODE actor.
 using node_actor = typed_actor_fwd<
   // Run an invocation in the node.
   caf::replies_to<atom::run, invocation>::with< //
     caf::message>,
+  // INTERNAL: Spawn component plugins.
+  caf::reacts_to<atom::internal, atom::spawn, atom::plugin>,
   // Run an invocation in the node that spawns an actor.
   caf::replies_to<atom::spawn, invocation>::with< //
     caf::actor>,
@@ -346,8 +414,7 @@ using node_actor = typed_actor_fwd<
   caf::replies_to<atom::get, atom::label, std::vector<std::string>>::with< //
     std::vector<caf::actor>>,
   // Retrieve the version of the process running the NODE.
-  caf::replies_to<atom::get, atom::version>::with< //
-    std::string>,
+  caf::replies_to<atom::get, atom::version>::with<record>,
   // Handle a signal.
   // TODO: Make this a signal_monitor_client_actor
   caf::reacts_to<atom::signal, int>>::unwrap;
@@ -358,12 +425,13 @@ using node_actor = typed_actor_fwd<
 
 CAF_BEGIN_TYPE_ID_BLOCK(vast_actors, caf::id_block::vast_atoms::end)
 
+  VAST_ADD_TYPE_ID((std::filesystem::path))
+
   VAST_ADD_TYPE_ID((vast::system::accountant_actor))
   VAST_ADD_TYPE_ID((vast::system::active_indexer_actor))
   VAST_ADD_TYPE_ID((vast::system::active_partition_actor))
   VAST_ADD_TYPE_ID((vast::system::analyzer_plugin_actor))
   VAST_ADD_TYPE_ID((vast::system::archive_actor))
-  VAST_ADD_TYPE_ID((vast::system::archive_client_actor))
   VAST_ADD_TYPE_ID((vast::system::disk_monitor_actor))
   VAST_ADD_TYPE_ID((vast::system::evaluator_actor))
   VAST_ADD_TYPE_ID((vast::system::exporter_actor))
@@ -371,20 +439,28 @@ CAF_BEGIN_TYPE_ID_BLOCK(vast_actors, caf::id_block::vast_atoms::end)
   VAST_ADD_TYPE_ID((vast::system::flush_listener_actor))
   VAST_ADD_TYPE_ID((vast::system::importer_actor))
   VAST_ADD_TYPE_ID((vast::system::index_actor))
-  VAST_ADD_TYPE_ID((vast::system::index_client_actor))
   VAST_ADD_TYPE_ID((vast::system::indexer_actor))
   VAST_ADD_TYPE_ID((vast::system::node_actor))
   VAST_ADD_TYPE_ID((vast::system::partition_actor))
-  VAST_ADD_TYPE_ID((vast::system::partition_client_actor))
   VAST_ADD_TYPE_ID((vast::system::query_map))
   VAST_ADD_TYPE_ID((vast::system::query_supervisor_actor))
   VAST_ADD_TYPE_ID((vast::system::query_supervisor_master_actor))
+  VAST_ADD_TYPE_ID((vast::system::receiver_actor<vast::atom::done>))
   VAST_ADD_TYPE_ID((vast::system::status_client_actor))
-  VAST_ADD_TYPE_ID((vast::system::stream_sink_actor<vast::table_slice>) )
+  VAST_ADD_TYPE_ID((vast::system::stream_sink_actor<vast::table_slice>))
   VAST_ADD_TYPE_ID(
-    (vast::system::stream_sink_actor<vast::table_slice, std::string>) )
+    (vast::system::stream_sink_actor<vast::table_slice, std::string>))
   VAST_ADD_TYPE_ID((vast::system::type_registry_actor))
 
 CAF_END_TYPE_ID_BLOCK(vast_actors)
 
+// Used in the interface of the meta_index actor.
+// We can't provide a meaningful implementation of `inspect()` for a shared_ptr,
+// so so we add these as `UNSAFE_MESSAGE_TYPE` to assure caf that they will
+// never be sent over the network.
+#define vast_uuid_synopsis_map std::map<vast::uuid, vast::partition_synopsis>
+CAF_ALLOW_UNSAFE_MESSAGE_TYPE(std::shared_ptr<vast_uuid_synopsis_map>)
+CAF_ALLOW_UNSAFE_MESSAGE_TYPE(std::shared_ptr<vast::partition_synopsis>)
+#undef vast_uuid_synopsis_map
+
 #undef VAST_ADD_TYPE_ID
