@@ -1,15 +1,10 @@
-/******************************************************************************
- *                    _   _____   __________                                  *
- *                   | | / / _ | / __/_  __/     Visibility                   *
- *                   | |/ / __ |_\ \  / /          Across                     *
- *                   |___/_/ |_/___/ /_/       Space and Time                 *
- *                                                                            *
- * This file is part of VAST. It is subject to the license terms in the       *
- * LICENSE file found in the top-level directory of this distribution and at  *
- * http://vast.io/license. No part of VAST, including this file, may be       *
- * copied, modified, propagated, or distributed except according to the terms *
- * contained in the LICENSE file.                                             *
- ******************************************************************************/
+//    _   _____   __________
+//   | | / / _ | / __/_  __/     Visibility
+//   | |/ / __ |_\ \  / /          Across
+//   |___/_/ |_/___/ /_/       Space and Time
+//
+// SPDX-FileCopyrightText: (c) 2016 The VAST Contributors
+// SPDX-License-Identifier: BSD-3-Clause
 
 #pragma once
 
@@ -17,6 +12,7 @@
 
 #include "vast/system/actors.hpp"
 #include "vast/system/instrumentation.hpp"
+#include "vast/transform.hpp"
 
 #include <caf/behavior.hpp>
 #include <caf/event_based_actor.hpp>
@@ -38,6 +34,7 @@ struct sink_state {
   accountant_actor accountant;
   vast::system::measurement measurement;
   format::writer_ptr writer;
+  transformation_engine transforms;
   const char* name = "writer";
 
   explicit sink_state(caf::event_based_actor* self_ptr);
@@ -48,4 +45,9 @@ struct sink_state {
 caf::behavior sink(caf::stateful_actor<sink_state>* self,
                    format::writer_ptr&& writer, uint64_t max_events);
 
+caf::behavior
+transforming_sink(caf::stateful_actor<sink_state>* self,
+                  format::writer_ptr&& writer,
+                  std::vector<transform>&& transforms, uint64_t max_events);
+
 } // namespace vast::system
