@@ -1,15 +1,10 @@
-/******************************************************************************
- *                    _   _____   __________                                  *
- *                   | | / / _ | / __/_  __/     Visibility                   *
- *                   | |/ / __ |_\ \  / /          Across                     *
- *                   |___/_/ |_/___/ /_/       Space and Time                 *
- *                                                                            *
- * This file is part of VAST. It is subject to the license terms in the       *
- * LICENSE file found in the top-level directory of this distribution and at  *
- * http://vast.io/license. No part of VAST, including this file, may be       *
- * copied, modified, propagated, or distributed except according to the terms *
- * contained in the LICENSE file.                                             *
- ******************************************************************************/
+//    _   _____   __________
+//   | | / / _ | / __/_  __/     Visibility
+//   | |/ / __ |_\ \  / /          Across
+//   |___/_/ |_/___/ /_/       Space and Time
+//
+// SPDX-FileCopyrightText: (c) 2016 The VAST Contributors
+// SPDX-License-Identifier: BSD-3-Clause
 
 #pragma once
 
@@ -23,23 +18,25 @@
 #include "vast/detail/operators.hpp"
 #include "vast/detail/range.hpp"
 #include "vast/detail/stack_vector.hpp"
+#include "vast/detail/type_traits.hpp"
 #include "vast/offset.hpp"
 #include "vast/operator.hpp"
-#include "vast/optional.hpp"
 #include "vast/time.hpp"
 
+#include <caf/detail/apply_args.hpp>
+#include <caf/detail/int_list.hpp>
 #include <caf/detail/type_list.hpp>
 #include <caf/error.hpp>
-#include <caf/expected.hpp>
 #include <caf/fwd.hpp>
 #include <caf/intrusive_cow_ptr.hpp>
 #include <caf/make_counted.hpp>
 #include <caf/meta/omittable.hpp>
 #include <caf/none.hpp>
 #include <caf/ref_counted.hpp>
-#include <caf/variant.hpp>
+#include <caf/sum_type.hpp>
 
 #include <functional>
+#include <optional>
 #include <string>
 #include <type_traits>
 #include <vector>
@@ -76,7 +73,7 @@ constexpr type_id_type invalid_type_id = -1;
 
 template <class T>
 constexpr type_id_type type_id() {
-  static_assert(caf::detail::tl_contains<concrete_types, T>::value,
+  static_assert(detail::contains_type_v<concrete_types, T>,
                 "type IDs only available for concrete types");
   return caf::detail::tl_index_of<concrete_types, T>::value;
 }
@@ -100,8 +97,8 @@ public:
   /// Constructs a type from a concrete instance.
   /// @tparam T a type that derives from @ref abstract_type.
   /// @param x An instance of a type.
-  template <class T, class = std::enable_if_t<
-                       caf::detail::tl_contains<concrete_types, T>::value>>
+  template <class T,
+            class = std::enable_if_t<detail::contains_type_v<concrete_types, T>>>
   type(T x) : ptr_{caf::make_counted<T>(std::move(x))} {
     // nop
   }
@@ -119,8 +116,8 @@ public:
   type& operator=(type&&) noexcept = default;
 
   /// Assigns a type from another instance
-  template <class T, class = std::enable_if_t<
-                       caf::detail::tl_contains<concrete_types, T>::value>>
+  template <class T,
+            class = std::enable_if_t<detail::contains_type_v<concrete_types, T>>>
   type& operator=(T x) {
     ptr_ = caf::make_counted<T>(std::move(x));
     return *this;
@@ -161,16 +158,16 @@ public:
   explicit operator bool() const;
 
   /// @returns the name of the type.
-  const std::string& name() const;
+  [[nodiscard]] const std::string& name() const;
 
   /// @returns The attributes of the type.
-  const std::vector<attribute>& attributes() const;
+  [[nodiscard]] const std::vector<attribute>& attributes() const;
 
   /// @cond PRIVATE
 
-  abstract_type_ptr ptr() const;
+  [[nodiscard]] abstract_type_ptr ptr() const;
 
-  const abstract_type* raw_ptr() const noexcept;
+  [[nodiscard]] const abstract_type* raw_ptr() const noexcept;
 
   const abstract_type* operator->() const noexcept;
 
@@ -304,6 +301,32 @@ public:
     return std::move(derived());
   }
 
+  Derived& update_attributes(std::vector<attribute> xs) & {
+    auto& attrs = this->attributes_;
+    for (auto& x : xs) {
+      auto i = std::find_if(attrs.begin(), attrs.end(),
+                            [&](auto& attr) { return attr.key == x.key; });
+      if (i == attrs.end())
+        attrs.push_back(std::move(x));
+      else
+        i->value = std::move(x).value;
+    }
+    return derived();
+  }
+
+  Derived update_attributes(std::vector<attribute> xs) && {
+    auto& attrs = this->attributes_;
+    for (auto& x : xs) {
+      auto i = std::find_if(attrs.begin(), attrs.end(),
+                            [&](auto& attr) { return attr.key == x.key; });
+      if (i == attrs.end())
+        attrs.push_back(std::move(x));
+      else
+        i->value = std::move(x).value;
+    }
+    return std::move(derived());
+  }
+
   friend bool operator==(const Derived& x, const Derived& y) {
     return x.equals(y);
   }
@@ -392,7 +415,7 @@ private:
 /// @relates type
 template <class Derived>
 struct basic_type : concrete_type<Derived> {
-  type_flags flags() const noexcept final {
+  [[nodiscard]] type_flags flags() const noexcept final {
     return type_flags::basic;
   }
 };
@@ -401,7 +424,7 @@ struct basic_type : concrete_type<Derived> {
 /// @relates basic_type type
 template <class Derived>
 struct complex_type : concrete_type<Derived> {
-  type_flags flags() const noexcept override {
+  [[nodiscard]] type_flags flags() const noexcept override {
     return type_flags::complex | type_flags::recursive;
   }
 };
@@ -412,7 +435,7 @@ template <class Derived>
 struct recursive_type : complex_type<Derived> {
   using super = complex_type<Derived>;
 
-  type_flags flags() const noexcept override {
+  [[nodiscard]] type_flags flags() const noexcept override {
     return type_flags::complex | type_flags::recursive;
   }
 };
@@ -435,12 +458,12 @@ struct nested_type : recursive_type<Derived> {
              x.value_type);
   }
 
-  bool equals(const abstract_type& other) const final {
+  [[nodiscard]] bool equals(const abstract_type& other) const final {
     return super::equals(other)
            && value_type == super::downcast(other).value_type;
   }
 
-  bool less_than(const abstract_type& other) const final {
+  [[nodiscard]] bool less_than(const abstract_type& other) const final {
     return super::less_than(other)
            && value_type < super::downcast(other).value_type;
   }
@@ -597,9 +620,9 @@ struct record_type final : recursive_type<record_type> {
   class each : public detail::range_facade<each> {
   public:
     struct range_state {
-      std::string key() const;
-      const class type& type() const;
-      size_t depth() const;
+      [[nodiscard]] std::string key() const;
+      [[nodiscard]] const class type& type() const;
+      [[nodiscard]] size_t depth() const;
 
       detail::stack_vector<const record_field*, 64> trace;
       vast::offset offset;
@@ -611,8 +634,8 @@ struct record_type final : recursive_type<record_type> {
     friend detail::range_facade<each>;
 
     void next();
-    bool done() const;
-    const range_state& get() const;
+    [[nodiscard]] bool done() const;
+    [[nodiscard]] const range_state& get() const;
 
     range_state state_;
     detail::stack_vector<const record_type*, 64> records_;
@@ -634,12 +657,12 @@ struct record_type final : recursive_type<record_type> {
   /// Attemps to resolve a key to an offset.
   /// @param key The key to resolve.
   /// @returns The offset corresponding to *key*.
-  caf::optional<offset> resolve(std::string_view key) const;
+  std::optional<offset> resolve(std::string_view key) const;
 
   /// Attemps to resolve an offset to a key.
   /// @param o The offset to resolve.
   /// @returns The key corresponding to *o*.
-  caf::optional<std::string> resolve(const offset& o) const;
+  std::optional<std::string> resolve(const offset& o) const;
 
   /// Finds a record field by exact name.
   /// @param field_name The name of the field to lookup.
@@ -666,6 +689,10 @@ struct record_type final : recursive_type<record_type> {
   const record_field* at(const offset& o) const&;
   const record_field* at(const offset& o) && = delete;
 
+  /// Replaces the field at a given offset with `field`.
+  [[nodiscard]] caf::expected<record_type>
+  assign(const offset& o, const record_field& field) const;
+
   /// Returns the field at the given offset with the full name as if the record
   /// was flattened.
   /// @param o The offset to resolve.
@@ -673,15 +700,15 @@ struct record_type final : recursive_type<record_type> {
   ///   auto r = record_type{{"x", record_type{"y", count_type{}}}};
   ///   ASSERT_EQ(r.flat_field_at({0,0}), record_field{"x.y", count_type{}});
   /// @endcode
-  caf::optional<record_field> flat_field_at(offset o) const;
+  std::optional<record_field> flat_field_at(offset o) const;
 
   /// Converts an offset into an index for the flattened representation.
   /// @param o The offset to resolve.
-  caf::optional<size_t> flat_index_at(offset o) const;
+  std::optional<size_t> flat_index_at(offset o) const;
 
   /// Converts an index for the flattened representation into an offset.
   /// @param i The index to resolve.
-  caf::optional<offset> offset_from_index(size_t i) const;
+  std::optional<offset> offset_from_index(size_t i) const;
 
   friend bool operator==(const record_type& x, const record_type& y);
   friend bool operator<(const record_type& x, const record_type& y);
@@ -718,9 +745,47 @@ record_type concat(const Rs&... rs) {
   (result.fields.insert(result.fields.end(), rs.fields.begin(),
                         rs.fields.end()),
    ...);
+  // TODO: This function is missing an integrity check that makes sure the
+  // result does not contain multiple fields with the same name.
+  // We should also add a differently named version that deduplicates completely
+  // identical fields and recurses into nested records under the same field
+  // name.
   return result;
 }
 
+/// Creates a new unnamed record_type containing the fields and attribues of lhs
+/// and rhs. Errors if a field of the same name but different types is present
+/// in both inputs. Errors is the inputs disagree over the value of an attribute
+/// with a certain name.
+/// @returns The combined record_type.
+/// @relates record_type
+caf::expected<record_type>
+merge(const record_type& lhs, const record_type& rhs);
+
+/// @relates priority_merge
+enum class merge_policy { prefer_left, prefer_right };
+
+/// Creates a new unnamed record_type containing the fields and attribues of lhs
+/// and rhs. Uses a merge_policy to decide wheter to use a field from lhs or rhs
+/// in case of a conflict.
+/// @returns The combined record_type.
+/// @relates record_type merge_policy
+record_type
+priority_merge(const record_type& lhs, const record_type& rhs, merge_policy p);
+
+/// Removes a field from a record_type by name.
+/// @param r The record to mutate.
+/// @param path The sequence of keys pointing to the target field.
+/// @returns A new type without the target field if it exists in `r`.
+/// @pre `!path.empty()`
+/// @relates record_type
+std::optional<record_type>
+remove_field(const record_type& r, std::vector<std::string_view> path);
+
+/// As above, but use an offset instead of a vector of string to specify
+/// the field to be removed.
+std::optional<record_type> remove_field(const record_type& r, offset o);
+
 /// Recursively flattens the arguments of a record type.
 /// @param rec The record to flatten.
 /// @returns The flattened record type.
@@ -744,15 +809,6 @@ size_t flat_size(const record_type& rec);
 /// Computes the size of a flat representation of `rec`.
 size_t flat_size(const type&);
 
-/// Unflattens a flattened record type.
-/// @param rec the record to unflatten.
-/// @returns The unflattened record type.
-/// @relates record_type
-record_type unflatten(const record_type& rec);
-
-/// @relates type record_type
-type unflatten(const type& t);
-
 // -- helpers ----------------------------------------------------------------
 
 /// Maps a concrete type to a corresponding data type.
@@ -783,6 +839,11 @@ VAST_TYPE_TRAIT(map);
 
 #undef VAST_TYPE_TRAIT
 
+template <>
+struct type_traits<vast::type> {
+  using data_type = vast::data;
+};
+
 template <>
 struct type_traits<none_type> {
   using data_type = caf::none_t;
@@ -904,9 +965,7 @@ bool has_attribute(const Type& t, std::string_view key) {
 
 /// Tests whether a type has a "skip" attribute.
 /// @relates has_attribute type
-inline bool has_skip_attribute(const type& t) {
-  return has_attribute(t, "skip");
-}
+bool has_skip_attribute(const type& t);
 
 /// @relates type
 bool convert(const type& t, data& d);
@@ -942,6 +1001,19 @@ struct sum_type_access<vast::type> {
     return static_cast<const T&>(*x);
   }
 
+  // NOTE: This overload for get_if breaks the shared abstract type semantics.
+  // This was noticed when the content of the type registry was mutated without
+  // it's knowledge when type operations on layouts are in play. The exact bug
+  // is not clear yet, so we'll refrain from using mutations on type value and
+  // construct new ones instead.
+  // template <class T, int Pos>
+  // static T* get_if(vast::type* x, sum_type_token<T, Pos>) {
+  //  x->ptr().unshare();
+  //  auto ptr = x->raw_ptr();
+  //  return ptr->index() == Pos ? const_cast<T*>(static_cast<const T*>(ptr))
+  //                             : nullptr;
+  //}
+
   template <class T, int Pos>
   static const T* get_if(const vast::type* x, sum_type_token<T, Pos>) {
     auto ptr = x->raw_ptr();
