@@ -1,42 +1,38 @@
-/******************************************************************************
- *                    _   _____   __________                                  *
- *                   | | / / _ | / __/_  __/     Visibility                   *
- *                   | |/ / __ |_\ \  / /          Across                     *
- *                   |___/_/ |_/___/ /_/       Space and Time                 *
- *                                                                            *
- * This file is part of VAST. It is subject to the license terms in the       *
- * LICENSE file found in the top-level directory of this distribution and at  *
- * http://vast.io/license. No part of VAST, including this file, may be       *
- * copied, modified, propagated, or distributed except according to the terms *
- * contained in the LICENSE file.                                             *
- ******************************************************************************/
+//    _   _____   __________
+//   | | / / _ | / __/_  __/     Visibility
+//   | |/ / __ |_\ \  / /          Across
+//   |___/_/ |_/___/ /_/       Space and Time
+//
+// SPDX-FileCopyrightText: (c) 2020 The VAST Contributors
+// SPDX-License-Identifier: BSD-3-Clause
 
-#include "vast/chunk.hpp"
-#include "vast/concept/printable/to_string.hpp"
-#include "vast/concept/printable/vast/type.hpp"
-#include "vast/concept/printable/vast/uuid.hpp"
-#include "vast/directory.hpp"
-#include "vast/fbs/index.hpp"
-#include "vast/fbs/partition.hpp"
-#include "vast/fbs/segment.hpp"
-#include "vast/fbs/utils.hpp"
-#include "vast/ids.hpp"
-#include "vast/io/read.hpp"
-#include "vast/path.hpp"
-#include "vast/qualified_record_field.hpp"
-#include "vast/table_slice.hpp"
-#include "vast/type.hpp"
-#include "vast/uuid.hpp"
+#include <vast/chunk.hpp>
+#include <vast/concept/printable/to_string.hpp>
+#include <vast/concept/printable/vast/type.hpp>
+#include <vast/concept/printable/vast/uuid.hpp>
+#include <vast/error.hpp>
+#include <vast/fbs/index.hpp>
+#include <vast/fbs/partition.hpp>
+#include <vast/fbs/segment.hpp>
+#include <vast/fbs/utils.hpp>
+#include <vast/ids.hpp>
+#include <vast/io/read.hpp>
+#include <vast/qualified_record_field.hpp>
+#include <vast/table_slice.hpp>
+#include <vast/type.hpp>
+#include <vast/uuid.hpp>
 
 #include <caf/binary_deserializer.hpp>
-
+#include <caf/error.hpp>
 #include <flatbuffers/flatbuffers.h>
 
 #include <cstddef>
+#include <filesystem>
 #include <iomanip>
 #include <iostream>
 #include <sstream>
 #include <string>
+#include <system_error>
 
 // clang-format off
 // TODO: Implement different output formats for human-readable and
@@ -113,13 +109,19 @@ struct formatting_options {
 
 struct indentation;
 
-typedef void (*printer)(vast::path, indentation&, const formatting_options&);
+typedef void (*printer)(const std::filesystem::path&, indentation&,
+                        const formatting_options&);
 
-void print_unknown(vast::path, indentation&, const formatting_options&);
-void print_vast_db(vast::path, indentation&, const formatting_options&);
-void print_partition(vast::path, indentation&, const formatting_options&);
-void print_index(vast::path, indentation&, const formatting_options&);
-void print_segment(vast::path, indentation&, const formatting_options&);
+void print_unknown(const std::filesystem::path&, indentation&,
+                   const formatting_options&);
+void print_vast_db(const std::filesystem::path&, indentation&,
+                   const formatting_options&);
+void print_partition(const std::filesystem::path&, indentation&,
+                     const formatting_options&);
+void print_index(const std::filesystem::path&, indentation&,
+                 const formatting_options&);
+void print_segment(const std::filesystem::path&, indentation&,
+                   const formatting_options&);
 
 static const std::map<Kind, printer> printers = {
   {Kind::Unknown, print_unknown}, {Kind::DatabaseDir, print_vast_db},
@@ -127,10 +129,28 @@ static const std::map<Kind, printer> printers = {
   {Kind::Segment, print_segment},
 };
 
-Kind classify(vast::path path) {
-  if (path.is_directory() && path.basename() == "vast.db")
-    return Kind::DatabaseDir;
-  if (!path.is_regular_file())
+caf::expected<Kind> classify(const std::filesystem::path& path) {
+  std::error_code err{};
+  const auto is_directory = std::filesystem::is_directory(path, err);
+  if (err)
+    return caf::make_error(vast::ec::filesystem_error,
+                           "Invalid path: " + path.string(), err.message());
+  if (is_directory) {
+    auto version_file = path / "VERSION";
+    auto is_regular_file = std::filesystem::is_regular_file(version_file, err);
+    if (err)
+      return caf::make_error(vast::ec::filesystem_error,
+                             "Invalid path: " + version_file.string(),
+                             err.message());
+    if (is_regular_file)
+      return Kind::DatabaseDir;
+  }
+  const auto is_regular_file = std::filesystem::is_regular_file(path, err);
+  if (err)
+    return caf::make_error(vast::ec::filesystem_error,
+                           "Not a file: ", err.message());
+
+  if (!is_regular_file)
     return Kind::Unknown;
   auto bytes = vast::io::read(path);
   if (!bytes)
@@ -169,7 +189,7 @@ struct flatbuffer_deleter {
 // automatically upon destruction.
 template <typename T>
 std::unique_ptr<const T, flatbuffer_deleter<T>>
-read_flatbuffer_file(vast::path path) {
+read_flatbuffer_file(const std::filesystem::path& path) {
   using result_t = std::unique_ptr<const T, flatbuffer_deleter<T>>;
   auto result
     = result_t(static_cast<const T*>(nullptr), flatbuffer_deleter<T>{});
@@ -177,6 +197,10 @@ read_flatbuffer_file(vast::path path) {
   if (!maybe_bytes)
     return result;
   auto bytes = std::move(*maybe_bytes);
+  flatbuffers::Verifier verifier{reinterpret_cast<const uint8_t*>(bytes.data()),
+                                 bytes.size()};
+  if (!verifier.template VerifyBuffer<T>())
+    return result;
   const auto* ptr = flatbuffers::GetRoot<T>(bytes.data());
   return result_t(ptr, flatbuffer_deleter<T>(std::move(bytes)));
 }
@@ -229,9 +253,9 @@ std::ostream& operator<<(std::ostream& out, const vast::fbs::uuid::v0* uuid) {
   return out;
 }
 
-void print_unknown(vast::path path, indentation& indent,
+void print_unknown(const std::filesystem::path& path, indentation& indent,
                    const formatting_options&) {
-  std::cout << indent << "(unknown " << path.str() << ")\n";
+  std::cout << indent << "(unknown " << path.string() << ")\n";
 }
 
 std::string print_bytesize(size_t bytes, const formatting_options& formatting) {
@@ -258,32 +282,51 @@ std::string print_bytesize(size_t bytes, const formatting_options& formatting) {
   return std::move(ss).str();
 }
 
-void print_vast_db(vast::path vast_db, indentation& indent,
+void print_vast_db(const std::filesystem::path& vast_db, indentation& indent,
                    const formatting_options& formatting) {
   // TODO: We should have some versioning for the layout
   // of the vast.db directory itself, so we can still read
   // older versions.
   auto index_dir = vast_db / "index";
-  std::cout << indent << index_dir.str() << "/\n";
+  std::cout << indent << index_dir.string() << "/\n";
   {
     indented_scope _(indent);
     std::cout << indent << "index.bin - ";
     print_index(index_dir / "index.bin", indent, formatting);
-    for (auto file : vast::directory{index_dir}) {
-      auto stem = file.basename(true).str();
-      if (stem == "index")
-        continue;
-      std::cout << indent << stem << " - ";
-      print_partition(file, indent, formatting);
+    std::error_code err{};
+    auto dir = std::filesystem::directory_iterator{index_dir, err};
+    if (err) {
+      std::cerr << "Failed to find vast db index directory: " + err.message()
+                << std::endl;
+    } else {
+      for (const auto& entry : dir) {
+        const auto stem = entry.path().stem();
+        if (stem == "index")
+          continue;
+        const auto extension = entry.path().extension();
+        // TODO: Print partition synopses.
+        if (extension == ".mdx")
+          continue;
+        std::cout << indent << stem << " - ";
+        print_partition(entry.path(), indent, formatting);
+      }
     }
   }
-  auto segments_dir = vast_db / "archive" / "segments";
-  std::cout << indent << segments_dir.str() << "/\n";
+  const auto segments_dir = vast_db / "archive" / "segments";
+  std::cout << indent << segments_dir.string() << "/\n";
   {
     indented_scope _(indent);
-    for (auto file : vast::directory{segments_dir}) {
-      std::cout << indent << file.basename(true).str() << " - ";
-      print_segment(file, indent, formatting);
+    std::error_code err{};
+    auto dir = std::filesystem::directory_iterator{segments_dir, err};
+    if (err) {
+      std::cerr << "Failed to find vast db segments directory: " + err.message()
+                << std::endl;
+    } else {
+      for (const auto& entry : dir) {
+        const auto stem = entry.path().stem();
+        std::cout << indent << stem << " - ";
+        print_segment(entry.path(), indent, formatting);
+      }
     }
   }
 }
@@ -375,11 +418,12 @@ void print_partition_v0(const vast::fbs::partition::v0* partition,
   }
 }
 
-void print_partition(vast::path path, indentation& indent,
+void print_partition(const std::filesystem::path& path, indentation& indent,
                      const formatting_options& formatting) {
   auto partition = read_flatbuffer_file<vast::fbs::Partition>(path);
   if (!partition) {
-    std::cout << "(error reading partition file " << path.str() << ")\n";
+    std::cout << "(error reading partition file " << path.string() << ")\n";
+    return;
   }
   switch (partition->partition_type()) {
     case vast::fbs::partition::Partition::v0:
@@ -421,11 +465,12 @@ void print_index_v0(const vast::fbs::index::v0* index, indentation& indent,
   std::cout << "\n";
 }
 
-void print_index(vast::path path, indentation& indent,
+void print_index(const std::filesystem::path& path, indentation& indent,
                  const formatting_options& formatting) {
   auto index = read_flatbuffer_file<vast::fbs::Index>(path);
   if (!index) {
-    std::cout << indent << "(error reading index file " << path.str() << ")\n";
+    std::cout << indent << "(error reading index file " << path.string()
+              << ")\n";
   }
   switch (index->index_type()) {
     case vast::fbs::index::Index::v0:
@@ -475,11 +520,11 @@ void print_segment_v0(const vast::fbs::segment::v0* segment,
   }
 }
 
-void print_segment(vast::path path, indentation& indent,
+void print_segment(const std::filesystem::path& path, indentation& indent,
                    const formatting_options& formatting) {
   auto segment = read_flatbuffer_file<vast::fbs::Segment>(path);
   if (!segment) {
-    std::cout << "(error reading segment file " << path.str() << ")\n";
+    std::cout << "(error reading segment file " << path.string() << ")\n";
   }
   switch (segment->segment_type()) {
     case vast::fbs::segment::Segment::v0:
@@ -518,14 +563,19 @@ int main(int argc, char** argv) {
   }
   if (raw_path.back() == '/')
     raw_path.resize(raw_path.size() - 1);
-  auto path = vast::path{raw_path};
-  auto kind = classify(path);
+  const auto path = std::filesystem::path{raw_path};
+  const auto kind = classify(path);
+  if (!kind) {
+    std::cerr << "Filesystem error with error code: " << kind.error().code()
+              << std::endl;
+    return 1;
+  }
   if (kind == Kind::Unknown) {
     std::cerr << "Could not determine type of " << argv[1] << std::endl;
     return 1;
   }
   struct indentation indent;
-  auto printer = printers.at(kind);
+  auto printer = printers.at(*kind);
   printer(path, indent, format);
   return 0;
 }
