@@ -115,10 +115,10 @@ def try_wait(process, timeout, expected_result):
             return Result.ERROR
         return Result.SUCCESS
     except subprocess.TimeoutExpired:
-        log = LOGGER.error
-        if expected_result != Result.TIMEOUT:
-            log = LOGGER.debug
-        log(f"timeout reached, terminating process")
+        if expected_result == Result.TIMEOUT:
+            LOGGER.debug(f"expected timeout reached, terminating process")
+        else:
+            LOGGER.error(f"timeout reached, terminating process")
         process.terminate()
         return Result.TIMEOUT
 
@@ -316,7 +316,7 @@ class Server:
         self.name = name
         self.cwd = work_dir / self.name
         self.port = port
-        command = [self.app]
+        command = [self.app, "--bare-mode"]
         if self.config_arg:
             command.append(self.config_arg)
         command = command + args
@@ -694,7 +694,7 @@ def main():
     LOGGER.addHandler(sh)
     # Load test set.
     if not args.set:
-        args.set = Path(__file__).resolve().parent / "default_set.yaml"
+        args.set = Path(__file__).resolve().parent / "vast_integration_suite.yaml"
     args.set = args.set.resolve()
     LOGGER.debug(f"resolved test set path to {args.set}")
     global SET_DIR
